#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Biometric database score manipulation

Manipulate biometric database of scores.
"""

# imports
import csv
import sys
import argparse
import numpy as np
import pandas as pd
import random

GALLERY = 0
PROBE = 1
SAMPLE = 2
SCORE = 3

def read_stream(stream):
    for row in csv.reader(stream):
        yield row

def output_data(data):
    writer = csv.writer(sys.stdout)
    writer.writerows(data)

def remove_gallery_individual(data, galleryid):
    for row in data:
        if int(row[GALLERY]) != int(galleryid):
            yield row

def remove_probe_individual(data, probeid):
    for row in data:
        if int(row[PROBE]) != int(probeid):
            yield row

def remove_individual(data, id):
    data = remove_gallery_individual(data, id)
    data = remove_probe_individual(data, id)
    return data

def remove_sample(data, probeid, sampleid):
    for row in data:
        if int(row[PROBE]) != int(probeid) and \
           int(row[SAMPLE]) != int(sampleid):
            yield row

def remove_comparison(data, galleryid, probeid, sampleid):
    for row in data:
        if int(row[GALLERY]) != int(galleryid) and \
           int(row[PROBE]) != int(probeid) and \
           int(row[SAMPLE]) != int(sampleid):
            yield row


def filterData(data, galleryid, probeid, sampleid):


    for row in data:
        if galleryid != '?' and galleryid != row[GALLERY]:
	    continue
        elif probeid != '?' and probeid != row[PROBE]:
	    continue
        elif sampleid != '?' and sampleid != row[SAMPLE]:
	    continue
	else:
	    yield row


def summary(data):
    df = pd.DataFrame(data)

    gallery = np.unique(df[GALLERY])
    probe = np.unique(df[PROBE])
    common = np.intersect1d(gallery, probe)
    all = np.unique(np.concatenate((gallery, probe)))


    print "Nb different individuals:", len(all)
    print "Nb individuals in gallery: ", len(gallery)
    print "Nb individuals in probe: ", len(probe)
    print "Nb individuals both in gallery AND probe", len(common)



def generateThumbnails(data, thumbnail):
    done = set()
    for row in data:
        key = (row[PROBE], row[SAMPLE])
        if key not in done:
            done.add(key)
            yield (row[PROBE], row[SAMPLE], random.choice(thumbnail))
            

# code
if __name__ == '__main__':

    parser = argparse.ArgumentParser(sys.argv[0])
    parser.add_argument("--filename", type=file)
    subparsers = parser.add_subparsers(dest='subparser_name')

    parser_rmindiv = subparsers.add_parser("rmindiv", help="Remove an individual")
    parser_rmsample = subparsers.add_parser("rmsample", help="Remove a sample")
    parser_rmcomp = subparsers.add_parser("rmcomparison", help="Remove a comparison")
    parser_summary = subparsers.add_parser("summary", help="Display various summaries")
    parser_filter = subparsers.add_parser("filter", help="filter database of scores")
    parser_fakethumbnails = subparsers.add_parser("fakethumbnails", help="generate a fake database of thumbnails")

    parser_rmindiv.add_argument("id", type=int)

    parser_rmsample.add_argument("probeid", type=int)
    parser_rmsample.add_argument("sampleid", type=int)

    parser_rmcomp.add_argument("galleryid", type=int)
    parser_rmcomp.add_argument("probeid", type=int)
    parser_rmcomp.add_argument("sampleid", type=int)

    parser_filter.add_argument("galleryid", type=str, help="Id of gallery individual (or ? for any)")
    parser_filter.add_argument("probeid", type=str, help="Id of probe individual (or ? for any)")
    parser_filter.add_argument("sampleid", type=str, help="Id of sample for probe (or ? for any)")

    parser_fakethumbnails.add_argument("pics", type=str, nargs='+')


    args = parser.parse_args(sys.argv[1:])
    if args.filename:
        stream = args.filename #open(args.filename)
    else:
        stream = sys.stdin

    data = read_stream(stream)

    if args.subparser_name == 'rmindiv':
        data = remove_individual(data, args.id)
        output_data(data)
    elif args.subparser_name == 'rmsample':
        data = remove_sample(data, args.probeid, args.sampleid)
        output_data(data)
    elif args.subparser_name == 'rmcomparison':
       data = remove_comparison(data, args.galleryid, args.probeid, args.sampleid)
       output_data(data)
    elif args.subparser_name == 'summary':
        summary(data)
    elif args.subparser_name == 'filter':
        data = filterData(data, args.galleryid, args.probeid, args.sampleid)
        output_data(data)
    elif args.subparser_name == "fakethumbnails":
        data = generateThumbnails(data, args.pics)
        output_data(data)

# metadata
__author__ = 'Romain Giot'
__copyright__ = 'Copyright 2017, LaBRI'
__credits__ = ['Romain Giot']
__licence__ = 'GPL'
__version__ = '0.1'
__maintainer__ = 'Romain Giot'
__email__ = 'romain.giot@u-bordeaux1.fr'
__status__ = 'Prototype'

