/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include <tulip/PluginLoaderTxt.h>
#include <tulip/PluginLibraryLoader.h>
#include <tulip/GlMainWidget.h>
#include <tulip/MouseInteractors.h>
#include <tulip/TlpQtTools.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/TulipViewSettings.h>
#include <tulip/GlGraphComposite.h>
#include <tulip/GlGraphRenderingParameters.h>
#include <tulip/ForEach.h>

#include <tulip/TlpTools.h>

#include "bpg/common.h"
#include "docopt.h"

#include <QApplication>

using namespace tlp;
using namespace std;


static const char USAGE[] =
  R"(Biometric Power Graph.

    Usage:
      biometricPowerGraph --load <inputfname> [--isDistance] [--save <outputfname> [--debug]] [--show] [--export <imagefname>]
      biometricPowerGraph --version

    Options:
      -h --help     Show this screen.
      --version     Show version.
      --load <inputfname>   Name of the CSV file to open
      --save <outputfname>  Name of the tulip graph to save
      --debug       If activated, all the graph is saved, not only the display graph
      --show        Open the visualisation interface
      --isDistance  If the dataset of score is distance based, we have to transform it as scores
)";

Graph *treat(const string &inputfname, bool isScore) {
  string error;

  // Load the file
  cout << "Load CSV file" << endl;
  DataSet dataset;
  dataset.set("file::filename", inputfname);
  dataset.set("isScore", isScore);

  Graph *g = importGraph("Biometric Power Graph Import", dataset, nullptr, nullptr);

  // Compute Biometric information
  cout << "Compute EER" << endl;
  dataset = DataSet();
  g->applyAlgorithm("Compute Biometric Properties", error, &dataset);

  // Compute visuals attributes to draw the graph
  double threshold = 0;
  g->getAttribute(ATTR_THR, threshold);
  cout << "Compute visual attributes" << endl;
  dataset = DataSet();
  dataset.set("threshold", threshold);
  g->applyAlgorithm("Draw Biometric Power Graph", error, &dataset);

  return g;
}


// That function sets some rendering parameters on the graph to visualize
void setGraphRenderingParameters(GlGraphComposite *glGraphComposite) {
  GlGraphRenderingParameters *renderingParameters = glGraphComposite->getRenderingParametersPointer();
  // Activate the display of edge extremities (arrows by default)
  renderingParameters->setViewArrow(false);
  // No color interpolation for the edges
  renderingParameters->setEdgeColorInterpolate(false);
  // Size interpolation for the edges
  renderingParameters->setEdgeSizeInterpolate(false);
  // Scale labels to node sizes
  renderingParameters->setLabelScaled(true);
  renderingParameters->setViewNodeLabel(false);
}


auto showGraph(QApplication &app, tlp::Graph *g, const bool showGraph, const std::string &fname = "") {


  // Creates the main widget that will display our graph
  GlMainWidget *mainWidget = new GlMainWidget(NULL);
  // Adds a layer to the scene
  GlLayer *mainLayer = mainWidget->getScene()->createLayer("Main");
  // Adds the graph to this layer
  mainLayer->addGraph(g, "Biometric Graph");
  // Sets some rendering parameters on the graph to visualize
  setGraphRenderingParameters(mainWidget->getScene()->getGlGraphComposite());
  // Display the widget
  mainWidget->show();
  // Flush event loop in order to let paint events pass through in order for the scene to be initialized.
  QApplication::processEvents();
  // Center the camera and draw the graph
  mainWidget->centerScene();


  if (fname.size() > 0) {
    constexpr int image_width = 4000;
    constexpr int image_height = 4000;
    mainWidget->createPicture(fname, image_width, image_height);
  }

  if (showGraph)
    mainWidget->draw();

  // Adds Zoom and pan navigation to the widget
  mainWidget->installEventFilter(new MouseNKeysNavigator);
  if (showGraph)
    return app.exec();
  else
    return 0;
}


int main(int argc, char **argv) {

  std::map< std::string, docopt::value > args = docopt::docopt(USAGE,
                                                               {argv + 1, argv + argc},
                                                               true, // show help if requested
                                                               "Biometric Power Graph v0.1"); // version string


  QApplication app(argc, argv);
  // tlp::initTulipLib(QStringToTlpString(QApplication::applicationDirPath()).c_str());
  tlp::initTulipSoftware();


  const std::string inputfname = args["--load"].asString();
  std::cout << "Will treat " << inputfname << endl;

  const bool is_distance = static_cast< bool >(args["--isDistance"]) && args["--isDistance"].asBool();

#if 1
  Graph *g = treat(inputfname, !is_distance);

  if (static_cast< bool >(args["--save"])) {
    cout << "Save result" << endl;
    if (static_cast< bool >(args["--debug"]) && args["--debug"].asBool()) {
      cout << "(in debug mode)" << endl;
      saveGraph(g, args["--save"].asString());
    } else {
      cout << "(only display)" << endl;
      saveGraph(g->getSubGraph("display"), args["--save"].asString());
    }
  }
#endif

  const bool show = static_cast< bool >(args["--show"]) && args["--show"].asBool();
  const bool exportf = static_cast< bool >(args["<imagefname>"]);

  std::cout << "Show " << show << std::endl;
  std::cout << "Export " << exportf << std::endl;

#if 1
  if (show || exportf) {
    // And show it
    if (exportf) {
      return showGraph(app, g->getSubGraph("display"), show, args["<imagefname>"].asString());
    } else {
      return showGraph(app, g->getSubGraph("display"), show, "");
    }
  } else {
    return EXIT_SUCCESS;
  }

#else
  return 0;
#endif
}
