/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include <tulip/Graph.h>
#include <tulip/LayoutProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/BoundingBox.h>
#include <tulip/DrawingTools.h>

#include <cassert>
#include <utility>
#include <vector>
#include <map>
#include "lineIntersection.h"

#include "common.h"
#include "bpg/BPGConfig.h"

class BPG_UTILS_SCOPE MultilevelEdgeBundling
{

private:
  std::map< std::string, tlp::node > location2node;


  tlp::node getArtificalNodeAt(tlp::Graph *const graph, tlp::LayoutProperty *const viewLayout, const tlp::Coord &coord) {
    const std::string key = tlp::PointType::toString(coord);


    auto const iter = location2node.find(key);
    if (location2node.end() == iter) {
      tlp::node added = graph->addNode();
      viewLayout->setNodeValue(added, coord);
      location2node.insert({key, added});
      return added;
    } else {
      return iter->second;
    }
  }


private:
  void bundleGraph(tlp::Graph *const graph, tlp::LayoutProperty *const viewLayout) {
    std::string error;


    // Compute the bundling on the individuals layout
    tlp::DataSet dataSet;
    dataSet.set("layout", viewLayout);
    dataSet.set("long_edges", 0.5);
    dataSet.set("iterations", 5);
    dataSet.set("split_ratio", 20);
    bool res = graph->applyAlgorithm(
      "Edge bundling",
      error,
      &dataSet,
      nullptr);

    if (!res)
      std::cerr << "[ERROR] Unable to do bundling " << error << std::endl;
  }


  /**
 * The bends can go INSIDE the quotient nodes ... we try to remove them
 */
  void cleanup_bends(tlp::Graph *const quotient, const tlp::edge quotientEdge, const tlp::node quotientSource, const tlp::node quotientTarget) {
    assert(quotient->isElement(quotientEdge));
    assert(quotient->isElement(quotientSource));
    assert(quotient->isElement(quotientTarget));
    assert(quotient->isMetaNode(quotientSource));
    assert(quotient->isMetaNode(quotientTarget));

    auto const viewLayout = quotient->getProperty< tlp::LayoutProperty >("viewLayout");
    auto const viewSize = quotient->getProperty< tlp::SizeProperty >("viewSize");

    // Prepare the bounding box of the quotient node
    auto compute_bb = [&](const tlp::node n) {
      auto const center = viewLayout->getNodeValue(n);
      auto const size = viewSize->getNodeValue(n);

      tlp::BoundingBox bb;

      bb.expand(tlp::Coord(center.x() + size.x() / 2, center.y() + size.y() / 2, 0));
      bb.expand(tlp::Coord(center.x() + size.x() / 2, center.y() - size.y() / 2, 0));
      bb.expand(tlp::Coord(center.x() - size.x() / 2, center.y() + size.y() / 2, 0));
      bb.expand(tlp::Coord(center.x() - size.x() / 2, center.y() - size.y() / 2, 0));

      return bb;
    };

    auto const bbSource = compute_bb(quotientSource);
    auto const bbTarget = compute_bb(quotientTarget);

    // We assume a node is a square
    tlp::Vec3f bbS[8];
    bbSource.getCompleteBB(bbS);
    const std::vector< std::pair< tlp::Coord, tlp::Coord > > contourS{
      std::make_pair(bbS[0], bbS[1]),
      std::make_pair(bbS[1], bbS[2]),
      std::make_pair(bbS[2], bbS[3]),
      std::make_pair(bbS[3], bbS[0])};
    tlp::Vec3f bbT[8];
    bbTarget.getCompleteBB(bbT);
    const std::vector< std::pair< tlp::Coord, tlp::Coord > > contourT{
      std::make_pair(bbT[0], bbT[1]),
      std::make_pair(bbT[1], bbT[2]),
      std::make_pair(bbT[2], bbT[3]),
      std::make_pair(bbT[3], bbT[0])};

    // Filter the bends
    std::vector< tlp::Coord > original_bends{viewLayout->getEdgeValue(quotientEdge)};
    // XXX Check if these two additions are mandatory
    original_bends.insert(original_bends.begin(), viewLayout->getNodeValue(quotientSource));
    original_bends.push_back(viewLayout->getNodeValue(quotientTarget));

    std::vector< tlp::Coord > new_bends;
    new_bends.reserve(original_bends.size());

    // XXX See if we can do that with filters
    // XXX Instead removing all points, add the intersection between the node and the last subpart of edge
    bool previous_in_source = true;
    bool previous_in_target = false;
    for (size_t i = 0; i < original_bends.size(); ++i) {
      const auto &bend = original_bends.at(i);
      const bool current_in_source = bbSource.contains(bend);
      const bool current_in_target = bbTarget.contains(bend);

      // Add the intersection if any
      if (previous_in_source && !current_in_source) {
        tlp::Coord intersectionPoint;
        const std::pair< tlp::Vec3f, tlp::Vec3f > edge_part = (0 == i)
                                                                ? std::make_pair(viewLayout->getNodeValue(quotientSource), bend)
                                                                : std::make_pair(original_bends.at(i - 1), bend);

        bool intersects = false;
        for (const auto &line : contourS) {
          if (segmentsIntersect(edge_part, line)) {
            tlp::Coord intersectionPoint;
            computeLinesIntersection(edge_part, line, intersectionPoint);
            new_bends.push_back(intersectionPoint); // XXX Is it really the closest one
            intersects = true;
            break;
          }
        }
        if (intersects) {
          // nothing to do
        } else {
        }
      }

      // Add the bend if needed
      if (!current_in_source && !current_in_target) {
        new_bends.push_back(bend);

      } else {
      }

      // Add the intersection if any
      if ((!previous_in_target && current_in_target) || (i == original_bends.size() - 1)) {
        const std::pair< tlp::Vec3f, tlp::Vec3f > edge_part = (!previous_in_target && current_in_target)
                                                                ? std::make_pair(original_bends.at(i - 1), bend)
                                                                : std::make_pair(original_bends.at(i - 1), viewLayout->getNodeValue(quotientTarget));


        bool intersects = false;
        for (const auto &line : contourT) {
          if (segmentsIntersect(edge_part, line)) {
            tlp::Coord intersectionPoint;
            computeLinesIntersection(edge_part, line, intersectionPoint);
            new_bends.push_back(intersectionPoint);
            intersects = true;
            break;
          }
        }
        if (intersects) {
          break;
        } else {
        }
      }

      // Store previous state
      previous_in_source = current_in_source;
      previous_in_target = current_in_target;
    }
    viewLayout->setEdgeValue(quotientEdge, new_bends);
  }

  /**
   * Pre-requsisites : the coordinates of the nodes of subgraphs are well positionned in comparison to the quotient nodes
   */
  auto manageArtificialNodes(tlp::Graph *const quotient) {
    auto const viewLayout = quotient->getProperty< tlp::LayoutProperty >("viewLayout");

    //TODO manage artificial nodes of different edges positionned at the same place (in this case, it is better to use the SAME node)
    std::map< tlp::node, tlp::node > original2artificial_nodes; // Map from original node to artificial node

    // loop over all the quotient edges
    std::vector< tlp::edge > quotientEdges{quotient->edges()};
    for (tlp::edge quotientEdge : quotientEdges) {
      // Get the appropriate subgraphs
      const tlp::node quotientSource = quotient->source(quotientEdge);
      const tlp::node quotientTarget = quotient->target(quotientEdge);
      assert(quotientSource.isValid());
      assert(quotientTarget.isValid());

      tlp::Graph *const sgSource = quotient->getNodeMetaInfo(quotientSource);
      tlp::Graph *const sgTarget = quotient->getNodeMetaInfo(quotientTarget);
      assert(nullptr != sgSource);
      assert(nullptr != sgTarget);
      assert(sgSource != sgTarget);

      auto const viewLayoutSource = sgSource->getLocalProperty< tlp::LayoutProperty >("viewLayout");
      auto const viewLayoutTarget = sgTarget->getLocalProperty< tlp::LayoutProperty >("viewLayout");

      // Remove the extra bends of each metaedge
      cleanup_bends(quotient, quotientEdge, quotientSource, quotientTarget);

      // Get the bends after modification of this meta edge
      const std::vector< tlp::Coord > bends = viewLayout->getEdgeValue(quotientEdge);


      assert(bends.size() > 0);

      // By construction the front and back nodes are within the border of the source and target node
      const tlp::Coord sourceBend = bends.front();
      const tlp::Coord targetBend = bends.back();

      // Add the artificial nodes in the source/Target subgraphs in order to compute the bundling inside them
      const tlp::node artificialSource = getArtificalNodeAt(sgSource, viewLayoutSource, sourceBend);
      const tlp::node artificialTarget = getArtificalNodeAt(sgTarget, viewLayoutTarget, targetBend);


      // Add the artificial edges from the real nodes of the subgraphs to these intersection nodes
      tlp::edge originalEdge;
      // XXX Not portable, do that otherwhise
      const tlp::Graph *const originalGraph = quotient->getRoot()->getSubGraph(LAYER_USERS)->getSubGraph("modified");
      stableForEach(originalEdge, quotient->getEdgeMetaInfo(quotientEdge)) {
        // XXX Do not knwo why there is this issue
        if (!originalGraph->isElement(originalEdge))
          continue;

        assert(originalGraph->isElement(originalEdge));
        const tlp::node origSource = originalGraph->source(originalEdge);
        const tlp::node origTarget = originalGraph->target(originalEdge);

        assert(sgSource->isElement(origSource));
        assert(sgTarget->isElement(origTarget));

        // Add these new extremity edges
        sgSource->addEdge(origSource, artificialSource);
        sgTarget->addEdge(artificialTarget, origTarget);

        // Do not forget to place them in the map
        original2artificial_nodes.emplace(origSource, artificialSource);
        original2artificial_nodes.emplace(origTarget, artificialTarget);
      }
    }

    return original2artificial_nodes;
  }


  void injectControlPoint(tlp::Graph *const quotient, tlp::Graph *const original, const std::map< tlp::node, tlp::node > &original2artificial_nodes, tlp::Graph *const display = nullptr) {
    tlp::edge quotientEdge, originalEdge;
    tlp::node quotientNode;

    auto const quotientLayout = quotient->getProperty< tlp::LayoutProperty >("viewLayout");
    auto const originalLayout = original->getLocalProperty< tlp::LayoutProperty >("viewLayout");


    const bool approximated = nullptr != display;
    auto displayLayout = approximated ? display->getLocalProperty< tlp::LayoutProperty >("viewLayout") : nullptr;

    if (approximated) {

#if USE_TLP_ITERATOR
      {
        std::vector< tlp::edge > edges(display->numberOfEdges());
        size_t i = 0;
        tlp::edge e;
        forEach (e, display->getEdges()) {
          edges[i++] = e;
        }
      }
#else
      display->delEdges(display->edges());
#endif

    }


    // Treat all the edges integrated inside a meta edge
    const std::vector< tlp::edge > quotientEdges{quotient->edges()};
    for (tlp::edge quotientEdge : quotientEdges) {
      const std::vector< tlp::Coord > middle{quotientLayout->getEdgeValue(quotientEdge)};

      const tlp::node quotientSource = quotient->source(quotientEdge);
      const tlp::node quotientTarget = quotient->target(quotientEdge);
      tlp::Graph *const sourceGraph = quotient->getNodeMetaInfo(quotientSource);
      tlp::Graph *const targetGraph = quotient->getNodeMetaInfo(quotientTarget);
      auto const viewLayoutSource = sourceGraph->getLocalProperty< tlp::LayoutProperty >("viewLayout");
      auto const viewLayoutTarget = targetGraph->getLocalProperty< tlp::LayoutProperty >("viewLayout");

      assert(nullptr != sourceGraph);
      assert(nullptr != targetGraph);

      tlp::node display_startbend, display_stopbend;
      if (approximated) {
        // Draw the middle part of an edge
        // XXX Manage duplicated nodes
        display_startbend = display->addNode();
        display_stopbend = display->addNode();

        displayLayout->setNodeValue(display_startbend, middle.front());
        displayLayout->setNodeValue(display_stopbend, middle.back());


        display->getProperty< tlp::StringProperty >("viewLabel")->setNodeValue(display_startbend, "src");
        display->getProperty< tlp::StringProperty >("viewLabel")->setNodeValue(display_stopbend, "tgt");
        displayLayout->setEdgeValue(
          display->addEdge(display_startbend, display_stopbend),
          middle);
      }

      stableForEach(originalEdge, quotient->getEdgeMetaInfo(quotientEdge)) {
        const tlp::node originalSource = original->source(originalEdge);
        const tlp::node originalTarget = original->target(originalEdge);

        std::vector< tlp::Coord > bends; // This vector is only filled on the not approximate graph

        size_t number_of_bends = 0;

        {
          // Add source control points
          const auto iterSrc = original2artificial_nodes.find(originalSource);
          if (iterSrc != original2artificial_nodes.end()) {
            const tlp::edge artificial_edge = sourceGraph->existEdge(originalSource, iterSrc->second, false);
            assert(artificial_edge.isValid());
            const std::vector< tlp::Coord > beginning = viewLayoutSource->getEdgeValue(artificial_edge);

            const tlp::node artificial_edge_source = sourceGraph->source(artificial_edge);
            const tlp::node artificial_edge_target = sourceGraph->target(artificial_edge);

            if (approximated) {

              // In the approximate case, we want to add the artificial nodes and edges
              // XXX The same work can be done several times here
              display->addNode(artificial_edge_source);
              display->addNode(artificial_edge_target);
              display->addEdge(artificial_edge);

              if (originalSource != artificial_edge_source) {
                if (beginning.size()) {
                  displayLayout->setNodeValue(artificial_edge_source, beginning.front());
                  displayLayout->setNodeValue(artificial_edge_target, beginning.back());
                  displayLayout->setEdgeValue(artificial_edge, beginning);
                } else {
                  displayLayout->setNodeValue(artificial_edge_source, displayLayout->getNodeValue(display_startbend));
                  displayLayout->setNodeValue(artificial_edge_target, displayLayout->getNodeValue(display_stopbend));
                }
              } else {
                if (beginning.size()) {
                  displayLayout->setNodeValue(artificial_edge_source, beginning.back());
                  displayLayout->setNodeValue(artificial_edge_target, beginning.front());
                  std::vector< tlp::Coord > reversed{beginning.rbegin(), beginning.rend()};
                  displayLayout->setEdgeValue(artificial_edge, reversed);
                } else {
                  displayLayout->setNodeValue(artificial_edge_target, displayLayout->getNodeValue(display_startbend));
                  displayLayout->setNodeValue(artificial_edge_source, displayLayout->getNodeValue(display_stopbend));
                }
              }

            } else {

              // Add the bends ...
              if (originalSource == artificial_edge_source) {
                bends.insert(
                  bends.end(),
                  beginning.begin(),
                  beginning.end());
              } else {
                assert(originalSource == artificial_edge_target);
                bends.insert(
                  bends.end(),
                  beginning.rbegin(),
                  beginning.rend());
              }

              // ... and the coords of the artifical_node
              bends.emplace_back(
                viewLayoutSource->getNodeValue(
                  iterSrc->second));

              number_of_bends += beginning.size() + 1;
#ifndef NDEBUG
              std::cerr << "Beginning: " << beginning.size();
#endif
            }
          } else {
#ifndef NDEBUG
            std::cerr << "Impossible to find start";
            assert(false); //XXX I do not understand why this is raised now :(
#endif
          }
        }

        {


          if (!approximated) {
            // Add middle control points
            bends.insert(
              bends.end(),
              middle.begin(),
              middle.end());
            number_of_bends += middle.size();
#ifndef NDEBUG
            std::cerr << "  Middle: " << middle.size();
#endif
          }
        }

        {
          // Add target control point
          const auto iterTarget = original2artificial_nodes.find(originalTarget);
          if (iterTarget != original2artificial_nodes.end()) {
            const tlp::edge artificial_edge = targetGraph->existEdge(originalTarget, iterTarget->second, false);
            assert(artificial_edge.isValid());
            const std::vector< tlp::Coord > ending = viewLayoutTarget->getEdgeValue(artificial_edge);

            const tlp::node artificial_edge_source = targetGraph->source(artificial_edge);
            const tlp::node artificial_edge_target = targetGraph->target(artificial_edge);

            if (approximated) {

              display->addNode(artificial_edge_source);
              display->addNode(artificial_edge_target);
              display->addEdge(artificial_edge);


              if (originalTarget == artificial_edge_target) {
                if (ending.size()) {
                  displayLayout->setEdgeValue(artificial_edge, ending);
                  display->addEdge(artificial_edge_target, display_stopbend);
                }
              } else {
                if (ending.size()) {
                  std::vector< tlp::Coord > reversed{ending.rbegin(), ending.rend()};
                  displayLayout->setEdgeValue(artificial_edge, reversed);
                  display->addEdge(artificial_edge_source, display_stopbend);
                }
              }


            } else {

              // Add the artifical point
              bends.emplace_back(
                viewLayoutTarget->getNodeValue(
                  iterTarget->second));


              if (originalTarget == targetGraph->source(artificial_edge)) {
                bends.insert(
                  bends.end(),
                  ending.rbegin(),
                  ending.rend());
              } else {
                bends.insert(
                  bends.end(),
                  ending.begin(),
                  ending.end());
              }
              number_of_bends += ending.size() + 1;
#ifndef NDEBUG
              std::cerr << "  Ending: " << ending.size();
              std::cerr << " Total: " << bends.size() << "/ Expected " << number_of_bends << std::endl;
#endif
            }
          } else {
#ifndef NDEBUG
            std::cerr << "Impossible to find end";
            assert(false); //XXX Why ? :(
#endif
          }
        }


        // Set the control points
        if (!approximated) {
          assert(number_of_bends == bends.size());
          originalLayout->setEdgeValue(originalEdge, bends);
        }
      }
    }


    // Treat all the edges totally inside a meta node
    for (tlp::node quotientNode : quotient->nodes()) {
      tlp::Graph *const clusterGraph = quotient->getNodeMetaInfo(quotientNode);
      const auto localViewLayout = clusterGraph->getProperty< tlp::LayoutProperty >("viewLayout");
      assert(nullptr != clusterGraph);
      assert(nullptr != localViewLayout);

      for (tlp::edge innerEdge : clusterGraph->edges()) {
        assert(original->isElement(innerEdge));
        const tlp::node from = clusterGraph->source(innerEdge);
        const tlp::node to = clusterGraph->target(innerEdge);

        // There are particular edges which come from the intersection nodes.
        // We are not interested by them and we do not want to treat them
        if (original2artificial_nodes.find(from) == original2artificial_nodes.end() ||
            original2artificial_nodes.find(to) == original2artificial_nodes.end()) {
          continue;
        }

        // XXX here we are sure to treat an edge from two real nodes (i.e. users)
        const std::vector< tlp::Coord > bends = localViewLayout->getEdgeValue(innerEdge);
#ifndef NDEBUG
        std::cerr << "Local bundling of " << bends.size() << " elements" << std::endl;
#endif
        originalLayout->setEdgeValue(innerEdge, bends);
      }
    }
  }


public:
  /**
 * - Step 1: Compute the bundling in the quotient graph => it will try to move around cluster (although with no guaranties)
 * - Step 2: In Each cluster, add artificial nodes which correspond to the entrance in the cluster
 * - Step 3: Compute the bundling inside the clusters
 * - Step 4: Inject the control points
 */
  void multilevelEdgeBundling(tlp::Graph *const quotient, tlp::Graph *const original, tlp::Graph *const drawing = nullptr) {

 
	// was deactivated for windows before
	// TODO remove this variable
    constexpr bool is_activated = true;

	if (is_activated) {
      auto const viewLayout = quotient->getProperty< tlp::LayoutProperty >("viewLayout");

      // Compute bundling on the quotient graph
      bundleGraph(quotient, viewLayout); // Step 1

      // Get the intersection between meta-edges and quotient nodes
      auto original2artificial_nodes = manageArtificialNodes(quotient); // Step 2

      // Compute the bundling in each quotient graph (we know all the nodes have been added)
      tlp::node quotientNode;
      const auto nbQuotientNodes = quotient->numberOfNodes();
      std::remove_const< decltype(nbQuotientNodes) >::type step = 0;

      stableForEach(quotientNode, quotient->getNodes()) {
        tlp::Graph *const sg = quotient->getNodeMetaInfo(quotientNode);
        bundleGraph(sg, sg->getProperty< tlp::LayoutProperty >("viewLayout"));
      }

      injectControlPoint(quotient, original, original2artificial_nodes, drawing); // Step 3


      if (nullptr == drawing) {
        // Remove the artifical points
        tlp::Graph *const root = original->getRoot();
        for (auto &couple : original2artificial_nodes) {
          if (root->isElement(couple.second))
            root->delNode(couple.second, true);
        }
      }
    }
  }
};
