/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#ifndef POWERGRAPHMANAGER_H
#define POWERGRAPHMANAGER_H


#include "tulip/Graph.h"

#include "utils.h"
#include "bpg/BPGConfig.h"


struct BPG_UTILS_SCOPE SampleInformation {

  int sampleId;
  int ownerId;

  bool isVerified;
  double genuineScore;

  int nbImpersonnateComparison;
  int nbImpersonnation;

  int nbGenuineComparison;
};


struct BPG_UTILS_SCOPE IndividualInformation {

  int id;
  double FNMR;
  double FMR;
  double FMR2;
};


struct BPG_UTILS_SCOPE RocNodeInformation {
  double threshold;
  double FNMR;
  double FMR;
};


class BPG_UTILS_SCOPE PowerGraphManager
{
private:
  int _gallery_to_filter;

  /**
     * @brief Root graph which contains the tulip structures of the power graph
     */
  tlp::Graph *_root = nullptr;

  /**
     * @brief thr Selected treshold chosen to be displayed
     */
  double thr;

  /**
     * @brief approximatedVisualisation is set to true if we display the approximated graph instead of the full graph.
     * The approximated graph displays far more edges
     */
  // bool approximatedVisualisation;


public:
  PowerGraphManager();
  PowerGraphManager(const PowerGraphManager &) = delete;
  PowerGraphManager(const PowerGraphManager &&) = delete;


  void backupState();
  void restoreState();

  // Set of utility functions
  bool isSample(tlp::node n) { return ::isSample(_root, n); }
  bool isIndividual(tlp::node n) { return ::isIndividual(_root, n); }
  int getSampleId(tlp::node n) { return _root->getProperty< tlp::IntegerProperty >(PROP_SAMPLE)->getNodeValue(n); }
  int getIndividualId(tlp::node n) { return _root->getProperty< tlp::IntegerProperty >(PROP_USER)->getNodeValue(n); }
  tlp::node getOwnerOfSample(tlp::node n) { return ::getOwner(_root, n); }
  inline double getDecisionThreshold() { return thr; }

  /**
     * This method must be called when we want to release the current graph from memory BECAUSE it is not displayed anymore
     */
  void safelyRealeaseData() {
    if (!isLoaded())
      return; // Nothing to do if nothing is loaded ...

    delete _root; // XXX hope it is the good way ...

    _root = nullptr;
    _gallery_to_filter = -1;
  }


  bool isLoaded() {
    return nullptr != _root;
  }


  bool displayAvailable() {
    return _root != nullptr && _root->getSubGraph(LAYER_DISPLAY) != nullptr;
  }

  /**
     * Load the CSV file in order to generate the Graph
     * @brief importCSV
     * @param fname
     * @return
     */
  bool importCSV(const std::string fname, const bool isScore, const std::string thumbnailFname = "");


  /**
     * Return the triplet, threshold, FNMR, FMR extracted from the ROC node
     */
  RocNodeInformation getInformationFromRocNode(tlp::node n);


  SampleInformation getInformationFromSampleNode(tlp::node n);
  IndividualInformation getInformationFromIndividualNode(tlp::node n);



  /**
   * @param thr Decision threshold
   * @param approximated Use the approximated version (does not currently work)
   * @param layoutGraph If specified, reproduce the layout of this graph to ease comparisons
   */
  bool buildDisplayGraphForThreshold(
    const double thr,
    const bool approximated,
    tlp::Graph* layoutGraph,

    const double partitionFNMRLowThr,
    const double partitionFNMRHighThr,

    const double partitionFMR_inLowThr,
    const double partitionFMR_inHighThr,

    const double partitionFMR_outLowThr,
    const double partitionFMR_outHighThr);

  /**
     * Returns the graph which represents the roc curve
     * @brief getROCGraph
     * @return
     */
  tlp::Graph *getROCGraph();


  /**
     * Returns the graph to display
     * @brief getDisplayGraph
     * @return
     */
  tlp::Graph *getDisplayGraph();

    std::pair< std::vector<double>, std::vector<double> > getIntraInterScores();

  /**
     * We do not want to display the display graph, but a filtered version of it.
     * This method :
     *  - destroy an existing filtered graph
     *  - create a new one according the specified criterias
     *  - returns the newly created graph
     */
  tlp::Graph *buildFilteredGraph();


  void setGalleryToFilter(int gallery, bool rebuild = true) {
    clearFilters(false); // Ensure all other filters are removed
    _gallery_to_filter = gallery; // Configure the filter

    // Rebuild the representation graph if needed
    if (rebuild) {
      buildFilteredGraph();
    }
  }


  void clearFilters(bool rebuild = true) {
    _gallery_to_filter = -1;
    if (rebuild) {
      buildFilteredGraph();
    }
  }

  tlp::Graph *getRoot() {
    return _root;
  }


std::pair<const std::vector<double>, const std::vector<double>> getZooPlotData();


};

#endif // POWERGRAPHMANAGER_H
