/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#ifndef __BIO_POWERGRAPH_UTILS_H__
#define __BIO_POWERGRAPH_UTILS_H__

#include <tulip/Graph.h>
#include <tulip/DoubleProperty.h>
#include <tulip/Color.h>
#include <vector>
#include <algorithm>

#include "common.h"


template < typename Container, typename Property >
inline void setSameValueFor(const Container &nodes, Property *property, const decltype(property->getNodeValue(tlp::node())) &value) {

  for (const tlp::node &n : nodes) {
    assert(n.isValid());
    property->setNodeValue(n, value);
  }
}

/**
 * Compute the nodes lower than the percentile (and higher than 1-percentile) 
 */
inline std::pair< std::vector< tlp::node >, std::vector< tlp::node > > getPercentilesNodes(tlp::Graph *graph, tlp::DoubleProperty *p, double percentage) {
  assert(percentage < 0.5 && percentage > 0);

  tlp::node n;
  std::vector< double > values;
  std::vector< tlp::node > above, below;

// Get all values
#if USE_TLP_ITERATOR
  forEach (n, graph->getNodes()) {
#else
  for (tlp::node n : graph->nodes()) {
#endif
    values.push_back(p->getNodeValue(n));
  }
  assert(values.size() == graph->numberOfNodes());

  // Sort them
  std::sort(values.begin(), values.end());
  assert(values.front() <= values.back());

  // Get the limits
  const size_t rank_below = static_cast< double >(values.size()) * percentage + 0.5;
  const size_t rank_above = static_cast< double >(values.size() * (1.0 - percentage)) + 0.5;

  const double value_below = values.at(rank_below);
  const double value_above = values.at(rank_above);

#ifndef NDEBUG
  std::cout << "0: " << values.front() << "--" << rank_below << ": " << value_below << " -- " << rank_above << ": " << value_above << "--" << values.size() << ": " << values.back() << std::endl;
#endif

// Associate the nodes
#if USE_TLP_ITERATOR
  forEach (n, graph->getNodes()) {
#else
  for (tlp::node n : graph->nodes()) {
#endif
    const double value = p->getNodeValue(n);

    if (value < value_below) {
      below.push_back(n);
    } else if (value > value_above) {
      above.push_back(n);
    }
  }


  assert(below.size() != 0);
  assert(above.size() != 0);

  return std::make_pair(below, above);
}


inline tlp::Color HSVColor(int h, int s, int v, int alpha) {
  unsigned char r;
  unsigned char g;
  unsigned char b;

  int i;
  int p, q, t;
  float f, sf = s / 255.0;

  if (v < 0)
    v = 0;
  else if (v > 255)
    v = 255;

  if (s < 0)
    s = 0;
  else if (s > 255)
    s = 255;

  //   h %= 360;
  //   if (h<0) h+=360;

  if (s == 0) {
    // achromatic (grey)
    r = g = b = v;
    return tlp::Color(r, g, b);
  }

  i = h / 60; // sector 0 to 5
  f = (h / 60.0f - i); // factorial part of h
  p = static_cast<int> (v * (1 - sf));
  q = static_cast<int> (v * (1 - sf * f));
  t = static_cast<int> (v * (1 - sf * (1 - f)));

  switch (i) {
    case 0:
      r = v;
      g = t;
      b = p;
      break;

    case 1:
      r = q;
      g = v;
      b = p;
      break;

    case 2:
      r = p;
      g = v;
      b = t;
      break;

    case 3:
      r = p;
      g = q;
      b = v;
      break;

    case 4:
      r = t;
      g = p;
      b = v;
      break;

    default: // case 5:
      r = v;
      g = p;
      b = q;
      break;
  }


  return tlp::Color(r, g, b, alpha);
}


inline std::vector< tlp::Coord > reduce_bends(const std::vector< tlp::Coord > orig, const double epsilon = 10e-3) {
  std::vector< tlp::Coord > novel;
  novel.reserve(orig.size());

  bool first = true;
  tlp::Coord last;
  for (const tlp::Coord &c : orig) {
    if (first) {
      last = c;
      novel.push_back(last);
      first = false;
    } else {
      if ((last - c).norm() <= epsilon) {
        // do nothing
      } else {
        novel.push_back(last);
        last = c;
      }
    }
  }

  return novel;
}


inline tlp::Color averagedColor(const tlp::Color &a, const tlp::Color &b){
          return tlp::Color{

              static_cast<unsigned char>((a.r() + b.r()) >> 1),
              static_cast<unsigned char>((a.g() + b.g()) >> 1),
              static_cast<unsigned char>((a.b() + b.b()) >> 1)
        };
      }
#endif
