/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include "bpg/BiometricDatabaseModel.h"
#include "bpg/common.h"

#include <QBrush>
#include <QColor>


  const QColor BiometricDatabaseModel::TrueMatch{0, 255, 0};
  const QColor BiometricDatabaseModel::FalseNonMatch{128, 128, 128};
  const QColor BiometricDatabaseModel::TrueNonMatch{0, 0, 255};
  const QColor BiometricDatabaseModel::FalseMatch{255, 0, 0};

BiometricDatabaseModel::BiometricDatabaseModel(QObject *parent)
: QAbstractTableModel{parent},
  _data{},
  _isScore{true} {
}

BiometricDatabaseModel::~BiometricDatabaseModel() {
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
// Redefinition of the mandatory methods

int BiometricDatabaseModel::columnCount(const QModelIndex &parent) const {
  return NB_COLUMNS;
}
int BiometricDatabaseModel::rowCount(const QModelIndex &parent) const {
  return _data.size();
}
#pragma GCC diagnostic pop

QVariant BiometricDatabaseModel::data(const QModelIndex &index, int role) const {
	const auto line = _data.at(index.row());
	const auto tag = line[4];
  if (role == Qt::DisplayRole) {
      if (index.column() == 4) {
          const auto symbol = [&]() -> QChar{
              if (tag == 0) {
                  return '-';
              }
              else if (tag == 1) {
                  return 0x2713;
              }
              else {
                  assert(tag == -1);
                  return 0x2717;
              }

          }();
          return symbol;

      }
      else {
        return _data.at(index.row())[index.column()];
      }
  }
  else if (Qt::BackgroundRole == role && tag !=0) {

	  const bool intra = line[0] == line[1];
	  const bool inter = !intra;
	  const bool accepted = line[4]  == 1;
	  const bool rejected = !accepted;

	  if (intra && accepted) {
 		return QBrush(TrueMatch.lighter());
	  }
	  else if (intra && rejected) {
 		return QBrush(FalseNonMatch.lighter());
	  }
	  if (inter && accepted) {
 		return QBrush(FalseMatch.lighter());
	  }
	  else if (inter && rejected) {
 		return QBrush(TrueNonMatch.lighter());
	  }
  }
  return QVariant::Invalid;
}


QVariant BiometricDatabaseModel::headerData(int section, Qt::Orientation orientation, int role) const {
  if (role == Qt::DisplayRole && orientation == Qt::Horizontal) {
    switch (section) {
      case 0:
        return QString("Gallery");
      case 1:
        return QString("Probe");
      case 2:
        return QString("Sample nb");
      case 3:
        return QString(_isScore ? "Score" : "Distance");
      case 4:
        return QString("Verified");
    }

    return QVariant::Invalid;
  }


  return QVariant::Invalid;
}


// Specific methods
void BiometricDatabaseModel::updateDataTable(const InnerType &data, bool isScore) {

  beginResetModel();
  _isScore = isScore;
  //    clear(); // XXX do we need to call it ?
  _data = data;
  endResetModel();
}


BiometricDatabaseModel &getUniqueQtBiometricDataModel() {
  static BiometricDatabaseModel model{nullptr};
  return model;
}

void BiometricDatabaseModel::updateUniqueDecisionThreshold(const double thr) {
    beginResetModel();
    std::for_each(
                _data.begin(),
                _data.end(),
                [&](auto & line) {
                    line[4] = isComparisonScoreAccepted(line[3], thr) ? 1 : -1 ;
                }

                );
    endResetModel();
}
