/**
 * Copyright University Bordeaux - CNRS UMR 5800
 * contributor(s) : Romain Giot (2017)
 *
 * Romain.Giot@u-bordeaux.fr
 *
 * Biometric Power Graph (bpg) is a computer program whose purpose is to
 * evaluate biometric authentication
 * systems by displaying with a graph-based representation after having
 * transformed the data.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info.
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include "bpg/powergraphmanager.h"
#include "bpg/common.h"

#include <iostream>

#include <cassert>

#include "tulip/DoubleProperty.h"
#include "tulip/IntegerProperty.h"

#include "unistd.h"


const std::string get_backup_name() {
  return "/tmp/power_graph_" + std::to_string(getpid()) + ".tlpb.gz";
}

PowerGraphManager::PowerGraphManager()
: _gallery_to_filter{-1} {}

bool PowerGraphManager::importCSV(const std::string fname, const bool isScore, const std::string thumbnailFname) {
  assert(!isLoaded() && "Import can only be done when no graph exists");

  // delete existing graph if needed
  {
    // XXX is it necessary to delete its properties ?
    if (nullptr != _root) {
      delete _root;
    }
  }
  _root = nullptr;


  // Load the CSV file
  {
    tlp::DataSet dataset;
    dataset.set("file::filename", fname);
    dataset.set("isScore", isScore);
    dataset.set("file::thumbnail", thumbnailFname);

    _root = tlp::importGraph("Biometric Power Graph Import", dataset, nullptr,
                             nullptr);
  }

  // Compute the ROC curve in case of success
  if (nullptr != _root) {

    tlp::DataSet dataset{};
    std::string error;

    const bool res =
      _root->applyAlgorithm("Compute Biometric Properties", error, &dataset);

    if (!res) {
      std::cerr << "[ERROR] Compute Biometric Properties failed: " << error
                << std::endl;
    }
    return res;
  } else {
    return false;
  }
}

tlp::Graph *PowerGraphManager::getROCGraph() {
  assert(_root);

  tlp::Graph *roc = _root->getSubGraph("ROC");
  assert(roc);

  return roc;
}

tlp::Graph *PowerGraphManager::getDisplayGraph() {
  assert(_root);
  assert(displayAvailable());

  tlp::Graph *display = _root->getSubGraph("display");
  assert(display);

  tlp::Graph *filtered = display->getSubGraph("filtered");
  if (!filtered)
    filtered = buildFilteredGraph();
  assert(filtered);

  return filtered;
}

tlp::Graph *PowerGraphManager::buildFilteredGraph() {
  assert(displayAvailable());

  tlp::Graph *display = _root->getSubGraph("display");
  tlp::Graph *filtered = display->getSubGraph("filtered");

  assert(display);

  auto userId = _root->getProperty< tlp::IntegerProperty >(PROP_USER);

  // Ensure the graph does not exists
  if (filtered) {
    display->delSubGraph(filtered);
  }

  if (_gallery_to_filter == -1) {
    filtered = display->addCloneSubGraph("filtered");
  } else {
    tlp::node gallery, probe, sample;
    std::vector< tlp::node > to_keep;

    for (tlp::node gallery : getUsers(_root)) {

      // Add the user of interest and the associate samples
      if (userId->getNodeValue(gallery) == _gallery_to_filter) {
        to_keep.push_back(gallery);
      }
    }

    filtered = display->inducedSubGraph(to_keep, display, "filtered");

    std::cout << "Filtered " << filtered->numberOfNodes() << "nodes / "
              << filtered->numberOfEdges() << "edges\n";
  }

  return filtered;
}

RocNodeInformation PowerGraphManager::getInformationFromRocNode(tlp::node n) {
  tlp::Graph *roc = getROCGraph();
  assert(roc->isElement(n));

  tlp::DoubleProperty *propFNMR =
    roc->getProperty< tlp::DoubleProperty >(PROP_FNMR);
  tlp::DoubleProperty *propFMR =
    roc->getProperty< tlp::DoubleProperty >(PROP_FMR);
  tlp::DoubleProperty *propThreshold =
    roc->getProperty< tlp::DoubleProperty >(PROP_THR);

  return {propThreshold->getNodeValue(n), propFNMR->getNodeValue(n),
          propFMR->getNodeValue(n)};
}


void PowerGraphManager::backupState() {
  // Save the current graph on disc
  _root->push();
  // assert (_root->getSubGraph(LAYER_USERS)->getSubGraph("modified") == nullptr);
  // assert (_root->getSubGraph(LAYER_DISPLAY) == nullptr);
}


void PowerGraphManager::restoreState() {
  _root->pop();
  // assert (_root->getSubGraph(LAYER_USERS)->getSubGraph("modified") != nullptr);

  //_root->getSubGraph(LAYER_USERS)->delSubGraph(_root->getSubGraph(LAYER_USERS)->getSubGraph("modified"));
  // _root->delSubGraph(_root->getSubGraph(LAYER_DISPLAY));
}


/**
 * As there is a side effect to display the power graph (remember that the drawing plugin modifies some graphs), a clone 
 * is first creted
 */
bool PowerGraphManager::buildDisplayGraphForThreshold(const double thr,
                                                      const bool approximated,
                                                      tlp::Graph * layoutGraph,
                                                      const double partitionFNMRLowThr,
                                                      const double partitionFNMRHighThr,

                                                      const double partitionFMR_inLowThr,
                                                      const double partitionFMR_inHighThr,

                                                      const double partitionFMR_outLowThr,
                                                      const double partitionFMR_outHighThr
                                                      ) {
  this->thr = thr;


  if (this->displayAvailable()) {
    restoreState();
  }

  backupState();

  // TODO Manage the fact that it may not be the first time the algorithm is
  // called
  std::string error;
  tlp::DataSet dataset;
  dataset.set("threshold", thr);
  dataset.set("approximated", approximated);

  dataset.set("partitionFNMRLowThr", partitionFNMRLowThr);
  dataset.set("partitionFNMRHighThr", partitionFNMRHighThr);

  dataset.set("partitionFMR_inLowThr", partitionFMR_inLowThr);
  dataset.set("partitionFMR_inHighThr", partitionFMR_inHighThr);

  dataset.set("partitionFMR_outLowThr", partitionFMR_outLowThr);
  dataset.set("partitionFMR_outHighThr", partitionFMR_outHighThr);


  dataset.set<tlp::Graph*>("layoutGraph", layoutGraph);

  if (nullptr != layoutGraph) {
    std::cout << "[Layout grpah used]" << std::endl;
  }

  const bool res =
    _root->applyAlgorithm("Draw Biometric Power Graph", error, &dataset);

  if (!res) {
    std::cerr << "[ERROR] Unable to draw the power graph" << std::endl;
  }

  return res;
}

SampleInformation
PowerGraphManager::getInformationFromSampleNode(tlp::node sample) {
  assert(::isSample(_root, sample));
  SampleInformation info;

  const tlp::node owner = getOwner(_root, sample);
  assert(::isIndividual(_root, owner));

  auto query2gallery = _root->getSubGraph(LAYER_QUERY2GALLERY);
  auto prop = _root->getProperty< tlp::DoubleProperty >(PROP_SCORE);
  auto sampleId = _root->getProperty< tlp::IntegerProperty >(PROP_SAMPLE);
  auto userId = _root->getProperty< tlp::IntegerProperty >(PROP_USER);

  int nbInterComparison{0};
  int nbIntraComparison{0};
  int nbInterAccepted{0};
  int nbIntraAccepted{0};

  tlp::edge e;
  forEach (e, query2gallery->getOutEdges(sample)) {
    const tlp::node gallery = _root->target(e);
    assert(::isIndividual(_root, gallery));

    const double score = prop->getEdgeValue(e);
    const bool accepted = isComparisonScoreAccepted(score, thr);
    const bool isIntraScore = gallery == owner;

    if (!isIntraScore) {
      ++nbInterComparison;
      if (accepted)
        ++nbInterAccepted;
    } else {
      ++nbIntraComparison;
      info.genuineScore = score;
      if (accepted)
        ++nbIntraAccepted;
    }
  }

  //assert(nbIntraComparison == 1);

  info.sampleId = sampleId->getNodeValue(sample);
  info.ownerId = userId->getNodeValue(owner);
  info.isVerified = 1 == nbIntraAccepted;
  info.nbImpersonnateComparison = nbInterComparison;
  info.nbImpersonnation = nbInterAccepted;
  info.nbGenuineComparison = nbIntraComparison;

  return info;
}

IndividualInformation
PowerGraphManager::getInformationFromIndividualNode(tlp::node individual) {
  assert(::isIndividual(_root, individual));

  auto userId = _root->getProperty< tlp::IntegerProperty >(PROP_USER);
  return {userId->getNodeValue(individual),
          _root->getProperty< tlp::DoubleProperty >(PROP_FNMR)->getNodeValue(individual),
          _root->getProperty< tlp::DoubleProperty >(PROP_IN_FMR)
            ->getNodeValue(individual),
          _root->getProperty< tlp::DoubleProperty >(PROP_OUT_FMR)
            ->getNodeValue(individual)};
}



std::pair<const std::vector<double>, const std::vector<double>> PowerGraphManager::getZooPlotData() {
    assert(_root != nullptr);

    tlp::Graph * const users = _root->getSubGraph(LAYER_USERS);

    std::vector<double> meanGenuine(users->numberOfNodes());
    std::vector<double> meanImpostor(users->numberOfNodes());

  auto const selfScore = _root->getProperty< tlp::DoubleProperty >(PROP_MEAN_SELF_SCORE);
  auto const inScore = _root->getProperty< tlp::DoubleProperty >(PROP_MEAN_IN_SCORE);


  size_t pos = 0;
  for(tlp::node user: users->nodes()) {
    meanGenuine[pos] = selfScore->getNodeValue(user);
    meanImpostor[pos] = inScore->getNodeValue(user);
    ++pos;
  }


  return {meanGenuine, meanImpostor};

}



std::pair< std::vector<double>, std::vector<double> > PowerGraphManager::getIntraInterScores() {

    std::vector<double> intra, inter;

    for(tlp::edge e: getRoot()->getSubGraph(LAYER_QUERY2GALLERY)->edges()) {
        const tlp::node sample = getRoot()->source(e);
        assert(isSample(sample));

        const tlp::node gallery = getRoot()->target(e);
        assert (isIndividual(gallery));

        const tlp::node sampleOwner =  getUserOfSample(getRoot(), sample);
        assert (isIndividual(sampleOwner));

        const double score = getComparisonScore(getRoot(), e) ;
        const bool isIntra = gallery == sampleOwner;

        if (isIntra) {
            intra.push_back(score);
        }
        else {
            inter.push_back(score);
        }
    }

    return make_pair(intra, inter);
}
