/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include <tulip/TulipPluginHeaders.h>
#include <tulip/TlpQtTools.h>
#include <tulip/TulipFontAwesome.h>
#include <tulip/TulipViewSettings.h>
#include <tulip/StaticProperty.h>

#include <vector>
#include <algorithm>
#include <numeric>

#include "bpg/common.h"

#include <omp.h>


using namespace std;
using namespace tlp;

pair< vector< double >, vector< double > > getIntraInterScores(Graph *const root) {
  vector< double > intra;
  vector< double > inter;

  // Properties to store the average values of each user
  auto const selfScore = root->getProperty< DoubleProperty >(PROP_MEAN_SELF_SCORE);
  auto const inScore = root->getProperty< DoubleProperty >(PROP_MEAN_IN_SCORE);
  auto const outScore = root->getProperty< DoubleProperty >(PROP_MEAN_OUT_SCORE);

  node tmp;
  const std::vector<node> & users = getUsers(root);
  const size_t nbUsers = users.size();
  std::cout << nbUsers << " indiv to treat" << std::endl;

#pragma omp parallel for shared(users)
  for(size_t userIdx=0; userIdx<nbUsers; ++userIdx){
    const node currentUser = users.at(userIdx);
    tlp::edge e;

    auto selfScoreT = tlp::NodeStaticProperty< double>(root);
    auto inScoreT = tlp::NodeStaticProperty< double >(root);
    auto outScoreT = tlp::NodeStaticProperty< double>(root);

    decltype(intra) intraT;
    decltype(inter) interT;

    const auto nbIntraInitial = intra.size();
    const auto nbInterInitial = inter.size();

    // Collect the intra/inter scores
    for(tlp::edge e: getIntraComparisonsToUser(root, currentUser)) {
      intraT.push_back(getComparisonScore(root, e));
    }

    for(tlp::edge e: getInterComparisonsToUser(root, currentUser)) {
      interT.push_back(getComparisonScore(root, e));
    }

    const auto nbIntraAdded = intraT.size();
    const auto nbInterAdded = interT.size();
    // Get their minvalue
    if (nbIntraAdded > 0) {
      selfScoreT.setNodeValue(
        currentUser,
        accumulate(
          intraT.begin(),
          intraT.end(),
          0) /
          static_cast< double >(nbIntraAdded));
    } else {
      std::cerr << "[WARNING] user node " << currentUser << " has no intra scores / change of dataset" << std::endl;
    }

    if (nbInterAdded > 0) {
      inScoreT.setNodeValue(
        currentUser,
        accumulate(
          interT.begin(),
          interT.end(),
          0) /
          static_cast< double >(nbInterAdded));
    } else {
      std::cerr << "[WARNING] user node " << currentUser << " has no inter scores / change of dataset" << std::endl;
    }


    size_t count = 0;
    double val = 0;
    forEach (e, getInterComparisonIterator2(root, currentUser)) {
      val += getComparisonScore(root, e);
      ++count;
    }

    if (0 == count) {
      std::cerr << "[WARNING] user node " << currentUser << " is not registered in the system / change of dataset !!" << std::endl;
      outScoreT.setNodeValue(currentUser, 0);
    } else {
      outScoreT.setNodeValue(currentUser, val / static_cast< double >(count));
    }

#pragma omp critical(copyToOrigin)
    {
	    // Move into the real property
	    if (nbIntraAdded > 0) selfScore->setNodeValue(currentUser, selfScoreT.getNodeValue(currentUser));
	    if (nbInterAdded > 0) inScore->setNodeValue(currentUser, inScoreT.getNodeValue(currentUser));
	    outScore->setNodeValue(currentUser, outScoreT.getNodeValue(currentUser));

	    std::copy(intraT.begin(), intraT.end(), std::back_inserter(intra));
	    std::copy(interT.begin(), interT.end(), std::back_inserter(inter));
    }
  }

  auto const query2galleryLayer = root->getSubGraph(LAYER_QUERY2GALLERY);
  cout << intra.size() << " intra scores" << endl;
  cout << inter.size() << " inter scores" << endl;
  cout << query2galleryLayer->numberOfEdges() << " total scores" << endl;

  assert(intra.size() + inter.size() == query2galleryLayer->numberOfEdges());

  return make_pair(intra, inter);
}


/**
 *
 * Compute the EER and build the ROC curve at the same time
 *
 * Fawcet method is used to have a very fast computation
 * XXX  we assume the comparison is score
 */
pair< double, double > computeEERStuff(const vector< double > &intra, const vector< double > &inter, Graph *ROC = nullptr) {
  assert(intra.size() > 0);
  assert(inter.size() > 0);

  DoubleProperty *const propFNMR = ROC ? ROC->getProperty< DoubleProperty >(PROP_FNMR) : nullptr;
  DoubleProperty *const propFMR = ROC ? ROC->getProperty< DoubleProperty >(PROP_FMR) : nullptr;
  DoubleProperty *const propThreshold = ROC ? ROC->getProperty< DoubleProperty >(PROP_THR) : nullptr;
  LayoutProperty *const viewLayout = ROC ? ROC->getProperty< LayoutProperty >("viewLayout") : nullptr;
  IntegerProperty *const occurences = ROC ? ROC->getProperty< IntegerProperty >("nbOccurences") : nullptr;
  double EER = 1;
  double thr = 0;
  double smallest_diff = 1;


  const auto nbintra = intra.size();
  const auto nbinter = inter.size();


  // Build the sorted set of scores
  std::vector< std::pair< bool, double > > scores{nbintra + nbinter};
#pragma omp parallel for
  for (unsigned int i = 0; i < nbintra; ++i) {
    scores[i] = {true, intra[i]};
  }
#pragma omp parallel for
  for (unsigned int i = 0; i < nbinter; ++i) {
    scores[i + nbintra] = {false, inter[i]};
  }
  std::sort(
    scores.begin(),
    scores.end(),
    [](const std::pair< bool, double > &a, const std::pair< bool, double > &b) {
      return std::get< 1 >(a) < std::get< 1 >(b);
    });

  int TR{0};
  int FR{0};
  double fprev = std::numeric_limits< double >::min();

  double FMR{0};
  double FNMR{0};

  node previous;
  node EERnode;

  unsigned int count = 1; // Count the number of time the current threshold has been met
  unsigned int pos = 0;
  unsigned int selected_threshold = 0;
  // loop over the sorted scores
  for (const auto &value : scores) {
    const double &fi = std::get< 1 >(value);
    const double &current_threshold = fi;
    const bool &positive = std::get< 0 >(value);


    if (fi != fprev) { // XXX add or if last element
      FNMR = FR / static_cast< double >(nbintra);
      FMR = 1 - TR / static_cast< double >(nbinter);
      fprev = fi;

      const double current_diff = fabs(FMR - FNMR);
      if (current_diff < smallest_diff) {
        EER = (FMR + FNMR) / 2.0;
        thr = current_threshold;
        selected_threshold = pos;
        smallest_diff = current_diff;
      }


      // Iteratively build the ROC curve
      if (ROC) {
        node current = ROC->addNode();

        propFMR->setNodeValue(current, FMR);
        propFNMR->setNodeValue(current, FNMR);
        propThreshold->setNodeValue(current, current_threshold);
        viewLayout->setNodeValue(current, Coord(FMR, 1.0 - FNMR, 0));
        occurences->setNodeValue(current, count);

        if (previous.isValid()) {
          ROC->addEdge(previous, current);
        }
        previous = current;

        if (current_diff == smallest_diff){
          EERnode = current;
	  ROC->getRoot()->setAttribute("EER_node_id", EERnode.id); // Store the id of the node to retreive it easier
	}
      }

      count = 1;
    } else {
      ++count;
    }


    if (positive) {
      ++FR;
    } else {
      ++TR;
    }

    ++pos;
  } // end thresholds loop


    // Need to change the EER threshold because here an individual is acepted if the score is higher OR EQUAL wheras in the rest of the application, it is STRICTLY higher 
    while(selected_threshold >0 && std::get<1>(scores[selected_threshold]) == thr) {
        std::cout << std::get<1>(scores[selected_threshold]) << std::endl;
        --selected_threshold;
    }
    thr = std::get<1>(scores[selected_threshold]);

  if (ROC) {
    auto viewSize = ROC->getProperty< SizeProperty >("viewSize");
    auto viewColor = ROC->getProperty< ColorProperty >("viewColor");

    viewSize->setAllNodeValue(Size(0.01, 0.01, 0));
    viewColor->setAllNodeValue(Color::Red);

    viewSize->setNodeValue(EERnode, Size(0.02, 0.02, 0));
    viewColor->setNodeValue(EERnode, Color::Blue);
  }

  return {EER, thr};
}


class ComputeBiometricProperties : public tlp::Algorithm
{
public:
  PLUGININFORMATION("Compute Biometric Properties", "Romain Giot", "23/11/2016", "Compute common values for the evaluation of biometric systems", "1.0", "Biometrics")
  ComputeBiometricProperties(tlp::PluginContext *context)
  : tlp::Algorithm(context) {
  }


  bool run() {
    assert(graph->getRoot() == graph);

    vector< double > intra, inter;
    std::tie(intra, inter) = getIntraInterScores(graph);

    double EER, thr;
    Graph *ROC = graph->addSubGraph(LAYER_ROC_CURVE);
    std::tie(EER, thr) = computeEERStuff(intra, inter, ROC);


    graph->setAttribute(ATTR_NB_INTRA, static_cast<int>(intra.size()));
    graph->setAttribute(ATTR_NB_INTER, static_cast<int>(inter.size()));
    graph->setAttribute(ATTR_NB_DISTINCT_SCORES, static_cast<int>(ROC->numberOfNodes()));
    graph->setAttribute(ATTR_EER, EER);
    graph->setAttribute(ATTR_THR, thr);

    cout << "EER value: " << EER << endl;
    cout << "EER threshold: " << thr << endl;


    computeMenagerieStuff();


    return true;
  }


private:
  /**
   * Weight configuration comes from Exploiting the “Doddington Zoo” Effect in Biometric Fusion
   * @todo XXX Do we have to take these definitions ? or definitions related to the number of edges ?
   */
  void computeMenagerieStuff() {

    cout << "DO NOT COMPUTE THE MENAGERIE" << endl;
    return;

    auto sgUsers = graph->getSubGraph(LAYER_USERS);

    {
      // Goats negatively impact False Rejection Rate
      auto selfScore = graph->getProperty< DoubleProperty >(PROP_MEAN_SELF_SCORE);
      auto isGoat = graph->getProperty< BooleanProperty >(PROP_IS_GOAT);
      vector< node > above, below;
      std::tie(below, above) = getPercentilesNodes(
        sgUsers,
        selfScore,
        0.30);
      setSameValueFor(below, isGoat, true); // XXX score only
      cout << below.size() << "/" << sgUsers->numberOfNodes() << " goats" << endl;
    }


    {
      // Lambs can be easily mimicked
      // @todo modify to work with distances
      auto lambScore = graph->getProperty< DoubleProperty >(PROP_MEAN_MAX_PAIR_IN_SCORE);
      auto isLamb = graph->getProperty< BooleanProperty >(PROP_IS_LAMB);

      for (tlp::node user: getUsers(graph)) {

        // Get the max score for current impersonator
        double max_values_summed = 0;
        double nb = 0;
        for(tlp::node impersonator: getUsers(graph)) {
          if (impersonator == user)
            continue;

          // Get the max value
          assert(impersonator != user);
          edge e;
          double max = 0;
          bool first = 1;
          forEach (e, getComparisons(graph, user, impersonator)) {
            if (first) {
              first = false;
              max = getComparisonScore(graph, e);
            } else {
              max = std::max(max, getComparisonScore(graph, e));
            }
          }

          // and store it to compute means later
          max_values_summed += max;
          ++nb;
        }

        // Store this score
        lambScore->setNodeValue(user, max_values_summed / nb);
      }

      // Search the corresponding users
      vector< node > above, below;
      std::tie(below, above) = getPercentilesNodes(
        sgUsers,
        lambScore,
        0.10);
      setSameValueFor(above, isLamb, true); // XXX score only
      cout << below.size() << "/" << sgUsers->numberOfNodes() << " lambs" << endl;
    }

    /*
    {
        // Resolve goats/lamb ambiguity
        auto isLamb  = graph->getProperty<BooleanProperty>(PROP_IS_LAMB);
        auto isGoat = graph->getProperty<BooleanProperty>(PROP_IS_GOAT);

        node user;
        forEach(user, getUsersIterator(graph)) {
            if (isLamb->getNodeValue(user) && isGoat->getNodeValue(user)) {
                isLamb->setNodeValue(false);
            }
        }
    
    }
    */


    {
      // wolves impersonate easily other individuals
      auto isWolf = graph->getProperty< BooleanProperty >(PROP_IS_WOLF);
      auto wolfScore = graph->getProperty< DoubleProperty >(PROP_MEAN_MAX_PAIR_OUT_SCORE);

      for(tlp::node user : getUsers(graph)) {

        // Get the max score for current impersonated
        double max_values_summed = 0;
        double nb = 0;
        for(tlp::node impersonated : getUsers(graph)) {
          if (impersonated == user)
            continue;

          // Get the max value
          assert(impersonated != user);
          edge e;
          double max = 0;
          bool first = 1;
          forEach (e, getComparisons(graph, impersonated, user)) {
            if (first) {
              first = false;
              max = getComparisonScore(graph, e);
            } else {
              max = std::max(max, getComparisonScore(graph, e));
            }
          }

          // and store it to compute means later
          max_values_summed += max;
          ++nb;
        }

        // Store this score
        wolfScore->setNodeValue(user, max_values_summed / nb);
      }

      vector< node > above, below;
      std::tie(below, above) = getPercentilesNodes(
        sgUsers,
        wolfScore,
        0.10);
      setSameValueFor(above, isWolf, true); // XXX score only
      cout << below.size() << "/" << sgUsers->numberOfNodes() << " wolves" << endl;
    }


    {
      // Sheeps performs well

      auto selfScore = graph->getProperty< DoubleProperty >(PROP_MEAN_SELF_SCORE);
      auto isSheep = graph->getProperty< BooleanProperty >(PROP_IS_SHEEP);
      vector< node > above, below;
      std::tie(below, above) = getPercentilesNodes(
        sgUsers,
        selfScore,
        0.30);
      setSameValueFor(above, isSheep, true); // XXX score only

      // A sheep cannot be a goat, or a lamb (maybe it can be a wolf ?)
      auto isGoat = graph->getProperty< BooleanProperty >(PROP_IS_GOAT);
      auto isLamb = graph->getProperty< BooleanProperty >(PROP_IS_LAMB);

      size_t nbSheeps = 0;
      for(tlp::node user: getUsers(graph)) {
        if (isGoat->getNodeValue(user) && isLamb->getNodeValue(user)) {
          isSheep->setNodeValue(user, false);
        }

        if (isSheep->getNodeValue(user)) {
          ++nbSheeps;
        }
      }
      cout << nbSheeps << "/" << sgUsers->numberOfNodes() << " sheeps " << endl;
    }
  }
};


PLUGIN(ComputeBiometricProperties)
