/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */


#include <tulip/TulipPluginHeaders.h>
#include <tulip/TlpQtTools.h>
#include <tulip/TulipFontAwesome.h>
#include <tulip/TulipViewSettings.h>
#include <tulip/DrawingTools.h>
#include <tulip/StaticProperty.h>
#include <tulip/Observable.h>

#include <vector>
#include <algorithm>
#include <numeric>
#include <thread>

#include "bpg/common.h"
#include "bpg/multilevelEdgeBundling.h"
#include "bpg/ExecutionMonitor.h"

#include "omp.h"

using namespace std;
using namespace tlp;

//const float userSizeFactor = 1.4142135623730951;
const float userSizeFactor = 1.4142135623730951 /** 1.3*/;
const float metaNodeSizeFactor = 1.3 * 1.4142135623730951;


constexpr unsigned int POWERGRAPH_SHAPE = 100;
constexpr unsigned int HALF_CIRCLESHAPE = 90;

class DrawBiometricPowerGraph : public tlp::Algorithm
{
public:
  PLUGININFORMATION("Draw Biometric Power Graph", "Romain Giot", "06/07/2017", "Compute the necessary stuff to draw the biometric power graph", "1.1", "Biometrics")
  DrawBiometricPowerGraph(tlp::PluginContext *context)
  : tlp::Algorithm(context),
    layoutGraph{nullptr},
    display{nullptr}, threshold{0} {
    addInParameter< double >("threshold", "Decision threshold", "0");
    addInParameter< bool >("approximated", "Enable to show an approximated graph (some edges are merged)", "false");
    addInParameter< tlp::Graph *>("layoutGraph", "Similar graph used to steal the layout", "0");

    /*
      addDependency("Edge bundling", "1.3");
      addDependency("Curve Builder", "0.1");
*/
  }


  void init() {
    // We want to be sure to use a new display graph
    display = graph->getSubGraph(LAYER_DISPLAY);
    if (nullptr != display)
      graph->delSubGraph(display);
    display = graph->addSubGraph(LAYER_DISPLAY);
  }



  void setLabels() {
  
    auto const viewLabel = display->getProperty<tlp::StringProperty>("viewLabel");
    auto const id        = graph->getProperty<IntegerProperty>(PROP_USER);
    auto const viewLabelColor = display->getProperty<tlp::ColorProperty>("viewLabelColor");
    auto const viewLabelBorderColor = display->getProperty<tlp::ColorProperty>("viewLabelBorderColor");

    // TODO parallelize
    for(tlp::node user: getUsers(graph)) {
	    viewLabel->setNodeValue(user, std::to_string( id->getNodeValue(user)) );
	    viewLabelColor->setNodeValue(user, tlp::Color::Black);
	    viewLabelBorderColor->setNodeValue(user, tlp::Color::Black);

    }

  }

  void computeSamplesColors() {


    const tlp::Color trueMatch = tlp::Color::Green; // (successful genuine match)
    const tlp::Color falseNonMatch = tlp::Color::Gray; // (failed to match genuine user)
    const tlp::Color trueNonMatch = tlp::Color::Blue; // (failed impostor attack)
    const tlp::Color falseMatch = tlp::Color::Red; //  (successful impostor attack)

    // Need to use graph instead of display because this structure uses nodes id as idx ...
    tlp::NodeStaticProperty<tlp::Color> viewLeftColor(graph);
    tlp::NodeStaticProperty<tlp::Color> viewRightColor(graph);
    tlp::NodeStaticProperty<tlp::Color> viewMergedColor(graph);
    std::map<tlp::node, int> node2shape;

    tlp::Graph *const query2GalleryLayer = this->graph->getSubGraph(LAYER_QUERY2GALLERY);
    assert(nullptr != query2GalleryLayer);

        auto innerLogic = [&](
            const tlp::node queryNode,
           const tlp::node userNode, 
            const bool testedAgainstGenuineRefrence,
            const bool testedAgainstImpostorsRefrence) -> std::pair< const tlp::Color, const tlp::Color > {
          tlp::Color leftColor;
          tlp::Color rightColor;

          if (testedAgainstGenuineRefrence && !testedAgainstImpostorsRefrence) {
            const bool genuineSampleAccepted = display->existEdge(queryNode, userNode).isValid();
            // This is the normal side of the side
            if (genuineSampleAccepted) {
              return {trueMatch, trueMatch};
            } else {
              return {falseNonMatch, falseMatch};
            }

          } else if (testedAgainstImpostorsRefrence && !testedAgainstGenuineRefrence) {
            const bool attacksSucceded = display->outdeg(queryNode) > 0; // There is an edge only if the sample is able to authenticate another individual
            if (attacksSucceded) {
              return {falseMatch, falseMatch};
            } else {
              return {trueNonMatch, trueNonMatch};
            }
          } else if ((!testedAgainstGenuineRefrence) && (!testedAgainstImpostorsRefrence)) {
            return {tlp::Color::Black, tlp::Color::Black};
          } else {
            assert(testedAgainstGenuineRefrence);
            assert(testedAgainstImpostorsRefrence);

#pragma omp critical(node2shape)
            node2shape.insert({queryNode, HALF_CIRCLESHAPE}); //  Samples nodes will be displayed with two colors

            const bool genuineSampleAccepted = display->existEdge(queryNode, userNode).isValid();
            const bool attacksSucceded = genuineSampleAccepted ? display->outdeg(queryNode) > 1 : display->outdeg(queryNode) > 0;

            if (genuineSampleAccepted) {
              if (attacksSucceded) {
                return {trueMatch, falseMatch};
              } else {
                return {trueMatch, trueNonMatch};
              }
            } else {
              if (attacksSucceded) {
                return {falseNonMatch, falseMatch};
              } else {
                return {falseNonMatch, trueNonMatch};
              }
            }
          }
        };

    const std::vector<tlp::node> &users = getUsers(graph);
    const size_t nbUsers = users.size();

#pragma omp parallel for
    for(size_t userIdx=0; userIdx<nbUsers; ++userIdx) {
	    const tlp::node userNode= users[userIdx];
      node queryNode;

      if (getNumberOfSamplesForIndividual(graph, userNode)) for (tlp::node queryNode: getSamples(graph, userNode)) {

        const bool testedAgainstGenuineRefrence = query2GalleryLayer->existEdge(queryNode, userNode).isValid();
        const bool testedAgainstImpostorsRefrence = testedAgainstGenuineRefrence ? query2GalleryLayer->outdeg(queryNode) > 1 : query2GalleryLayer->outdeg(queryNode) > 0;


        const std::pair< const tlp::Color, const tlp::Color > colors = innerLogic(
            queryNode,
            userNode,
            testedAgainstGenuineRefrence,
            testedAgainstImpostorsRefrence);

        const auto & genuineColor  = std::get< 0 > (colors);
        const auto & impostorColor = std::get< 1 > (colors);

        viewLeftColor.setNodeValue(queryNode, genuineColor);
        viewRightColor.setNodeValue(queryNode, impostorColor);


        if (testedAgainstGenuineRefrence && testedAgainstImpostorsRefrence) {
            viewMergedColor.setNodeValue(queryNode, averagedColor(genuineColor, impostorColor));
        }
        else if (testedAgainstGenuineRefrence && !testedAgainstImpostorsRefrence){
            viewMergedColor.setNodeValue(queryNode, genuineColor);
        }
        else if (!testedAgainstGenuineRefrence && testedAgainstImpostorsRefrence) {
            viewMergedColor.setNodeValue(queryNode, impostorColor);
        }
        else {
            assert(false);
        }

      }
    }

    // Copy back properties
    tlp::ColorProperty *const viewLeftColorG = display->getProperty< ColorProperty >("viewBorderColor");
    tlp::ColorProperty *const viewRightColorG = display->getProperty< ColorProperty >("viewLabelColor");
    tlp::ColorProperty *const viewMergedColorG = display->getProperty< ColorProperty >("viewColor");

#if MORE_PARALLEL_STUFF
auto copyP = [&](const std::vector<tlp::node> & nodes, auto * from, auto * to) {
	    for_each(
			    nodes.begin(),
			    nodes.end(),
			    [&](tlp::node n) {
			    	to->setNodeValue(n, from->getNodeValue(n));
			    }
		    );
    };

    // Users
    std::thread thr1(copyP, std::ref(getUsers(graph)), (&viewLeftColor), (viewLeftColorG));
    std::thread thr2(copyP, std::ref(getUsers(graph)), (&viewRightColor), (viewRightColorG));
    std::thread thr3(copyP, std::ref(getUsers(graph)), (&viewMergedColor), (viewMergedColorG));
    // Samples
    std::thread thr4(copyP, std::ref(getSamples(graph)), (&viewLeftColor), (viewLeftColorG));
    std::thread thr5(copyP, std::ref(getSamples(graph)), (&viewRightColor), (viewRightColorG));
    std::thread thr6(copyP, std::ref(getSamples(graph)), (&viewMergedColor), (viewMergedColorG));

    thr1.join();
    thr2.join();
    thr3.join();


    thr4.join();
    thr5.join();
    thr6.join();
#else

    // Copy users properties
    for_each(
        getUsers(graph).begin(),
        getUsers(graph).end(),
        [&](tlp::node n) {
          viewLeftColorG->setNodeValue(n, viewLeftColor.getNodeValue(n));
          viewRightColorG->setNodeValue(n, viewRightColor.getNodeValue(n));
          viewMergedColorG->setNodeValue(n, viewMergedColor.getNodeValue(n));
        }
   );

    // Copy samples properties
    for_each(
        getSamples(graph).begin(),
        getSamples(graph).end(),
        [&](tlp::node n) {
          viewLeftColorG->setNodeValue(n, viewLeftColor.getNodeValue(n));
          viewRightColorG->setNodeValue(n, viewRightColor.getNodeValue(n));
          viewMergedColorG->setNodeValue(n, viewMergedColor.getNodeValue(n));
        }
   );




#endif

    // TODO do it in real time as we ause a critical section
    // TODO parallellize
    auto const viewShapeG = graph->getProperty< IntegerProperty >("viewShape");
    for_each(
        node2shape.begin(),
        node2shape.end(),
        [&](auto &item){
          viewShapeG->setNodeValue(item.first, item.second);
        });

  }


  /**
   * Fill the new structures and filter them thanks to the given threshold.
   * The process are:
   *
   *  - Add to display of all the samples
   *
   *  - XXX Attention, here we add links to the user graph
   *
   *  XXX Complete
   */
  void fillAndFilter() {
    auto const viewShape = graph->getProperty< IntegerProperty >("viewShape");
    auto const viewColor = graph->getProperty< ColorProperty >("viewColor");
    edge e;

    viewColor->setAllEdgeValue(Color(0, 0, 0, 100));

    // Add samples and users
    display->addNodes(getUsers(graph)); // Add all nodes in a row
    for (const tlp::node userNode : getUsers(graph)) {
      viewShape->setNodeValue(userNode, POWERGRAPH_SHAPE); // specific stuff for powergraphs
      if (getNumberOfSamplesForIndividual(graph, userNode)) {
	      display->addNodes(getSamples(graph, userNode)); // Add all the samples of the user
      }
    }


#ifndef NDEBUG
    const auto nbEdgesInitial = display->numberOfEdges();
#endif

    // Add the edges from a user to a sample
    // XXX TODO Stack an additional filter to not add edges where query and gallery are the same user (should speed up everything BUT imposes to modify the coloring code of the samples)

    const std::vector< tlp::edge > &acceptedComparisons = getAcceptedComparisonVector(graph, threshold);
    display->addEdges(acceptedComparisons);

#ifndef NDEBUG
    const auto nbEdgesAdded = display->numberOfEdges() - nbEdgesInitial;
    assert(nbEdgesAdded == acceptedComparisons.size());
#endif

    // Set the accepted boolean property to all the edges in query2gallery
    auto isQueryAccepted = graph->getProperty< BooleanProperty >(PROP_IS_QUERY_ACCEPTED);
    isQueryAccepted->setAllNodeValue(false);
    std::for_each(
      acceptedComparisons.begin(),
      acceptedComparisons.end(),
      [&](const tlp::edge e) {
        isQueryAccepted->setEdgeValue(e, true);
      });


    // Add edges between nodes WHEN they are linked thhanks to queries
    // XXX think to remove them later
    set< pair< int, int > > links;
    auto propUser = graph->getProperty< IntegerProperty >(PROP_USER);

    for (edge e : display->edges()) {
      const node source = display->source(e);
      const node target = display->target(e);

      links.insert({propUser->getNodeValue(source), propUser->getNodeValue(target)});
    }

    assert(!(graph->getSubGraph(LAYER_USERS)->getSubGraph("modified") != nullptr));

    this->modifiedUsers = graph->getSubGraph(LAYER_USERS)->addCloneSubGraph("modified");
    std::cout << this->modifiedUsers << this->modifiedUsers->getName() << std::endl;
    graph->getSubGraph(LAYER_USERS)->addCloneSubGraph("original");
    assert(modifiedUsers->numberOfEdges() == 0);
    for (auto &e : links) {
      modifiedUsers->addEdge(
          getLut(graph).getUser(e.first),
          getLut(graph).getUser(e.second)
          );
    }
  }


  /**
   * Each user owns a subgraph (the content of the power node). It is necessary to compute the local layout of these subgraphs before displaying the users
   */
  void drawLocalUsersGraph() {



    auto const viewSize = graph->getProperty< SizeProperty >("viewSize");

     // Adapt the size of the samples to their ability to recognize users
      Graph *const display = graph->getSubGraph(LAYER_DISPLAY);
      Graph *const samples = graph->getSubGraph(LAYER_SAMPLES);


      constexpr double min_sample_size = 1.0;
      constexpr double max_sample_size = 5.0;


      DoubleProperty *const degree = graph->getProperty< DoubleProperty >("degree");

      // Compute the degree of each sample without taking into account the edge with its owner
      // TODO parallelize
      for (tlp::node userNode : getUsers(graph)) {
        if (getNumberOfSamplesForIndividual(graph, userNode)) for(tlp::node queryNode: getSamples(graph, userNode)) {
          const int outdeg = display->outdeg(queryNode);
          const int realdeg = display->existEdge(queryNode, userNode).isValid() ? outdeg - 1 : outdeg;

          degree->setNodeValue(queryNode, realdeg);
        }
      }


      /**
       * In the standard case, the size of a sample depends on its degree
       */
      auto setRealSampleSize = [&](){
          DataSet dataSet = DataSet();
          string error;
          dataSet.set("property", degree);
          dataSet.set("min size", min_sample_size);
          dataSet.set("max size", max_sample_size);
          bool res = samples->applyPropertyAlgorithm("Size Mapping", viewSize, error, &dataSet);
          if (!res) {
              cout << "[ERROR] Unable to compute the size of the samples: " << error << "Set them to a fixed value" << endl;
              const tlp::Size normalSize(1,1,1);
              for(tlp::node sample: samples->nodes()){
                viewSize->setNodeValue(sample, normalSize);
              }

          }
      };

      /**
       * In the imported layout case, the size of a sample depends on the imported layout size
       */
      auto setOriginalSampleSize = [&](){
          assert(layoutGraph != nullptr);
          auto const viewSizeSource = layoutGraph->getProperty< SizeProperty >("viewSize");
          for(tlp::node n: layoutGraph->nodes()) {
            viewSize->setNodeValue(n, viewSizeSource->getNodeValue(n));
          }
      };

      if (layoutGraph) {
        setOriginalSampleSize();
      }else {
        setRealSampleSize();
      }


    // Draw the local graphs
    // XXX TODO integrate this code in the upper loop
    std::vector<tlp::node> indivWithoutSamples;
    tlp::Size smallestSize(std::numeric_limits<int>::max(), std::numeric_limits<int>::max(), std::numeric_limits<int>::max());
    for (tlp::node userNode : getUsers(graph)) {
	    if (getNumberOfSamplesForIndividual(graph, userNode) == 0) {
		    indivWithoutSamples.push_back(userNode);
		    continue;
	    } 

      Graph *const userGraph = getGraphForUser(graph, userNode);

      if (layoutGraph) {
	// For imported graphs, dumbly copy/paste the coordiantes
     	Graph *const importedUserGraph = getGraphForUser(layoutGraph, userNode);
	assert (importedUserGraph != nullptr);


	auto const viewLayout = userGraph->getLocalProperty< LayoutProperty >("viewLayout");
	auto const importedViewLayout = importedUserGraph->getLocalProperty< LayoutProperty >("viewLayout");

	for(tlp::node localSample: importedUserGraph->nodes()) {
		viewLayout->setNodeValue(localSample, 
				importedViewLayout->getNodeValue(localSample));
	}
      }
      else {
      	drawThisIndividualGraph(userGraph);
      }

      // Get the size of the local layout
      auto const localLayout = userGraph->getLocalProperty< LayoutProperty >("viewLayout");

      const BoundingBox bb = computeBoundingBox(
        userGraph,
        localLayout,
        graph->getProperty< SizeProperty >("viewSize"),
        graph->getProperty< DoubleProperty >("viewRotation"),
        nullptr);

      // and store it in the size property of the user node
      // XXX Buggy => does not take into account nodes size
      viewSize->setNodeValue(userNode, Size{bb.width() * userSizeFactor, bb.height() * userSizeFactor, bb.depth()});
      smallestSize.setX(std::min(bb.width(), smallestSize.getX()));
      smallestSize.setY(std::min(bb.height(), smallestSize.getY()));
      smallestSize.setZ(std::min(bb.depth(), smallestSize.getZ()));
    }

    // Set the size of the indiviudals without samples
    // TODO parallelize
    for(tlp::node userNode: indivWithoutSamples){
    	viewSize->setNodeValue(userNode, smallestSize);
    }


   }

  /**
   * Each user is repsented by a subgraph already drawn whose size has been reported to the node size of a user
   *
   * We want to cluster the users by their type of behavior. To do that a cluster score is computed from their appartenance to menagery stuff
   * The drawing of this graph consists to draw a cluster Power graph ...
   */
  void drawGlobalUserGraph() {
    Graph *const sg = this->modifiedUsers;
    Graph *const users = sg;
    Graph *quotientGraph = nullptr;
    assert(nullptr != sg);
    assert(nullptr != users);


    std::cout << users << std::endl;
    std::cout << users->getName() << std::endl;
    string error;
    DataSet dataSet;

    auto const usersLayout = users->getLocalProperty< LayoutProperty >("viewLayout");
    auto const clusterValue = graph->getProperty< DoubleProperty >(PROP_CLUSTER_VALUE);

    constexpr bool useThresholdedMenagerie = true;

    // Generate the clusters
    if (useThresholdedMenagerie) {
      auto const inFMR = graph->getProperty< DoubleProperty >(PROP_IN_FMR);
      auto const outFMR = graph->getProperty< DoubleProperty >(PROP_OUT_FMR);
      auto const FNMR = graph->getProperty< DoubleProperty >(PROP_FNMR);

      auto const clusterInFMR = graph->getProperty< IntegerProperty >(PROP_IN_FMR_CLUSTER);
      auto const clusterOutFMR = graph->getProperty< IntegerProperty >(PROP_OUT_FMR_CLUSTER);
      auto const clusterFNMR = graph->getProperty< IntegerProperty >(PROP_FNMR_CLUSTER);


      /**
       * Compute the average value of a property by taking into account the number of error
       * @todo It would be nice to parallel that
       */
      auto averagedValue = [&](DoubleProperty *const prop) {
        double sum = 0;
        size_t total = 0;

        for (tlp::node n : getUsers(graph)) {
          const double value = prop->getNodeValue(n);
          if (!isnan(value)) {
            sum += value;
            ++total;
          }
        }

        if (total == 0) {
          return static_cast< double >(NAN);
        } else {
          return sum / static_cast< double >(total);
        }
      };

      // Compute the averged value of FMR/FNMR amiong users
      const double averagedInFMR = averagedValue(inFMR);
      const double averagedOutFMR = averagedValue(outFMR);
      const double averagedFNMR = averagedValue(FNMR);


      std::cout << "averagedInFMR: " << averagedInFMR << std::endl;
      std::cout << "averagedOutFMR: " << averagedOutFMR << std::endl;
      std::cout << "averagedFNMR: " << averagedFNMR << std::endl;



        // When systems perform very badly, we HAVE TO change the value of these thresholds
        const double highInFMR = (partitionFMR_inHighThr < averagedInFMR) ? averagedInFMR : partitionFMR_inHighThr ;
        const double highOutFMR = (partitionFMR_outHighThr < averagedOutFMR) ? averagedOutFMR : partitionFMR_outHighThr ;
        const double highFNMR = (partitionFNMRHighThr < averagedFNMR) ? averagedFNMR : partitionFNMRHighThr;

        const double lowInFMR = (partitionFMR_inLowThr > averagedInFMR) ? averagedInFMR : partitionFMR_inLowThr ;
        const double lowOutFMR = (partitionFMR_outLowThr > averagedOutFMR) ? averagedOutFMR : partitionFMR_outLowThr ;
        const double lowFNMR = (partitionFNMRLowThr > averagedFNMR) ? averagedFNMR : partitionFNMRLowThr ;


         std::cout << "Low thresholds " << lowInFMR << "/" << lowOutFMR << "/" << lowFNMR << std::endl;
         std::cout << "High thresholds " << highInFMR << "/" << highOutFMR << "/" << highFNMR << std::endl;


      for (tlp::node user : getUsers(graph)) {
        // XXX Add these thresholds as parameters

        /**
         * Compute a score for the current value.
         * A high score is better tjan a low score.
         */
        auto score2code = [](const double lowThr, const double avrThr, const double highThr, const double value) {
          if (isnan(value)) {
            return 4;
          } else if (value < lowThr || value == 0) { // very best
            return 3;
          } else if (value < avrThr) { // best
            return 2;
          } else if (value < highThr) { //worst
            return 1;
          } else {
            return 0; // very worst
          }
        };


        const double currentInFMR = inFMR->getNodeValue(user);
        const double currentOutFMR = outFMR->getNodeValue(user);
        const double currentFNMR = FNMR->getNodeValue(user);

        const int inFMRCode = score2code(lowInFMR, averagedInFMR, highInFMR, currentInFMR);
        const int outFMRCode = score2code(lowOutFMR, averagedOutFMR, highOutFMR, currentOutFMR);
        const int FNMRCode = score2code(lowFNMR, averagedFNMR, highFNMR, currentFNMR);

        const int value = inFMRCode + (outFMRCode << 4) + (FNMRCode << 8);

        clusterValue->setNodeValue(user, value);

        clusterInFMR->setNodeValue(user, inFMRCode);
        clusterOutFMR->setNodeValue(user, outFMRCode);
        clusterFNMR->setNodeValue(user, FNMRCode);
      }

    } else {

      // In the first step, we want to cluster the graph
      auto isLamb = graph->getProperty< BooleanProperty >(PROP_IS_LAMB);
      auto isSheep = graph->getProperty< BooleanProperty >(PROP_IS_SHEEP);
      auto isGoat = graph->getProperty< BooleanProperty >(PROP_IS_GOAT);
      auto isWolf = graph->getProperty< BooleanProperty >(PROP_IS_WOLF);

      for (tlp::node user : getUsers(graph)) {
        int value = 0;

        value += isLamb->getNodeValue(user) ? 1 : 0;
        value += isSheep->getNodeValue(user) ? 2 : 0;
        value += isGoat->getNodeValue(user) ? 4 : 0;
        value += isWolf->getNodeValue(user) ? 8 : 0;

        clusterValue->setNodeValue(user, value);
      }
    }

    {

      // Generate the clusters
      if (layoutGraph) {
          // If a display graph is specified, the clusters MUST be reproduced from it 
          auto const clusterValueSource = layoutGraph->getProperty< DoubleProperty >(PROP_CLUSTER_VALUE);
          auto const clusterValueTarget = graph->getProperty< DoubleProperty >(PROP_CLUSTER_VALUE + "_imported");
          //Graph *const sgSource = layoutGraph->getSubGraph(LAYER_USERS)->getSubGraph("modified");

          // We assume the nodes numbering is exactly the same
          for(tlp::node n: getUsers(graph)) {
            clusterValueTarget->setNodeValue(n, clusterValueSource->getNodeValue(n));
          }
        dataSet.set("Property", clusterValueTarget);
      }
      else {
        dataSet.set("Property", clusterValue);
      }
      sg->applyAlgorithm("Equal Value", error, &dataSet, nullptr);


      // Draw each cluster in order to have its size
      Graph *cluster;
      unsigned int sgNb = 0;
      // TODO parallelize
      forEach (cluster, sg->getSubGraphs()) {

        sgNb += 1;
        auto clusterLayout = cluster->getLocalProperty< LayoutProperty >("viewLayout");


	// Quotient clustering
        if (nullptr != layoutGraph) {
            /*
             * Currently deactivate because this stuff is deported later
            auto clusterSize = cluster->getLocalProperty< SizeProperty >("viewSize");


            // Reproduce the layout from the original graph
            Graph *const sgSource = layoutGraph->getSubGraph(LAYER_USERS)->getSubGraph("modified");


            auto clusterSource = sgSource->getNthSubGraph(sgNb-1);
            std::cout << cluster->getName() << "/" << clusterSource->getName() << std::endl;
            assert(nullptr != clusterSource);
            auto clusterSourceLayout = clusterSource->getLocalProperty< LayoutProperty >("viewLayout");
            auto clusterSourceSize = clusterSource->getLocalProperty< SizeProperty >("viewSize");

            for(tlp::node n: cluster->nodes()) {
                assert(clusterSource->isElement(n));
                clusterLayout->setNodeValue(n, clusterSourceLayout->getNodeValue(n));
                clusterSize->setNodeValue(n, clusterSourceSize->getNodeValue(n));
            }
            */

        } else {
            cluster->applyPropertyAlgorithm(
                    "Fast Multipole Multilevel Embedder (OGDF)",
                    clusterLayout,
                    error);

            // Ensure there are no overlaps
            constexpr double overlap_border = 20.0;
            DataSet dataSet;
            dataSet.set("x border", overlap_border);
            dataSet.set("y border", overlap_border);
            cluster->applyPropertyAlgorithm(
                    "Fast Overlap Removal",
                    clusterLayout,
                    error,
                    &dataSet);

            // forward the unconnected components
            cluster->applyPropertyAlgorithm(
                    "Connected Component Packing",
                    clusterLayout,
                    error);
        }

      }


      // Create the quotient graph
      dataSet = DataSet();
      dataSet.set("use name of subgraph", true);
      sg->applyAlgorithm("Quotient Clustering", error, &dataSet, nullptr);

      const string quotientName = string("quotient of ") + sg->getName();
      quotientGraph = graph->getSubGraph(quotientName);
      auto const quotientLayout = quotientGraph->getLocalProperty< LayoutProperty >("viewLayout");
      auto const quotientSize = quotientGraph->getProperty< SizeProperty >("viewSize");
      {

        if (layoutGraph != nullptr) {
            auto const quotientSourceGraph = layoutGraph->getSubGraph(quotientName);
            auto const quotientSourceLayout = quotientSourceGraph->getLocalProperty< LayoutProperty >("viewLayout");
            auto const quotientSourceSize = quotientSourceGraph->getProperty< SizeProperty >("viewSize");

            for (node quotientNode : quotientGraph->nodes()) {

                quotientLayout->setNodeValue(
                    quotientNode,
                    quotientSourceLayout->getNodeValue(quotientNode));
	    quotientSize->getNodeValue(quotientNode);
	    quotientSourceSize->getNodeValue(quotientNode);
                }

        }
        else {

    // Fix node size of the quotient graph
            for (node quotientNode : quotientGraph->nodes()) {

            const auto newSize = metaNodeSizeFactor * quotientLayout->getNodeValue(quotientNode);
            quotientLayout->setNodeValue(
                quotientNode,
                newSize);
            }
        }
      }
      {



        quotientGraph->applyPropertyAlgorithm(
          "Bubble Tree", // XXX We have bundling issues tFMR\_{score}^{in}here => a circular layout hide them
          //"Circular",  // XXX We have bundling issues there => a circular layout hide them
          quotientLayout,
          error);
        cout << error;


/*
               constexpr double overlap_border  = 3.0;
               dataSet = DataSet() ;
               dataSet.set("x border", overlap_border); // XXX use a parameter ?
               dataSet.set("y border", overlap_border);
               quotientGraph->applyPropertyAlgorithm(
               "Fast Overlap Removal",
               quotientLayout,
               error,
               nullptr,
               &dataSet
               );
               */

// Move each cluster at the position of its representant

        for (node quotientNode : quotientGraph->nodes()) {
          const Coord center = quotientLayout->getNodeValue(quotientNode);

          Graph *cluster = quotientGraph->getNodeMetaInfo(quotientNode);
          auto clusterLayout = cluster->getLocalProperty< LayoutProperty >("viewLayout");
          clusterLayout->center(center, cluster);


          for (node n : cluster->nodes()) {
            usersLayout->setNodeValue(n, clusterLayout->getNodeValue(n));
          }
        }
      }



      if (layoutGraph) {
        std::cout << "Need to report sizes and positions" << std::endl;

	// Layout the quotient graph
        const string quotientName = string("quotient of ") + sg->getName();
        tlp::Graph * quotientGraph = graph->getSubGraph(quotientName); 
        tlp::Graph * quotientSourceGraph = layoutGraph->getSubGraph(quotientName);

	assert(nullptr != quotientGraph);
	assert(nullptr != quotientSourceGraph);

	const auto viewOtherLabel = quotientGraph->getProperty<StringProperty>("viewLabel");
	const auto viewSourceLabel = quotientSourceGraph->getProperty<StringProperty>("viewLabel");



	// XXX TODO Use a map to speed up that when it will really work
	auto getNodeWithName = [&](const tlp::Graph * g, const std::string &srcname) {
		static const std::string pattern = "_imported";
		const auto viewLabel = (g==quotientGraph) ? viewOtherLabel : viewSourceLabel;
		for(const tlp::node n: g->nodes()) {
			auto tgtname = viewLabel->getNodeValue(n);
			auto start_position_to_erase = tgtname.find(pattern);
			if (start_position_to_erase != std::string::npos) {
				tgtname.erase(start_position_to_erase, pattern.size());
			}
			if(srcname == tgtname){
				return n;
			}
		}

		return tlp::node();
	};

	assert(quotientGraph->numberOfNodes() == quotientSourceGraph->numberOfNodes());
#ifndef NDEBUG
	for (const tlp::node n: quotientSourceGraph->nodes()) {
		const std::string label = viewSourceLabel->getNodeValue(n);
		assert(label.size()>0);

		const tlp::node orig = getNodeWithName(quotientSourceGraph, label);
		const tlp::node other = getNodeWithName(quotientGraph, label);

		assert (orig.isValid());
		assert (other.isValid());
	}
#endif
        auto const quotientLayout = quotientGraph->getLocalProperty< LayoutProperty >("viewLayout");
        auto const quotientSourceLayout = quotientSourceGraph->getLocalProperty< LayoutProperty >("viewLayout");

        auto const quotientSize = quotientGraph->getLocalProperty< SizeProperty >("viewSize");
        auto const quotientSourceSize = quotientSourceGraph->getLocalProperty< SizeProperty >("viewSize");

	// TODO parallelize
        for(const tlp::node orig: quotientSourceGraph->nodes()) {
		const std::string label = viewSourceLabel->getNodeValue(orig);
		const tlp::node other = getNodeWithName(quotientGraph, label);
            quotientLayout->setNodeValue(other, quotientSourceLayout->getNodeValue(orig));
            quotientSize->setNodeValue(other, quotientSourceSize->getNodeValue(orig));
        }


        // Layout the nodes graph
        Graph *const sgSource = layoutGraph->getSubGraph(LAYER_USERS)->getSubGraph("modified");
        auto const sgLayout = sg->getLocalProperty< LayoutProperty >("viewLayout");
        auto const sgSourceLayout = sgSource->getLocalProperty< LayoutProperty >("viewLayout");
        auto const sgSize = sg->getLocalProperty< SizeProperty >("viewSize");
        auto const sgSourceSize = sgSource->getLocalProperty< SizeProperty >("viewSize");

	// TODO Parallelize
        for(tlp::node n: sg->nodes()) {
            assert (sgSource->isElement(n));
            sgLayout->setNodeValue(n, sgSourceLayout->getNodeValue(n));
            sgSize->setNodeValue(n, sgSourceSize->getNodeValue(n));
        }


      }

      std::cout << "Before bundling" << std::endl;

      // Manage the multilevel edge bundling
      MultilevelEdgeBundling bundling;
      if (approximatedDrawing) {
        bundling.multilevelEdgeBundling(quotientGraph, users, display);
      } else {
        bundling.multilevelEdgeBundling(quotientGraph, users, nullptr);
      }

      //graph->delSubGraph(quotientGraph);
    }


    { // Color each individual

      assert(nullptr != quotientGraph);
      assert(useThresholdedMenagerie); // Probably need to do differenty in this case

      // XXX Do a color mapping which depend on the threshold stuff
      // Set node color depending on the cluster value
      auto borderColor = users->getRoot()->getProperty< ColorProperty >("viewBorderColor");
      auto innerColor = users->getRoot()->getProperty< ColorProperty >("viewColor");
      auto borderSize = users->getRoot()->getProperty< DoubleProperty >("viewBorderWidth");
      auto description = users->getRoot()->getProperty< StringProperty >("Description");


      auto code_to_description = [](int code) {

        assert(code >= 0);
        assert(code < 4);

        const array< string, 4 > ranked = {{"very worst", "worst", "better", "very better"}};
        return ranked[code];
      };

      auto byte_to_description = [&](int value) {
        auto description1 = code_to_description(value & 0b11);
        auto description2 = code_to_description((value & 0b1100) >> 2);
        auto description3 = code_to_description((value & 0b110000) >> 4);

        return description1 + string(" inFMR") + string("\n") + description2 + string(" outFMR") + string("\n") + description3 + string(" FNMR");
      };


      // Color each quotient node and apply the same stuff to all its members
      const double fixed_size = 2;

      // TODO parallelize
      for (node quotientNode : quotientGraph->nodes()) {
        tlp::Graph *clusterGraph = quotientGraph->getNodeMetaInfo(quotientNode);

        // Get one node of the cluster (they all share the same information)
        tlp::node oneNode;
        // We may ahve bends nodes ...
        forEach (oneNode, clusterGraph->getNodes()) {
          if (isIndividual(graph, oneNode)) {
            break;
          }
        }
        assert(oneNode.isValid() && isIndividual(graph, oneNode));

        // Compute the cluster_value of interest
        const int cluster_value = static_cast< int >(clusterValue->getNodeValue(oneNode));
#if 0 
            const tlp::Color cluster_color = byte_to_color(cluster_value);
#else
        const tlp::Color cluster_color(tlp::Color::White);
#endif
        const std::string cluster_description = byte_to_description(cluster_value);

        std::cout<< " Quotient node " << quotientNode
                  << " cluster value : " << clusterValue
                  << " Color: " << cluster_color
                  << " Descriptioh: " << cluster_description
                  << " Nb users : " << clusterGraph->numberOfNodes()
                  << std::endl;

        // Apply it one the quotient graph ...
        borderColor->setNodeValue(quotientNode, cluster_color);
        borderSize->setNodeValue(quotientNode, fixed_size);
        description->setNodeValue(quotientNode, cluster_description);

// ... and on each user of this group

        for (node userNode : clusterGraph->nodes()) {
          innerColor->setNodeValue(userNode, cluster_color);
          borderSize->setNodeValue(userNode, fixed_size);
          description->setNodeValue(userNode, cluster_description);


          /*
           * XXX Temporarly removed as it overrite the half circle
          const double userFNMR = getFNMR(graph->getRoot(), userNode);
          borderColor->setNodeValue(userNode, tlp::Color(
                                                userFNMR * 255,
                                                0,
                                                (1 - userFNMR) * 255));
        */
        }
      }
    }
  }

  void translateSamples() {
    auto const displayLayoutG = display->getLocalProperty< LayoutProperty >("viewLayout");
    auto const userLayout = modifiedUsers->getLocalProperty< LayoutProperty >("viewLayout");

    // XXX unoptimal as it need a big array
    tlp::NodeStaticProperty<tlp::Coord> displayLayout(graph);

    const auto users = getUsers(graph);
    const size_t nbUsers = users.size();

#pragma omp parallel for
    for(size_t userIdx = 0; userIdx<nbUsers; ++userIdx){
      const tlp::node userNode = users[userIdx];

	    auto const userCenter = userLayout->getNodeValue(userNode);
	    displayLayout.setNodeValue(userNode, userCenter);

	    if (getNumberOfSamplesForIndividual(graph, userNode) == 0) {
		    continue;
	    }
	    Graph *const userGraph = getGraphForUser(graph, userNode);
	    assert (userGraph != nullptr);


	    auto const localLayout = userGraph->getLocalProperty< LayoutProperty >("viewLayout");


	    for(tlp::node sampleNode: getSamples(graph, userNode)) {
		    displayLayout.setNodeValue(
				    sampleNode,
				    localLayout->getNodeValue(sampleNode) + userCenter);
	    }
    }

    // copy back display
    for(tlp::node n: display->nodes()){
      displayLayoutG->setNodeValue(n, displayLayout.getNodeValue(n));
    }
  }

  // Revert the possible modifications in the graph
  void cleanup() {
    // TODO delete the modified user graph
  }


  /**
   * The drawing of each user graph is local to this graph
   * Graph is always centered
   */
  void drawThisIndividualGraph(Graph *userGraph) {
    string error;

    auto const viewLayout = userGraph->getLocalProperty< LayoutProperty >("viewLayout");

    userGraph->applyPropertyAlgorithm(
      "GRIP",
      viewLayout,
      error);


    viewLayout->center(userGraph);
  }

  /**
   * The edge bundling is a bit strange : bundling has been already done on the graph of users and control points are reported to the edges of samples
   */
  void edgeBundling() {
    Graph *const userGraph = modifiedUsers;
    Graph *const displayGraph = graph->getSubGraph(LAYER_DISPLAY);


    auto const viewLayoutUser = userGraph->getLocalProperty< LayoutProperty >("viewLayout");
    auto const viewLayoutDisplay = displayGraph->getLocalProperty< LayoutProperty >("viewLayout");

/*
    string error;
 // XXX Edge bundling has already been computed
    // Compute the bundling on the individuals layout
    DataSet dataSet;
    dataSet.set("layout", viewLayoutUser);
    dataSet.set("long_edges", 1.5);
    dataSet.set("iterations", 5);
    dataSet.set("plit_ratio", 20);
    userGraph->applyAlgorithm(
        "Edge bundling",
        error,
        &dataSet,
        nullptr
    );
*/
// Report it in the samples layout

    for (edge sample2UserEdge : displayGraph->edges()) {
      node sampleNode, userDestNode, userSourceNode;
      edge user2userEdge;

      // Retreive the user source sample
      std::tie(sampleNode, userDestNode) = displayGraph->ends(sample2UserEdge);
      assert(sampleNode.isValid());
      assert(userDestNode.isValid());
      userSourceNode = getUserOfSample(graph, sampleNode);

      // And the associated edge
      user2userEdge = userGraph->existEdge(userSourceNode, userDestNode, true);
      assert(user2userEdge.isValid() && "By construction, the edge must exist");

      // Copy the propery
      viewLayoutDisplay->setEdgeValue(
        sample2UserEdge,
        reduce_bends(viewLayoutUser->getEdgeValue(user2userEdge)));
    }


    if (false) {
      // XXX BUG does not work
      std::cout << "Curve simplification procedure" << std::endl;
      std::string error;


      StringCollection curveType("Bezier;Catmull-Rom;Cubic B-Spline");
      curveType.setCurrent(2);
      DataSet dataset;
      dataset.set("curve type", curveType);
      dataset.set("max bends", 20);
      dataset.set("min dev", sin(M_PI / (360. * 2)));
      dataset.set("max dev", sin(M_PI / (36. / 2)));

      // Display edges as polylines after simplification of CubicBSplineRendering
      displayGraph->applyPropertyAlgorithm(
        "Curve Builder",
        viewLayoutDisplay,
        error,
        &dataset);
    } else {

      // Simplification of the curves by removing some points / display as a CubicBSpline
      std::cout<< "TODO simplify the curves to speed up there generation" << std::endl;
      displayGraph->getProperty< IntegerProperty >("viewShape")->setAllEdgeValue(EdgeShape::CubicBSplineCurve);
    }

    displayGraph->getProperty< ColorProperty >("viewColor")->setAllEdgeValue(tlp::Color(0, 0, 0, 100));
  }


  void orderDrawing() {
    static const double sample_order = 20;
    static const double user_order = 10;
    static const double edge_order = 5;

    auto order = display->getProperty< DoubleProperty >("order");
    auto layout = display->getProperty< LayoutProperty >("viewLayout");

    node n;


    auto setZN = [&](const tlp::node &n, double z) {
      tlp::Coord c = layout->getNodeValue(n);
      c.setZ(z);
      layout->setNodeValue(n, c);
    };

    auto setZE = [&](const tlp::edge &e, double z) {
      std::vector< tlp::Coord > v = layout->getEdgeValue(e);
      for (tlp::Coord &c : v)
        c.setZ(z);
      layout->setEdgeValue(e, v);
    };

    for (node n : getUsers(graph)) {
      order->setNodeValue(n, user_order);
      setZN(n, -0.05);
    }

    for (node n : getSamples(graph)) {
      order->setNodeValue(n, sample_order);
      setZN(n, +0.05);
    }

    for (edge e : display->edges()) {
      order->setEdgeValue(e, edge_order);
      setZE(e, -1);
    }
  }

  void fixBadEdgeDisplay() {
    auto viewColor = graph->getProperty< ColorProperty >("viewColor");
    const Color c(0, 0, 0, 0);
    for (node u : getUsers(graph)) {
      for(tlp::edge e: getIntraComparisonsToUser(graph, u)) {
        viewColor->setEdgeValue(e, c);
      }
    }
  }


  /**
     * The biometric menagerie is computed by analysing scores without taking into account the selected configuration.
     * Here we try to mimick the menagarie for a specific configuration.
     *
     * Here we loop over all the edges of the query to gallery layer and check if they exist in the display graph which is a filtered version of it
     */
  void computeThresholdedMenagerie() {
    // Reducer fonction used in the accumulate code on boolean values
    auto reducer = [](const double &sum, const bool &value) {
      return sum + (value ? 1.0 : 0.0);
    };



    auto inFMR = tlp::NodeStaticProperty<double>(graph);
    auto outFMR = tlp::NodeStaticProperty<double>(graph);
    auto FNMR = tlp::NodeStaticProperty<double>(graph);


    // temporary vectors which keep the number of **errors**


    // Inout graphs for this code
    Graph *const query2gallery = graph->getSubGraph(LAYER_QUERY2GALLERY);

    // Loop over the individuals to compute their properties
    const std::vector< tlp::node > & users = getUsers(graph);
    const size_t nbUsers = users.size();

#pragma parallel for
    for (size_t userIdx=0; userIdx<nbUsers; ++userIdx){
      const tlp::node user = users[userIdx];
        assert(isIndividual(graph, user));

        vector< bool > intra, inter, inter2; // XXX put them out of the loop
        // List of nodes to manipulate
        tlp::node probe, gallery, sample, sample2, owner;
        tlp::edge comparison, ownership;
        /*
        { // Init
          intra.clear();
          inter.clear();
          inter2.clear();
        }*/

      // Get the comparison from one sample to this user
      forEach (comparison, query2gallery->getInEdges(user)) {

        std::tie(sample, gallery) = query2gallery->ends(comparison);
        assert(gallery == user);
        probe = getUserOfSample(graph, sample);

        if (gallery == probe) {
          // All intra comparisons SHOULD be present
          intra.push_back(!display->isElement(comparison));
        } else {
          // All inter comparisons SHOULD be absent
          inter.push_back(display->isElement(comparison));
        }
      }

      // Get the comparison from one sample of this user to another one
      if (getNumberOfSamplesForIndividual(graph, user)) for(tlp::node sample: getSamples(graph, user)){
        assert(user == getOwner(graph, sample));
        forEach (comparison, query2gallery->getOutEdges(sample)) {
          std::tie(sample2, gallery) = query2gallery->ends(comparison);
          assert(sample2 == sample);

          // Ensure we are not using a comparison between the sample and its owner
          if (user != gallery) {
            //inter2.push_back(display->isElement(comparison));
            inter2.push_back(::isComparisonScoreAccepted(::getComparisonScore(graph, comparison), threshold));
          }
        }
      }

      {
        // Compute the error metrics
        // Set nan, when it is impossible to compute them
        FNMR.setNodeValue(
          user,
          (intra.size() > 0) ? std::accumulate(intra.begin(), intra.end(), 0.0, reducer) / static_cast< double >(intra.size()) : NAN);

        inFMR.setNodeValue(
          user,
          inter.size() > 0 ? std::accumulate(inter.begin(), inter.end(), 0.0, reducer) / static_cast< double >(inter.size()) : NAN);


        outFMR.setNodeValue(
          user,
          inter2.size() > 0 ? std::accumulate(inter2.begin(), inter2.end(), 0.0, reducer) / static_cast< double >(inter2.size()) : NAN);


      }
    }
 
    // List of properties to fill
    auto inFMRG = graph->getProperty< DoubleProperty >(PROP_IN_FMR);
    auto outFMRG = graph->getProperty< DoubleProperty >(PROP_OUT_FMR);
    auto FNMRG = graph->getProperty< DoubleProperty >(PROP_FNMR);

    // copy in the real datastracture
    for(tlp::node n: getUsers(graph)) {
      inFMRG->setNodeValue(n, inFMR.getNodeValue(n));
      outFMRG->setNodeValue(n, outFMR.getNodeValue(n));
      FNMRG->setNodeValue(n, FNMR.getNodeValue(n));
    }
  }

  bool run() {
    execution_monitor::start("Layout::nodes_computation");
    assert(graph->getRoot() == graph);

    Observable::disableEventNotification();

    partitionFNMRLowThr = 0.05;
    partitionFNMRHighThr = 0.08;

    partitionFMR_inLowThr = 0.05;
    partitionFMR_inHighThr = 0.08;

    partitionFMR_outLowThr = 0.05;
    partitionFMR_outHighThr = 0.08;


    if (nullptr != dataSet) {
      dataSet->get< double >("threshold", threshold);
      dataSet->get< bool >("approximated", approximatedDrawing);
      dataSet->get< tlp::Graph * >("layoutGraph", layoutGraph);


      dataSet->get<double>("partitionFNMRLowThr", partitionFNMRLowThr);
      dataSet->get<double>("partitionFNMRHighThr", partitionFNMRHighThr);

      dataSet->get<double>("partitionFMR_inLowThr", partitionFMR_inLowThr);
      dataSet->get<double>("partitionFMR_inHighThr", partitionFMR_inHighThr);

      dataSet->get<double>("partitionFMR_outLowThr", partitionFMR_outLowThr);
      dataSet->get<double>("partitionFMR_outHighThr", partitionFMR_outHighThr);


    }

    // Save the display parameters in the graph in order to reuse them later
    graph->setAttribute("display_threshold", threshold);
    graph->setAttribute("display_partitionFNMRLowThr", partitionFNMRLowThr);
    graph->setAttribute("display_partitionFNMRHighThr", partitionFNMRHighThr);

    graph->setAttribute("display_partitionFMR_inLowThr", partitionFMR_inLowThr);
    graph->setAttribute("display_partitionFMR_inHighThr", partitionFMR_inHighThr);

    graph->setAttribute("display_partitionFMR_outLowThr", partitionFMR_outLowThr);
    graph->setAttribute("display_partitionFMR_outHighThr", partitionFMR_outHighThr);



    std::cout << "DrawBiometricPowerGraph plugin\n";
    std::cout << "\tthreshold: " << threshold << "\n";
    std::cout << "\tapproximated: " << approximatedDrawing << "\n";
    std::cout << "\tlayoutGraph: " << (nullptr ==layoutGraph ? "nullptr" : "set") << "\n";

    cout << "Init" << std::endl;
    init();
    
    cout << "Filtering threshold of " << threshold << endl;
    fillAndFilter();
 
    cout << "Colors" << std::endl;
    computeSamplesColors();

    cout << "Compute menagerie  " << endl;
    computeThresholdedMenagerie();

    cout << "Draw local users graph" << endl;
    drawLocalUsersGraph();

    cout << "Draw global users graph" << endl;
    drawGlobalUserGraph();

    if (!layoutGraph)
    {
	    cout << "Translate samples" << endl;
	    translateSamples();
    }
    else {
    	tlp::Graph * src = layoutGraph->getRoot()->getSubGraph(LAYER_DISPLAY);
    	tlp::Graph * tgt = graph->getRoot()->getSubGraph(LAYER_DISPLAY);

	assert (src!=nullptr);
	assert (tgt!=nullptr);
	assert(src->existProperty("viewLayout"));
	assert(tgt->existProperty("viewLayout"));
	auto srcLayout = src->getProperty<LayoutProperty>("viewLayout");
	auto tgtLayout = tgt->getProperty<LayoutProperty>("viewLayout");

	for (const tlp::node n: src->nodes()) {
		assert(tgt->isElement(n));
		tgtLayout->setNodeValue(n, srcLayout->getNodeValue(n));
		std::cout << tgtLayout->getNodeValue(n) << std::endl;
	}
    }
    cout << "Manage bundling" << endl;
    execution_monitor::start("Layout::edges_computation");
    approximatedDrawing = false;
    if (!approximatedDrawing) {
      edgeBundling();
    } else {
      Graph *displayGraph = graph->getSubGraph(LAYER_DISPLAY);
      displayGraph->getProperty< IntegerProperty >("viewShape")->setAllEdgeValue(EdgeShape::CubicBSplineCurve);
    }
    orderDrawing();
    fixBadEdgeDisplay();

    setLabels();
    // cleanup();


    execution_monitor::stop();
    Observable::enableEventNotification();
    return true;
  }


private:
  Graph *layoutGraph;
  Graph *display;
  Graph *modifiedUsers = nullptr;

  double threshold;
  bool approximatedDrawing; // true: bundling intersection nodes are added to the graph


  double  partitionFNMRLowThr;
  double  partitionFNMRHighThr;

  double  partitionFMR_inLowThr;
  double  partitionFMR_inHighThr;

  double  partitionFMR_outLowThr;
  double  partitionFMR_outHighThr;

};


PLUGIN(DrawBiometricPowerGraph)
