/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */


#include <tulip/TulipPluginHeaders.h>
#include <tulip/TlpQtTools.h>
#include <tulip/TulipFontAwesome.h>
#include <tulip/TulipViewSettings.h>
#include <tulip/GraphProperty.h>

#include <map>

#include <iterator>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <map>

#include "csv.h"

#include "bpg/common.h"
#include "bpg/BiometricDatabaseModel.h"
#include "bpg/BPGLookup.h"
#include "bpg/ExecutionMonitor.h"

#include <QApplication>
#include <QDebug>
#include <QFileInfo>
#include <QDir>

using namespace tlp;
using namespace std;


/**
 * @brief The BiometricPowerGraph class
 *
 * Load the database of scores and produces:
 *  - the graph data able to build the biometric power graph (yes, the name is wrongly chosen, the power graph is not build there)
 *  - the qt data to fill tables
 */
class ImportBiometricPowerGraph : public tlp::ImportModule
{

public:
  PLUGININFORMATION("Biometric Power Graph Import", "Romain Giot", "22/11/2016", "Imports a biometric powergraph from a Schucker Matrix", "0.2", "Biometrics")
  ImportBiometricPowerGraph(tlp::PluginContext *context)
  : ImportModule(context),
    usersLayer(nullptr), samplesLayer(nullptr), sample2ownerLayer{nullptr}, query2galleryLayer{nullptr},
    propScore(nullptr), propUserId{nullptr}, propUserSampleId{nullptr}, propUserGraph{nullptr} {
    addInParameter< std::string >("file::filename", "CSV file to import", "");
    addInParameter< bool >("isScore", "Set to true if it is a score and false if it is a distance", "true");
  }


  ~ImportBiometricPowerGraph() {
  }

private:
  Graph *usersLayer;
  Graph *samplesLayer;
  Graph *sample2ownerLayer;
  Graph *query2galleryLayer;

  DoubleProperty *propScore;
  IntegerProperty *propUserId;
  IntegerProperty *propUserSampleId;
  GraphProperty *propUserGraph;

  BPGLookup * lut = new BPGLookup();

public:
  bool importGraph() {
    string filename = "";
    string thumbnailFname = "";
    bool isScore = true;

    if (dataSet) {
      dataSet->get< string >("file::filename", filename);
      dataSet->get< bool >("isScore", isScore);
      dataSet->get< string >("file::thumbnail", thumbnailFname);
    }

    if (filename.empty()) {
      pluginProgress->setError("Filename is empty.");
      return false;
    }

    // Create the needed structures

    usersLayer = graph->addSubGraph(LAYER_USERS);
    samplesLayer = graph->addSubGraph(LAYER_SAMPLES);
    query2galleryLayer = graph->addSubGraph(LAYER_QUERY2GALLERY);
    sample2ownerLayer = graph->addSubGraph(LAYER_SAMPLE2OWNER);

    propScore = graph->getProperty< DoubleProperty >(PROP_SCORE);
    propUserId = graph->getProperty< IntegerProperty >(PROP_USER);
    propUserSampleId = graph->getProperty< IntegerProperty >(PROP_SAMPLE);
    propUserGraph = graph->getProperty< GraphProperty >(PROP_USER_GRAPH);

    try {
      importCSVContent(filename, isScore, thumbnailFname);
    } catch (const io::error::too_many_columns &error) {
      std::cerr << "[ERROR] Unable to load file. " << error.what() << std::endl;
      return false;
    }
    return true;
  }


private:
  void checkAssertions() {
    assert(nullptr != usersLayer);
    assert(nullptr != samplesLayer);
    assert(nullptr != query2galleryLayer);
    assert(nullptr != samplesLayer);

    assert(nullptr != propScore);
    assert(nullptr != propUserId);
    assert(nullptr != propUserId);
  }

  node getUserNode(unsigned int userId) {
    checkAssertions();

    node userNode;

    const auto it = lut->_userId2node.find(userId);
    if (lut->_userId2node.end() != it) {
      // Returns the node if it exists
      userNode = it->second;
    } else {
      // Create it otherwise
      userNode = usersLayer->addNode();
      lut->_userId2node.insert({userId, userNode});
      lut->_users.push_back(userNode);
      lut->_userNode2samplesNodes.insert({userNode, {}});
      propUserId->setNodeValue(userNode, userId);

      // As well as its underlying subgraph
      Graph *const userGraph = graph->addSubGraph("User: " + to_string(userId));
      lut->_userId2subgraph.insert({userId, userGraph});
      propUserGraph->setNodeValue(userNode, userGraph);

    }

    assert(userNode.isValid());
    return userNode;
  }


  node getSampleNode(const pair< unsigned int, unsigned int > sample_key) {
    checkAssertions();

    node sampleNode;
    // Returns the node if it exists
    auto it = lut->_sampleId2node.find(sample_key);
    if (lut->_sampleId2node.end() != it) {
      sampleNode = it->second;
    } else {
      // Create it otherwise
      const auto owner_id = get< 0 >(sample_key);
      const auto sample_id = get< 1 >(sample_key);

      const node owner_node = getUserNode(owner_id);
      sampleNode = samplesLayer->addNode();
      lut->_sampleId2node.insert({sample_key, sampleNode});
      lut->_samples.push_back(sampleNode);
      lut->_userNode2samplesNodes.at(owner_node).push_back(sampleNode);
      Graph *userGraph = getUserSubgraph(owner_id);

      propUserId->setNodeValue(sampleNode, owner_id);
      propUserSampleId->setNodeValue(sampleNode, sample_id);

      // And add it in the proper subgraph
      userGraph->addNode(sampleNode);

      // And additional layers
      sample2ownerLayer->addNode(sampleNode);
      sample2ownerLayer->addNode(owner_node);
      sample2ownerLayer->addEdge(sampleNode, owner_node);

    }

    return sampleNode;
  }

  Graph *getUserSubgraph(const node userNode) {
    return getUserSubgraph(propUserId->getNodeValue(userNode));
  }

  Graph *getUserSubgraph(const unsigned int userId) {
    checkAssertions();
    return lut->_userId2subgraph.at(userId);
  }

  /**
   * Read the samples information from the CSV file
   */
  void importCSVContent(const string &filename, const bool isScore, const string &thumbnailFname) {


    io::CSVReader< 4 > in(filename);
    in.set_header(
      CSV_GALLERY_USER,
      CSV_PROBE_USER,
      CSV_PROBE_USER_SAMPLE,
      CSV_SCORE); // XXX Check it is the real order


    double gallery_id, probe_id, sample_id;
    double score;

    if (!isScore) {
      qDebug() << QString("Need to transforme distances as scores");
    }


    size_t nbIntra = 0;
    size_t nbInter = 0;

    execution_monitor::start("Import::read_csv");
    BiometricDatabaseModel::InnerType rows;
    while (in.read_row(gallery_id, probe_id, sample_id, score)) {
	    rows.emplace_back(std::array< double, 5 >{{gallery_id, probe_id, sample_id, score, 0}});

	    // Transform distances to scores
	    if (!isScore)
		    score = -score;


	    const auto sample_key = make_pair(probe_id, sample_id);

	    //  Get or create the realted nodes
	    const node gallery_node = getUserNode(gallery_id);
	    const node probe_node = getUserNode(probe_id);
	    const node query_node = getSampleNode(sample_key);


	    assert(gallery_node.isValid());
	    assert(probe_node.isValid());
	    assert(query_node.isValid());

	    // Add the edge from the sample to the gallery
	    if (gallery_node == probe_node) {
		    query2galleryLayer->addNodes({probe_node, query_node});
		    ++nbIntra;
	    } else {
		    query2galleryLayer->addNodes({gallery_node, probe_node, query_node});
		    ++nbInter;
	    }
	    const edge e = query2galleryLayer->addEdge(query_node, gallery_node);
	    propScore->setEdgeValue(e, score);


	    // As we do not use a specific thread... unsure the UI does not hang out...
	    // XXX Remove that to load with a specific thread
	    if (rows.size() % 500) {
		    QCoreApplication::processEvents();
	    }
    }


    execution_monitor::start("Import::read_thumbnails");
  if (thumbnailFname.size() > 0) {
        qDebug() << QString("Import the pictures");
      tlp::StringProperty * thumbnailProp = graph->getProperty<tlp::StringProperty>("thumbnail");
      io::CSVReader< 3 > thumbnails(thumbnailFname);
      double probe_id, sample_id;
      std::string path;

      const std::string basename = (QFileInfo(QString::fromStdString(thumbnailFname)).absoluteDir().canonicalPath() + QDir::separator()).toStdString();
      while (thumbnails.read_row(probe_id, sample_id, path)) {
          const std::pair<unsigned int, unsigned int> sample_key = std::make_pair(probe_id, sample_id);
          auto it = lut->_sampleId2node.find(sample_key);
          assert(it != lut->_sampleId2node.end());

          // TODO transform path as the addition of the basename of thumbnailFname and path
          thumbnailProp->setNodeValue(it->second, basename + path);
      }
  }


  // Update the QT model (for the grid)

    execution_monitor::start("Import::generate_qt_model");
  BiometricDatabaseModel &qtmodel = getUniqueQtBiometricDataModel();
  qtmodel.updateDataTable(rows, isScore);

  qDebug() << QString("File read");


  std::cout << "After reading\nNb Intra: " << nbIntra << "\nNb Inter: " << nbInter << std::endl;

  assert(query2galleryLayer->numberOfEdges() == nbIntra + nbInter);

  usersLayer->addCloneSubGraph("clean");


  graph->setAttribute("LUT", lut);
  execution_monitor::stop();
  execution_monitor::print();

#ifndef NDEBUG
  lut->checkup(graph);
#endif
}
}
;


PLUGIN(ImportBiometricPowerGraph)
