/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

/**
 *
 * This file is part of Tulip (www.tulip-software.org)
 *
 * Authors: David Auber and the Tulip development Team
 * from LaBRI, University of Bordeaux
 *
 * Tulip is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Tulip is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 */

#include <QCursor>
#include <QEvent>
#include <QThread>
#include <QApplication>
#include <QTimeLine>
#include <QMouseEvent>

#include <tulip/TulipPluginHeaders.h>
#include <tulip/MouseInteractors.h>
#include <tulip/GlMainWidget.h>
#include <tulip/GlMainView.h>
#include <tulip/ForEach.h>
#include <tulip/GlCircle.h>
#include <tulip/OcclusionTest.h>
#include <tulip/GlBoundingBoxSceneVisitor.h>
#include <tulip/GlRect.h>
#include <tulip/GlTextureManager.h>
#include <tulip/GlGraphStaticData.h>
#include <tulip/Rectangle.h>
#include <tulip/QtGlSceneZoomAndPanAnimator.h>
#include <tulip/GlNode.h>
#include <tulip/GlEdge.h>
#include <tulip/NodeLinkDiagramComponent.h>
#include <tulip/GlLayer.h>

#include "BiometricNeighborhoodHighlighterInteractor.h"
#include "PluginNames.h"

#include <algorithm>

using namespace std;

namespace tlp {

class GraphLayoutMorphing : public AdditionalGlSceneAnimation
{

public:
  virtual ~GraphLayoutMorphing() {}
  GraphLayoutMorphing(Graph *graph, LayoutProperty *srcLayout, LayoutProperty *destLayout, LayoutProperty *viewLayout)
  : graph(graph), srcLayout(srcLayout), destLayout(destLayout), viewLayout(viewLayout) {}

  void animationStep(int animationStep) {
    for (tlp::node n : graph->getNodes()) {
      const Coord &startPos = srcLayout->getNodeValue(n);
      const Coord &endPos = destLayout->getNodeValue(n);
      viewLayout->setNodeValue(n, startPos + (animationStep / static_cast< float >(nbAnimationSteps) * (endPos - startPos)));
    }

    for (tlp::edge e : graph->getEdges()) {
      const vector< Coord > &srcBends = srcLayout->getEdgeValue(e);
      const vector< Coord > &destBends = destLayout->getEdgeValue(e);
      vector< Coord > newBends;

      for (size_t i = 0; i < destBends.size(); ++i) {
        newBends.push_back(srcBends[i] + (animationStep / static_cast< float >(nbAnimationSteps) * (destBends[i] - srcBends[i])));
      }

      viewLayout->setEdgeValue(e, newBends);
    }
  }

private:
  Graph *graph;
  LayoutProperty *srcLayout, *destLayout, *viewLayout;
};

class BiometricNeighborNodesEdgeLengthOrdering : public binary_function< node, node, bool >
{

public:
  BiometricNeighborNodesEdgeLengthOrdering(node centralNode, LayoutProperty *layout)
  : centralNode(centralNode), layout(layout) {}

  bool operator()(tlp::node n1, tlp::node n2) const {
    Coord centralNodeCoord = layout->getNodeValue(centralNode);
    Coord n1Coord = layout->getNodeValue(n1);
    Coord n2Coord = layout->getNodeValue(n2);
    float centralToN1Dist = centralNodeCoord.dist(n1Coord);
    float centralToN2Dist = centralNodeCoord.dist(n2Coord);
    return centralToN1Dist < centralToN2Dist;
  }

private:
  node centralNode;
  LayoutProperty *layout;
};

class MouseEventDiscardFilter : public QObject
{

public:
  bool eventFilter(QObject *, QEvent *event) {
    if (event->type() == QEvent::MouseMove ||
        event->type() == QEvent::MouseButtonDblClick ||
        event->type() == QEvent::MouseButtonPress ||
        event->type() == QEvent::MouseButtonRelease) {
      return true;
    }

    return false;
  }
};

const unsigned char maxCircleAlphaValue = 200;

PLUGIN(BiometricNeighborhoodHighlighterInteractor)

BiometricNeighborhoodHighlighterInteractor::BiometricNeighborhoodHighlighterInteractor(const PluginContext *)
: GLInteractorComposite(QIcon(":/i_biometric_neighborhood_highlighter.png"), "Highlight node neighborhood"), configWidget(NULL) {
}

BiometricNeighborhoodHighlighterInteractor::~BiometricNeighborhoodHighlighterInteractor() {
  delete configWidget;
}

bool BiometricNeighborhoodHighlighterInteractor::isCompatible(const std::string &viewName) const {
  return ((viewName == NodeLinkDiagramComponent::viewName) || (viewName == ViewName::HistogramViewName)
          //Interactor does not work
          //||(viewName==ViewName::MatrixViewName)
          || (viewName == ViewName::ScatterPlot2DViewName));
}

void BiometricNeighborhoodHighlighterInteractor::construct() {
  configWidget = new BiometricNeighborhoodHighlighterConfigWidget();

  push_back(new MousePanNZoomNavigator());
  _neighborhoodHighlighter.setConfigWidget(configWidget);
  push_back(&_neighborhoodHighlighter);
}

BiometricNeighborhoodHighlighter::BiometricNeighborhoodHighlighter()
: originalGraph(NULL), originalGlGraphComposite(NULL),
  neighborhoodGraph(NULL), glBiometricNeighborhoodGraph(NULL),
  neighborhoodGraphLayout(NULL), neighborhoodGraphCircleLayout(NULL),
  neighborhoodGraphOriginalLayout(NULL), neighborhoodGraphColors(NULL),
  neighborhoodGraphBackupColors(NULL), centralNodeLocked(false),
  circleLayoutSet(false), glWidget(NULL), neighborhoodDist(1),
  configWidget(NULL), circleAlphaValue(maxCircleAlphaValue),
  startAlpha(0), endAlpha(255), nbAnimSteps(0) {
}

BiometricNeighborhoodHighlighter::BiometricNeighborhoodHighlighter(const BiometricNeighborhoodHighlighter &neighborhoodHighlighter)
: neighborhoodGraph(NULL), glBiometricNeighborhoodGraph(NULL), neighborhoodGraphLayout(NULL), neighborhoodGraphCircleLayout(NULL),
  neighborhoodGraphOriginalLayout(NULL), neighborhoodGraphColors(NULL), neighborhoodGraphBackupColors(NULL),
  centralNodeLocked(false), circleLayoutSet(false), neighborhoodDist(1), circleAlphaValue(maxCircleAlphaValue) {
  configWidget = neighborhoodHighlighter.configWidget;
  connect(configWidget, SIGNAL(updateBiometricNeighborhoodGraph()), this, SLOT(updateBiometricNeighborhoodGraph()));
}

BiometricNeighborhoodHighlighter::~BiometricNeighborhoodHighlighter() {
  cleanupBiometricNeighborhoodGraph();
}

void BiometricNeighborhoodHighlighter::viewChanged(View *view) {
  originalGraph = NULL;

  if (view == NULL) {
    glWidget = NULL;
    return;
  }

  GlMainView *glView = static_cast< GlMainView * >(view);
  glWidget = glView->getGlMainWidget();
}


void BiometricNeighborhoodHighlighter::focusNode(tlp::node n) {
	std::cout << "Focus on " << n << std::endl;
}

bool BiometricNeighborhoodHighlighter::eventFilter(QObject *qo, QEvent *e) {


  _pgInteractor->eventFilter(qo, e);

  if (!glWidget->getScene()->getGlGraphComposite()) {
    std::cout << "BiometricNeighborhoodHighlighter::eventFilter disabled" << std::endl;
    return true;
  }

  checkIfGraphHasChanged();
  if (originalGraph == NULL) {
    originalGraph = glWidget->getScene()->getGlGraphComposite()->getGraph();
    originalGlGraphComposite = glWidget->getScene()->getGlGraphComposite();

    if (!glWidget->hasMouseTracking()) {
      glWidget->setMouseTracking(true);
    }

    Iterator< string > *it = originalGraph->getProperties();

    while (it->hasNext()) {
      string propertyName = it->next();
      PropertyInterface *property = originalGraph->getProperty(propertyName);

      if (property->getTypename() == "double") {
        configWidget->setPropertyToUse(propertyName);
      }
    }

    delete it;
  }

  SelectedEntity selectedEntity;

  /*
   // This code is currently deactivated for our purpose
  if (e->type() == QEvent::Wheel && centralNodeLocked && !circleLayoutSet) {
    QWheelEvent *we = static_cast<QWheelEvent *>(e);

    if (selectInAugmentedDisplayGraph(we->x(), we->y(), selectedEntity) && selectedEntity.getEntityType() == SelectedEntity::NODE_SELECTED) {
      if (selectedEntity.getComplexEntityId() == neighborhoodGraphCentralNode.id) {
        int numSteps = we->delta() / 120;
        neighborhoodDist += numSteps;

        if (neighborhoodDist < 1) {
          neighborhoodDist = 1;
        }

        configWidget->setCurrentMaxDistanceForReachableNodes(neighborhoodDist);
        //neighborhoodGraph->updateWithDistance(neighborhoodDist); // XXX Deactivated
        updateBiometricNeighborhoodGraphLayoutAndColors();
        delete glBiometricNeighborhoodGraph;
        updateGlBiometricNeighborhoodGraph();
        glWidget->redraw();
        return true;
      }
    }

  }
  else*/

  if (e->type() == QEvent::MouseMove) {
    QMouseEvent *qMouseEv = static_cast< QMouseEvent * >(e);

    if (!centralNodeLocked) {
      node tmpNode = selectNodeInOriginalGraph(glWidget, qMouseEv->x(), qMouseEv->y());

      if (tmpNode != selectedNode) {
        buildBiometricNeighborhoodGraph(tmpNode, originalGraph);
        selectedNode = tmpNode;
      }
    } else {
      *neighborhoodGraphColors = *neighborhoodGraphBackupColors;

      if (selectInAugmentedDisplayGraph(qMouseEv->x(), qMouseEv->y(), selectedEntity) && selectedEntity.getEntityType() == SelectedEntity::NODE_SELECTED) {
        if (selectedEntity.getComplexEntityId() != neighborhoodGraphCentralNode.id) {
          neighborhoodGraphColors->setNodeValue(node(selectedEntity.getComplexEntityId()), Color(0, 255, 0));
        } else {
          neighborhoodGraphColors->setNodeValue(node(selectedEntity.getComplexEntityId()), Color(0, 0, 255));
        }
      }

      if (selectedEntity.getEntityType() == SelectedEntity::NODE_SELECTED)
        selectedBiometricNeighborNode = node(selectedEntity.getComplexEntityId());
      else
        selectedBiometricNeighborNode = node();
    }

    glWidget->redraw();
    return true;
  } else if (e->type() == QEvent::MouseButtonPress && static_cast< QMouseEvent * >(e)->buttons() == Qt::LeftButton) {
    if (neighborhoodGraphCentralNode.isValid() && !centralNodeLocked) {
      // The user wants to lock one node
      emit individualOrSampleSelected(neighborhoodGraphCentralNode);
      centralNodeLocked = true;
    } else if (neighborhoodGraphCentralNode.isValid() && centralNodeLocked &&
               !circleLayoutSet && selectedBiometricNeighborNode.isValid() &&
               selectedBiometricNeighborNode == neighborhoodGraphCentralNode && neighborhoodDist == 1) {
      // The node is locked, the user wants to create the circle layout
      computeBiometricNeighborhoodGraphCircleLayout();
      computeBiometricNeighborhoodGraphBoundingBoxes();
      GraphLayoutMorphing *layoutMorphing = new GraphLayoutMorphing(neighborhoodGraph, neighborhoodGraphOriginalLayout, neighborhoodGraphCircleLayout, neighborhoodGraphLayout);
      performZoomAndPan(neighborhoodGraphCircleLayoutBB, layoutMorphing);
      delete layoutMorphing;
      circleLayoutSet = true;
    } else if (neighborhoodGraphCentralNode.isValid() && circleLayoutSet && !selectedBiometricNeighborNode.isValid()) {
      circleLayoutSet = false;
      centralNodeLocked = false;
      neighborhoodGraphCentralNode = node();
      neighborhoodDist = 1;
      configWidget->setCurrentMaxDistanceForReachableNodes(neighborhoodDist);
      glWidget->redraw();
    } else if (selectedBiometricNeighborNode.isValid()) {
      if (circleLayoutSet) {
        if (selectedBiometricNeighborNode == neighborhoodGraphCentralNode) {
          GraphLayoutMorphing *layoutMorphing = new GraphLayoutMorphing(neighborhoodGraph, neighborhoodGraphCircleLayout, neighborhoodGraphOriginalLayout, neighborhoodGraphLayout);
          performZoomAndPan(neighborhoodGraphOriginalLayoutBB, layoutMorphing);
          delete layoutMorphing;
          circleLayoutSet = false;
        } else {
          node currentCentralNode = neighborhoodGraphCentralNode;
          buildBiometricNeighborhoodGraph(selectedBiometricNeighborNode, originalGraph);
          computeBiometricNeighborhoodGraphCircleLayout();
          computeBiometricNeighborhoodGraphBoundingBoxes();
          BoundingBox destBB;

          if (configWidget->bringAndGoAnimation1()) {
            destBB = neighborhoodGraphOriginalLayoutBB;
          } else {
            destBB = neighborhoodGraphCircleLayoutBB;
          }

          buildBiometricNeighborhoodGraph(currentCentralNode, originalGraph);
          computeBiometricNeighborhoodGraphCircleLayout();
          computeBiometricNeighborhoodGraphBoundingBoxes();
          *neighborhoodGraphLayout = *neighborhoodGraphCircleLayout;
          GraphLayoutMorphing *layoutMorphing = new GraphLayoutMorphing(neighborhoodGraph, neighborhoodGraphCircleLayout, neighborhoodGraphOriginalLayout, neighborhoodGraphLayout);
          performZoomAndPan(destBB, layoutMorphing);
          delete layoutMorphing;

          if (configWidget->bringAndGoAnimation1()) {
            morphCircleAlpha(maxCircleAlphaValue, 0);
          }

          buildBiometricNeighborhoodGraph(selectedBiometricNeighborNode, originalGraph);
          computeBiometricNeighborhoodGraphCircleLayout();
          computeBiometricNeighborhoodGraphBoundingBoxes();

          if (configWidget->bringAndGoAnimation1()) {
            morphCircleAlpha(0, maxCircleAlphaValue);
          }

          layoutMorphing = new GraphLayoutMorphing(neighborhoodGraph, neighborhoodGraphOriginalLayout, neighborhoodGraphCircleLayout, neighborhoodGraphLayout);
          performZoomAndPan(neighborhoodGraphCircleLayoutBB, layoutMorphing);
          delete layoutMorphing;
        }
      } else if (selectedBiometricNeighborNode != neighborhoodGraphCentralNode) {
        neighborhoodDist = 1;
        configWidget->setCurrentMaxDistanceForReachableNodes(neighborhoodDist);
        buildBiometricNeighborhoodGraph(selectedBiometricNeighborNode, originalGraph);
        glWidget->redraw();
      }
    } else {
      centralNodeLocked = false;
      circleLayoutSet = false;
      neighborhoodDist = 1;
      configWidget->setCurrentMaxDistanceForReachableNodes(neighborhoodDist);
      neighborhoodGraphCentralNode = node();
      glWidget->redraw();
    }

    return true;
  }

  return false;
}
node BiometricNeighborhoodHighlighter::selectNodeInOriginalGraph(GlMainWidget *glWidget, int x, int y) {
  node n;
  glWidget->makeCurrent();
  vector< SelectedEntity > selectedElements;
  glWidget->getScene()->selectEntities(static_cast< RenderingEntitiesFlag >(RenderingNodes | RenderingWithoutRemove), glWidget->screenToViewport(x - 1), glWidget->screenToViewport(y - 1), glWidget->screenToViewport(3), glWidget->screenToViewport(3), NULL, selectedElements);

  if (!selectedElements.empty()) {
    n = node(selectedElements[0].getComplexEntityId());
  }

  return n;
}

bool BiometricNeighborhoodHighlighter::selectInAugmentedDisplayGraph(const int x, const int y, SelectedEntity &selectedEntity) {
  GlLayer *l = glWidget->getScene()->getLayer("Main");
  GlGraphComposite *graphComposite = static_cast< GlGraphComposite * >(l->findGlEntity("graph"));
  l->deleteGlEntity("graph");
  l->addGlEntity(glBiometricNeighborhoodGraph, "graph");
  bool ret = glWidget->pickNodesEdges(x, y, selectedEntity);
  l->deleteGlEntity("graph");
  l->addGlEntity(graphComposite, "graph");
  return ret;
}

void BiometricNeighborhoodHighlighter::performZoomAndPan(const BoundingBox &destBB, AdditionalGlSceneAnimation *additionalAnimation) {
  MouseEventDiscardFilter medf;
  glWidget->installEventFilter(&medf);
  QtGlSceneZoomAndPanAnimator sceneZoomAndPan(glWidget, destBB);

  if (additionalAnimation != NULL) {
    sceneZoomAndPan.setAdditionalGlSceneAnimation(additionalAnimation);
  }

  sceneZoomAndPan.animateZoomAndPan();
  glWidget->removeEventFilter(&medf);
}

void BiometricNeighborhoodHighlighter::cleanupBiometricNeighborhoodGraph() {
  delete glBiometricNeighborhoodGraph;
  glBiometricNeighborhoodGraph = NULL;
  delete neighborhoodGraph;
  neighborhoodGraph = NULL;
  delete neighborhoodGraphLayout;
  neighborhoodGraphLayout = NULL;
  delete neighborhoodGraphCircleLayout;
  neighborhoodGraphCircleLayout = NULL;
  delete neighborhoodGraphOriginalLayout;
  neighborhoodGraphOriginalLayout = NULL;
  delete neighborhoodGraphColors;
  neighborhoodGraphColors = NULL;
  delete neighborhoodGraphBackupColors;
  neighborhoodGraphBackupColors = NULL;
}

void BiometricNeighborhoodHighlighter::buildBiometricNeighborhoodGraph(node n, Graph *g) {

  neighborhoodGraphCentralNode = node();

  cleanupBiometricNeighborhoodGraph();

  if (n.isValid()) {

    neighborhoodGraphCentralNode = n;

    //neighborhoodGraph = new NodeBiometricNeighborhoodView(g, n, configWidget->getBiometricNeighborsType(), neighborhoodDist, configWidget->computeReachableSubGraph(), configWidget->propertyToUse(), configWidget->numberOfNodesToBring());
    neighborhoodGraph = new BiometricNodeNeighborhoodView(g, n);

    neighborhoodGraphLayout = new LayoutProperty(neighborhoodGraph);
    neighborhoodGraphCircleLayout = new LayoutProperty(neighborhoodGraph);
    neighborhoodGraphOriginalLayout = new LayoutProperty(neighborhoodGraph);
    neighborhoodGraphColors = new ColorProperty(neighborhoodGraph);
    neighborhoodGraphBackupColors = new ColorProperty(neighborhoodGraph);

    updateBiometricNeighborhoodGraphLayoutAndColors();
    updateGlBiometricNeighborhoodGraph();

    circleCenter = neighborhoodGraphLayout->getNodeValue(n);
  }
}

void BiometricNeighborhoodHighlighter::updateBiometricNeighborhoodGraphLayoutAndColors() {

  if (originalGlGraphComposite != NULL) {
    GlGraphInputData *originalInputData = originalGlGraphComposite->getInputData();
    LayoutProperty *origGraphLayout = originalInputData->getElementLayout();
    ColorProperty *origGraphColors = originalInputData->getElementColor();

    for (tlp::node n2 : neighborhoodGraph->nodes()) {
      auto c = origGraphLayout->getNodeValue(n2);
      c.setZ(3);
      neighborhoodGraphOriginalLayout->setNodeValue(n2, c);
      neighborhoodGraphBackupColors->setNodeValue(n2, origGraphColors->getNodeValue(n2));
    }

    for (tlp::edge e : neighborhoodGraph->getEdges()) {
      neighborhoodGraphOriginalLayout->setEdgeValue(e, origGraphLayout->getEdgeValue(e));
      neighborhoodGraphBackupColors->setEdgeValue(e, origGraphColors->getEdgeValue(e));
    }

    *neighborhoodGraphLayout = *neighborhoodGraphOriginalLayout;
    *neighborhoodGraphColors = *neighborhoodGraphBackupColors;
  }
}

void BiometricNeighborhoodHighlighter::updateGlBiometricNeighborhoodGraph() {
  GlGraphInputData *originalInputData = originalGlGraphComposite->getInputData();
  glBiometricNeighborhoodGraph = new GlGraphComposite(neighborhoodGraph);
  GlGraphInputData *glBiometricNeighborhoodGraphInputData = glBiometricNeighborhoodGraph->getInputData();
  glBiometricNeighborhoodGraphInputData->setElementBorderColor(originalInputData->getElementBorderColor());
  glBiometricNeighborhoodGraphInputData->setElementBorderWidth(originalInputData->getElementBorderWidth());
  glBiometricNeighborhoodGraphInputData->setElementColor(neighborhoodGraphColors);
  glBiometricNeighborhoodGraphInputData->setElementLabel(originalInputData->getElementLabel());
  glBiometricNeighborhoodGraphInputData->setElementLabelColor(originalInputData->getElementLabelColor());
  glBiometricNeighborhoodGraphInputData->setElementLabelBorderColor(originalInputData->getElementLabelBorderColor());
  glBiometricNeighborhoodGraphInputData->setElementLabelPosition(originalInputData->getElementLabelPosition());
  glBiometricNeighborhoodGraphInputData->setElementLayout(neighborhoodGraphLayout);
  glBiometricNeighborhoodGraphInputData->setElementRotation(originalInputData->getElementRotation());
  glBiometricNeighborhoodGraphInputData->setElementSelected(originalInputData->getElementSelected());
  glBiometricNeighborhoodGraphInputData->setElementShape(originalInputData->getElementShape());
  glBiometricNeighborhoodGraphInputData->setElementSize(originalInputData->getElementSize());
  glBiometricNeighborhoodGraphInputData->setElementTexture(originalInputData->getElementTexture());
  glBiometricNeighborhoodGraphInputData->setElementFont(originalInputData->getElementFont());
  glBiometricNeighborhoodGraphInputData->setElementFontSize(originalInputData->getElementFontSize());
  glBiometricNeighborhoodGraphInputData->setElementSrcAnchorShape(originalInputData->getElementSrcAnchorShape());
  glBiometricNeighborhoodGraphInputData->setElementSrcAnchorSize(originalInputData->getElementSrcAnchorSize());
  glBiometricNeighborhoodGraphInputData->setElementTgtAnchorShape(originalInputData->getElementTgtAnchorShape());
  glBiometricNeighborhoodGraphInputData->setElementTgtAnchorSize(originalInputData->getElementTgtAnchorSize());
}

void BiometricNeighborhoodHighlighter::updateBiometricNeighborhoodGraph() {
  if (neighborhoodGraphCentralNode.isValid()) {
    buildBiometricNeighborhoodGraph(neighborhoodGraphCentralNode, originalGraph);
    glWidget->redraw();
  }
}

void BiometricNeighborhoodHighlighter::computeBiometricNeighborhoodGraphBoundingBoxes() {
  Coord centralNodeCoord = neighborhoodGraphOriginalLayout->getNodeValue(neighborhoodGraphCentralNode);
  float originalLayoutRadius = computeBiometricNeighborhoodGraphRadius(neighborhoodGraphOriginalLayout);
  float circleLayoutRadius = computeBiometricNeighborhoodGraphRadius(neighborhoodGraphCircleLayout);
  neighborhoodGraphOriginalLayoutBB[0] = centralNodeCoord - Coord(originalLayoutRadius, originalLayoutRadius);
  neighborhoodGraphOriginalLayoutBB[1] = centralNodeCoord + Coord(originalLayoutRadius, originalLayoutRadius);
  neighborhoodGraphCircleLayoutBB[0] = centralNodeCoord - Coord(circleLayoutRadius, circleLayoutRadius);
  neighborhoodGraphCircleLayoutBB[1] = centralNodeCoord + Coord(circleLayoutRadius, circleLayoutRadius);
}

void BiometricNeighborhoodHighlighter::computeBiometricNeighborhoodGraphCircleLayout() {
  // TODO Remove tulip iterators

  Size centralNodeSize = originalGlGraphComposite->getInputData()->getElementSize()->getNodeValue(neighborhoodGraphCentralNode);
  Coord centralNodeCoord = neighborhoodGraphLayout->getNodeValue(neighborhoodGraphCentralNode);
  neighborhoodGraphCircleLayout->setNodeValue(neighborhoodGraphCentralNode, centralNodeCoord);

  vector< node > neighborsNodes;
  vector< node > extraNodes; // these extra nodes correspond to the sample nodes when the locked node is an individual
  node n;

  if (neighborhoodGraph->getType() == BiometricNodeNeighborhoodView::NeighborhoodType::SampleCenter) {
    // When the center is an individual, we need to draw samples WITH their owner
    for (tlp::node n : neighborhoodGraph->nodes()) {
      if (n != neighborhoodGraphCentralNode) {
        neighborsNodes.push_back(n);
      }
    }
  } else {
    assert(neighborhoodGraph->getType() == BiometricNodeNeighborhoodView::NeighborhoodType::IndividualCenter);
    for (tlp::node n : neighborhoodGraph->nodes()) {
      if (n != neighborhoodGraphCentralNode) {
        if (!neighborhoodGraph->nodeIsSample(n)) {
          neighborsNodes.push_back(n);
        } else {
          extraNodes.push_back(n);
        }
      }
    }
  }

  // TODO Find a better sort mesure ?
  sort(neighborsNodes.begin(), neighborsNodes.end(), BiometricNeighborNodesEdgeLengthOrdering(neighborhoodGraphCentralNode, neighborhoodGraphLayout));

  BoundingBox centralNodeBB(Coord(centralNodeCoord.getX() - centralNodeSize.getW() / 2, centralNodeCoord.getY() - centralNodeSize.getH() / 2, 0),
                            Coord(centralNodeCoord.getX() + centralNodeSize.getW() / 2, centralNodeCoord.getY() + centralNodeSize.getH() / 2, 0));
  vector< BoundingBox > neighborhoodNodesNewLayoutBB;
  neighborhoodNodesNewLayoutBB.resize(neighborsNodes.size());

  for (size_t i = 0; i < neighborsNodes.size(); ++i) {

    Size neighborNodeSize = originalGlGraphComposite->getInputData()->getElementSize()->getNodeValue(neighborsNodes[i]);
    Coord neighborNodeCoord = neighborhoodGraphLayout->getNodeValue(neighborsNodes[i]);
    float edgeLength = centralNodeCoord.dist(neighborNodeCoord);

    bool nodePosOk = false;
    unsigned int j = 1;

    while (!nodePosOk) {
      float ratio = ((centralNodeSize.getW() / 2) + j++ * neighborNodeSize.getW()) / edgeLength;
      Coord newNodeCoord = centralNodeCoord + ratio * (neighborNodeCoord - centralNodeCoord);
      neighborhoodNodesNewLayoutBB[i] = BoundingBox(Coord(newNodeCoord.getX() - neighborNodeSize.getW() / 2, newNodeCoord.getY() - neighborNodeSize.getH() / 2, 0),
                                                    Coord(newNodeCoord.getX() + neighborNodeSize.getW() / 2, newNodeCoord.getY() + neighborNodeSize.getH() / 2, 0));
      nodePosOk = true;

      if (i > 0) {
        for (unsigned int k = 0; k < i; ++k) {
          nodePosOk = nodePosOk && !neighborhoodNodesNewLayoutBB[i].intersect(neighborhoodNodesNewLayoutBB[k]) && !neighborhoodNodesNewLayoutBB[i].intersect(centralNodeBB);
        }
      }

      if (nodePosOk) {
        neighborhoodGraphCircleLayout->setNodeValue(neighborsNodes[i], newNodeCoord);
      }
    }
  }

  // Fix the position of the remaining nodes
  for (tlp::node sample : extraNodes) {
    // By definition there is no need to update the boudning box as the samples are INSIDE an individual
    const tlp::node owner = neighborhoodGraph->getOwner(sample);

    const tlp::Coord initial_sample_pos = neighborhoodGraphLayout->getNodeValue(sample);
    const tlp::Coord initial_owner_pos = neighborhoodGraphLayout->getNodeValue(owner);
    const tlp::Coord current_owner_pos = neighborhoodGraphCircleLayout->getNodeValue(owner);

    const tlp::Coord current_sample_pos = initial_sample_pos + (current_owner_pos - initial_owner_pos);

    neighborhoodGraphCircleLayout->setNodeValue(sample, current_sample_pos);
  }

  /*
   // Currently edges are not drawn
  edge e;
  forEach(e, neighborhoodGraph->getEdges()) {
    const std::pair<node, node>& eEnds = neighborhoodGraph->ends(e);
    node srcNode = eEnds.first;
    node tgtNode = eEnds.second;
    Coord finalBendsCoord;

    if (srcNode != neighborhoodGraphCentralNode) {
      Coord srcNodeCoord = neighborhoodGraphCircleLayout->getNodeValue(srcNode);
      finalBendsCoord = circleCenter + (srcNodeCoord - circleCenter) / 2.f;
    }
    else {
      Coord tgtNodeCoord = neighborhoodGraphCircleLayout->getNodeValue(tgtNode);
      finalBendsCoord = circleCenter + (tgtNodeCoord - circleCenter) / 2.f;
    }

    vector<Coord> edgesBends = neighborhoodGraphLayout->getEdgeValue(e);

    for (size_t i = 0 ; i < edgesBends.size() ; ++i) {
      edgesBends[i] = finalBendsCoord;
    }

    neighborhoodGraphCircleLayout->setEdgeValue(e, edgesBends);
  }
    */
}

void BiometricNeighborhoodHighlighter::selectIndividualOrSample(tlp::node n){
	buildBiometricNeighborhoodGraph(n, originalGraph);
}

void BiometricNeighborhoodHighlighter::morphCircleAlpha(unsigned char startA, unsigned endA, int nbAnimationSteps) {
  startAlpha = startA;
  endAlpha = endA;
  nbAnimSteps = nbAnimationSteps;
  QTimeLine timeLine(500);
  timeLine.setFrameRange(0, nbAnimSteps);
  connect(&timeLine, SIGNAL(frameChanged(int)), this, SLOT(morphCircleAlphaAnimStep(int)));
  timeLine.start();

  while (timeLine.state() != QTimeLine::NotRunning) {
    QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
  }
}

void BiometricNeighborhoodHighlighter::morphCircleAlphaAnimStep(int animStep) {
  circleAlphaValue = static_cast< unsigned char >(startAlpha + animStep / static_cast< float >(nbAnimSteps) * (endAlpha - startAlpha));
  glWidget->redraw();
}

float BiometricNeighborhoodHighlighter::computeBiometricNeighborhoodGraphRadius(LayoutProperty *neighborhoodGraphLayoutProp) {
  float radius = 0;
  node n;
  Coord centralNodeCoord = neighborhoodGraphLayoutProp->getNodeValue(neighborhoodGraphCentralNode);
  for (tlp::node n : neighborhoodGraph->nodes()) {
    Coord nodeCoord = neighborhoodGraphLayoutProp->getNodeValue(n);
    Size nodeSize = originalGlGraphComposite->getInputData()->getElementSize()->getNodeValue(n);
    float dist = centralNodeCoord.dist(nodeCoord) + nodeSize.getW();

    if (dist > radius) {
      radius = dist;
    }
  }
  return radius;
}

void BiometricNeighborhoodHighlighter::checkIfGraphHasChanged() {
  if (glWidget->getScene()->getGlGraphComposite() && glWidget->getScene()->getGlGraphComposite()->getGraph() != originalGraph) {
    neighborhoodGraphCentralNode = node();
    centralNodeLocked = false;
    circleLayoutSet = false;
    cleanupBiometricNeighborhoodGraph();
    originalGraph = glWidget->getScene()->getGlGraphComposite()->getGraph();
    originalGlGraphComposite = glWidget->getScene()->getGlGraphComposite();
  }
}

bool BiometricNeighborhoodHighlighter::draw(GlMainWidget *glMainWidget) {

  checkIfGraphHasChanged();

  if (neighborhoodGraphCentralNode.isValid() && glBiometricNeighborhoodGraph != NULL) {

    Camera *camera = &(glMainWidget->getScene()->getLayer("Main")->getCamera());
    camera->initGl();

    glLineWidth(1.0);
    glPointSize(1.0);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glDepthFunc(GL_LEQUAL);
    glPolygonMode(GL_FRONT, GL_FILL);
    glColorMask(1, 1, 1, 1);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glIndexMask(UINT_MAX);
    glClearStencil(0xFFFF);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    glClear(GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);

    GlCircle circle(
                    circleCenter, 
                    computeBiometricNeighborhoodGraphRadius(neighborhoodGraphLayout), 
                    Color(0, 0, 0, circleAlphaValue),
                    Color(128, 128, 128, circleAlphaValue), 
                    true, 
                    true, 
                    0.0, 
                    60);
    circle.draw(0, 0);

    GlGraphRenderingParameters renderingParameters = glWidget->getScene()->getGlGraphComposite()->getRenderingParameters();
    renderingParameters.setNodesStencil(1);
    renderingParameters.setNodesLabelStencil(1);
    renderingParameters.setDisplayEdges(configWidget->isdisplayEdgesCBChecked());
    glBiometricNeighborhoodGraph->setRenderingParameters(renderingParameters);
    glBiometricNeighborhoodGraph->draw(10, camera);
  }

  return true;
}
}
