/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

/**
 *
 * This file is part of Tulip (www.tulip-software.org)
 *
 * Authors: David Auber and the Tulip development Team
 * from LaBRI, University of Bordeaux
 *
 * Tulip is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Tulip is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 */

#ifndef NEIGHBOURHOODHIGHLIGHTERINTERACTOR_H_
#define NEIGHBOURHOODHIGHLIGHTERINTERACTOR_H_

#include "BiometricNeighborhoodHighlighterConfigWidget.h"
#include "BiometricNodeNeighborhoodView.h"
#include "StandardInteractorPriority.h"

#include "bpg/interactors/PowerGraphInteractor.h"

#include <QObject>

#include <tulip/GLInteractor.h>
#include <tulip/GlGraphComposite.h>
#include <tulip/GlLayer.h>
#include <tulip/GlScene.h>

namespace tlp {

class AdditionalGlSceneAnimation;

class BiometricNeighborhoodHighlighter : public GLInteractorComponent
{

  Q_OBJECT

public:
  BiometricNeighborhoodHighlighter();

  BiometricNeighborhoodHighlighter(const BiometricNeighborhoodHighlighter &neighborhoodHighlighter);

  ~BiometricNeighborhoodHighlighter();

  bool eventFilter(QObject *widget, QEvent *e);

  bool draw(GlMainWidget *glMainWidget);

  void viewChanged(View *view);
  void focusNode(tlp::node n);

  void setConfigWidget(BiometricNeighborhoodHighlighterConfigWidget *configWidget) {
    this->configWidget = configWidget;
  }

  void addPgInteractor(PowerGraphInteractor *i) {
    _pgInteractor = i;
  }



signals:
  void individualOrSampleSelected(tlp::node);

public slots:

  void updateBiometricNeighborhoodGraph();

  void morphCircleAlphaAnimStep(int animStep);

  void selectIndividualOrSample(tlp::node);

private:
  node selectNodeInOriginalGraph(GlMainWidget *glWidget, int x, int y);

  void buildBiometricNeighborhoodGraph(node n, Graph *g);

  void computeBiometricNeighborhoodGraphCircleLayout();

  float computeBiometricNeighborhoodGraphRadius(LayoutProperty *neighborhoodGraphLayoutProp);

  void cleanupBiometricNeighborhoodGraph();

  bool selectInAugmentedDisplayGraph(const int x, const int y, SelectedEntity &);

  void updateBiometricNeighborhoodGraphLayoutAndColors();

  void updateGlBiometricNeighborhoodGraph();

  void computeBiometricNeighborhoodGraphBoundingBoxes();

  void performZoomAndPan(const BoundingBox &destBB, AdditionalGlSceneAnimation *additionalAnimation = NULL);

  void morphCircleAlpha(unsigned char startAlpha, unsigned endAlpha, int nbAnimationSteps = 40);

  void checkIfGraphHasChanged();

  Graph *originalGraph;
  GlGraphComposite *originalGlGraphComposite;
  node selectedNode;
  BiometricNodeNeighborhoodView *neighborhoodGraph;
  GlGraphComposite *glBiometricNeighborhoodGraph;
  Coord circleCenter;

  node neighborhoodGraphCentralNode;
  LayoutProperty *neighborhoodGraphLayout;
  LayoutProperty *neighborhoodGraphCircleLayout;
  LayoutProperty *neighborhoodGraphOriginalLayout;
  ColorProperty *neighborhoodGraphColors;
  ColorProperty *neighborhoodGraphBackupColors;

  bool centralNodeLocked;
  bool circleLayoutSet;
  GlMainWidget *glWidget;
  node selectedBiometricNeighborNode;
  unsigned int neighborhoodDist;

  BiometricNeighborhoodHighlighterConfigWidget *configWidget;

  BoundingBox neighborhoodGraphCircleLayoutBB, neighborhoodGraphOriginalLayoutBB;

  unsigned char circleAlphaValue;
  unsigned char startAlpha, endAlpha;
  int nbAnimSteps;


  PowerGraphInteractor *_pgInteractor = nullptr;
};


/*@{*/
/** \file
 *  \brief  Tulip Node BiometricNeighbourhood Highlighter

 * This interactor plugin allow to get information regarding the neighbourhood of a node by highlighting
 * the nodes connected to it. A "Bring and Go" feature is also implemented allowing to navigate
 */
class BiometricNeighborhoodHighlighterInteractor : public GLInteractorComposite
{

public:
  PLUGININFORMATION("BiometricNeighborhoodHighlighterInteractor", "Antoine Lambert", "19/05/2009", "Node neighborhood highlighter", "1.0", "Navigation")

  BiometricNeighborhoodHighlighterInteractor(const PluginContext *);

  virtual ~BiometricNeighborhoodHighlighterInteractor();

BiometricNeighborhoodHighlighter& highlighter() {
	return _neighborhoodHighlighter;
}
  
  void construct();

  QWidget *configurationWidget() const {
    return configWidget;
  }

  unsigned int priority() const {
    return StandardInteractorPriority::NeighborhoodHighlighter;
  }

  void addPgInteractor(PowerGraphInteractor *i) {
    _neighborhoodHighlighter.addPgInteractor(i);
  }

  virtual bool isCompatible(const std::string &viewName) const;

  // PowerGraphInteractor * pgInteractor() {return &_pgInteractor;}
  BiometricNeighborhoodHighlighter *neighInteractor() { return &_neighborhoodHighlighter; }

private:
  BiometricNeighborhoodHighlighterConfigWidget *configWidget;
  BiometricNeighborhoodHighlighter _neighborhoodHighlighter;
};
}

#endif /* NEIGHBOURHOODHIGHLIGHTERINTERACTOR_H_ */
