/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */


#include "BiometricNodeNeighborhoodView.h"


#include <climits>
#include <tulip/ForEach.h>
#include <tulip/Iterator.h>
#include <tulip/BooleanProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/ColorProperty.h>

#include "bpg/common.h"


using namespace std;

template < typename T >
class StlVectorIterator : public Iterator< T >
{

public:
  explicit StlVectorIterator(const vector< T > &stlVector)
  : stlVector(stlVector), curIdx(0) {}

  T next() {
    return stlVector[curIdx++];
  }

  bool hasNext() {
    return curIdx < stlVector.size();
  }

private:
  vector< T > stlVector;
  unsigned int curIdx;
};

BiometricNodeNeighborhoodView::BiometricNodeNeighborhoodView(Graph *graph, node n)
: GraphDecorator(graph), centralNode(n), root{graph->getRoot()} {
  graphViewNodes.push_back(n);
  getNeighbors(n);
}

void BiometricNodeNeighborhoodView::getNeighbors(node n) {

  assert(graphViewNodes.size() == 1);
  assert(graphViewEdges.size() == 0);

  std::set< tlp::node > nodesToAdd;

  /*
    In destructor
   Need to remove the artifical edges previously added
    root->delEdges(graphViewEdges);
    graphViewEdges.clear();
*/

  auto isQueryAccepted = root->getProperty< tlp::BooleanProperty >(PROP_IS_QUERY_ACCEPTED);
  auto query2gallery = root->getSubGraph(LAYER_QUERY2GALLERY);


  if (isIndividual(root, n)) {
    type = NeighborhoodType::IndividualCenter;
    const node &indiv = n;
    std::cout << "Individual selected => need to search samples linked to it as well as its owners" << std::endl;
    assert(query2gallery->isElement(n));

    edge e;
    forEach (e, query2gallery->getInEdges(indiv)) {
      if (isQueryAccepted->getEdgeValue(e)) {
        const node &query = query2gallery->source(e);
        assert(isSample(root, query));
        nodesToAdd.insert(query);
#if 1
        // Also add the individuals
        nodesToAdd.insert(::getOwner(root, query));
#endif
        // graphViewEdges.push_back(this->addEdge(query, indiv));
        //
      }
    }
  } else if (isSample(root, n)) {
    type = NeighborhoodType::SampleCenter;
    // OK correct behavior
    const node &query = n;
    std::cout << "Sample selected => need to search individuals linked to it" << std::endl;
    assert(query2gallery->isElement(query));

    edge e;
    forEach (e, query2gallery->getOutEdges(query)) {
      if (isQueryAccepted->getEdgeValue(e)) {
        const node &indiv = query2gallery->target(e);
        assert(isIndividual(root, indiv));
        nodesToAdd.insert(indiv);
        //       graphViewEdges.push_back(this->addEdge(query, indiv));
      }
    }
  }

  std::copy(nodesToAdd.begin(), nodesToAdd.end(), std::back_inserter(graphViewNodes));
}


bool BiometricNodeNeighborhoodView::nodeIsSample(tlp::node n) const {
  return isSample(root, n);
}

tlp::node BiometricNodeNeighborhoodView::getOwner(tlp::node n) const {
  return ::getOwner(root, n);
}

bool BiometricNodeNeighborhoodView::isElement(const node n) const {
  return find(graphViewNodes.begin(), graphViewNodes.end(), n) != graphViewNodes.end();
}


unsigned int BiometricNodeNeighborhoodView::nodePos(const node n) const {
  auto nbNodes = graphViewNodes.size();

  for (unsigned int i = 0; i < nbNodes; ++i)
    if (graphViewNodes[i] == n)
      return i;

  return UINT_MAX;
}


unsigned int BiometricNodeNeighborhoodView::edgePos(const edge e) const {
  auto nbEdges = graphViewEdges.size();

  for (unsigned int i = 0; i < nbEdges; ++i)
    if (graphViewEdges[i] == e)
      return i;

  return UINT_MAX;
}



bool BiometricNodeNeighborhoodView::isElement(const edge e) const {
  return find(graphViewEdges.begin(), graphViewEdges.end(), e) != graphViewEdges.end();
}

Iterator< node > *BiometricNodeNeighborhoodView::getNodes() const {
  return new StlVectorIterator< node >(graphViewNodes);
}

Iterator< node > *BiometricNodeNeighborhoodView::getInNodes(const node n) const {
  vector< node > inNodes;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (target(graphViewEdges[i]) == n) {
      inNodes.push_back(source(graphViewEdges[i]));
    }
  }

  return new StlVectorIterator< node >(inNodes);
}

Iterator< node > *BiometricNodeNeighborhoodView::getOutNodes(const node n) const {
  vector< node > outNodes;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (source(graphViewEdges[i]) == n) {
      outNodes.push_back(target(graphViewEdges[i]));
    }
  }

  return new StlVectorIterator< node >(outNodes);
}

Iterator< node > *BiometricNodeNeighborhoodView::getInOutNodes(const node n) const {
  vector< node > inNodes;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (target(graphViewEdges[i]) == n) {
      inNodes.push_back(source(graphViewEdges[i]));
    }
  }

  vector< node > outNodes;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (source(graphViewEdges[i]) == n) {
      outNodes.push_back(target(graphViewEdges[i]));
    }
  }

  inNodes.insert(inNodes.end(), outNodes.begin(), outNodes.end());
  return new StlVectorIterator< node >(inNodes);
}

Iterator< edge > *BiometricNodeNeighborhoodView::getEdges() const {
  return new StlVectorIterator< edge >(graphViewEdges);
}

Iterator< edge > *BiometricNodeNeighborhoodView::getOutEdges(const node n) const {
  vector< edge > outEdges;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (source(graphViewEdges[i]) == n) {
      outEdges.push_back(graphViewEdges[i]);
    }
  }

  return new StlVectorIterator< edge >(outEdges);
}

Iterator< edge > *BiometricNodeNeighborhoodView::getInOutEdges(const node n) const {
  vector< edge > inEdges;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (target(graphViewEdges[i]) == n) {
      inEdges.push_back(graphViewEdges[i]);
    }
  }

  vector< edge > outEdges;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (source(graphViewEdges[i]) == n) {
      outEdges.push_back(graphViewEdges[i]);
    }
  }

  inEdges.insert(inEdges.end(), outEdges.begin(), outEdges.end());
  return new StlVectorIterator< edge >(inEdges);
}

Iterator< edge > *BiometricNodeNeighborhoodView::getInEdges(const node n) const {
  vector< edge > inEdges;

  for (unsigned int i = 0; i < graphViewEdges.size(); ++i) {
    if (target(graphViewEdges[i]) == n) {
      inEdges.push_back(graphViewEdges[i]);
    }
  }

  return new StlVectorIterator< edge >(inEdges);
}
