/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#ifndef BIOMETRICPOWERGRAPHWINDOW_H
#define BIOMETRICPOWERGRAPHWINDOW_H

#include <memory>

#include <QMainWindow>
#include <QSortFilterProxyModel>
#include <QtHelp/QHelpEngine>
#include <QtHelp/QHelpIndexWidget>
#include <QDockWidget>
#include <QtHelp/QHelpContentWidget>
#include <QTextBrowser>

#include <tulip/MouseInteractors.h>
#include <tulip/MouseSelectionEditor.h>
#include <tulip/View.h>

#include "bpg/powergraphmanager.h"

#include "bpg/datasetinformation.h"
#include "bpg/graphparameters.h"
#include "bpg/PowerGraphForm.h"
#include "bpg/GenuineScoreEvolution.h"
#include "bpg/ImpostorScoreEvolution.h"
#include "bpg/SamplesWidget.h"

#include <QSettings>


class HelpBrowser : public QTextBrowser
{
public:
    HelpBrowser(QHelpEngine* helpEngine, QWidget* parent = 0);
    QVariant loadResource (int type, const QUrl& name);
private:
    QHelpEngine* helpEngine;
};



class BiometricDatabaseForm;
class ROCForm;

class QSortFilterProxyModel;

namespace Ui {
class BiometricPowerGraphWindow;
}


class SampleFilterProxyModel : public QSortFilterProxyModel
{
  Q_OBJECT
public:
  explicit SampleFilterProxyModel(QObject *parent = nullptr)
  : QSortFilterProxyModel(parent), state{StateFilter::none} {
  }

  bool filterAcceptsRow(int sourceRow, const QModelIndex &sourceParent) const override {
    QModelIndex galleryIndex = sourceModel()->index(sourceRow, 0, sourceParent);
    QModelIndex probeIndex = sourceModel()->index(sourceRow, 1, sourceParent);
    QModelIndex sampleIndex = sourceModel()->index(sourceRow, 2, sourceParent);

    QString gallery = sourceModel()->data(galleryIndex).toString();
    QString probe = sourceModel()->data(probeIndex).toString();
    QString sample = sourceModel()->data(sampleIndex).toString();


    switch (state) {
      case StateFilter::gallery:
        return gallery.contains(galleryRegExp);
      case StateFilter::probe:
        return probe.contains(probeRegExp);
      case StateFilter::sample:
        return probe.contains(probeRegExp) && sample.contains(sampleRegExp);
      default:
        return true;
    }
  }

public slots:
  void setGalleryFilter(const QString &regExp) {
    this->galleryRegExp.setPattern(regExp);
    state = StateFilter::gallery;

    invalidateFilter();
  }
  void setProbeFilter(const QString &regExp) {
    this->probeRegExp.setPattern(regExp);
    state = StateFilter::probe;


    invalidateFilter();
  }
  void setSampleFilter(const QString &probeRegExp, const QString &sampleRegExp) {
    this->probeRegExp.setPattern(probeRegExp);
    this->sampleRegExp.setPattern(sampleRegExp);
    state = StateFilter::sample;

    invalidateFilter();
  }

  void setNoFilter() {
    state = StateFilter::none;
    invalidateFilter();
  }

private:
  enum class StateFilter
  {
    gallery,
    probe,
    sample,
    none
  };

  StateFilter state;

  QRegExp galleryRegExp;
  QRegExp probeRegExp;
  QRegExp sampleRegExp;
};


class BiometricPowerGraphWindow : public QMainWindow
{
  Q_OBJECT

public:
  explicit BiometricPowerGraphWindow(QWidget *parent = 0);
  ~BiometricPowerGraphWindow();


private:
  Ui::BiometricPowerGraphWindow *ui;
  PowerGraphManager _pgManager;

  DatasetInformation *_datasetInfo;
  GraphParameters *_graphParameters;
  BiometricDatabaseForm *_biometricDatabaseForm;
  GenuineScoreEvolution *_genuineScoreEvolution;
  ImpostorScoreEvolution *_impostorScoreEvolution;
  SamplesWidget *_samplesWidget;


  // ROC display management
  ROCForm *_rocWidget;

  // Pgraph display management
  PowerGraphForm *_pgraphWidget;
  std::unique_ptr< tlp::MouseNKeysNavigator > _pgraphNavigator;


  // Filter used for the database of scores
  SampleFilterProxyModel _filterDatabase;

  tlp::Graph *const dummyGraph;

  bool infoToggled = true;

  QSettings _settings;

public:
  void importCSV(const QString fname, const bool isScore, const QString thumbnailFile, const bool openEER=false);
  void exportTLP(const QString fname);


  void dragEnterEvent(QDragEnterEvent *event) override;
  void dropEvent(QDropEvent *event) override;

private slots:
  void openClusterColorLegend();
  void openSampleColorLegend();

  void openZooGraph();
  void openZooPlot();
  void openRoc();
  void openFnmrFmrGraph();
  void openScoresDistribution();

  void layoutImported();

void helpContent();

  /**
      The user wants to import a CSV file
     * @brief importCSV
     */
  void importCSV();
  void exportTLP();

  /**
     * The CSV file has been successfully imported
     * @brief fileImported
     */
  void fileImported(const bool openEER=false);


  // Interractions on the ROC curve
  void setRocNodeHovered(tlp::node n, int x, int y);
  void setRocNodeSelected(tlp::node n);
  void setDecisionThreshold(double thr);


  // Interractions on the power graph
  void setPgraphSampleHovered(tlp::node n, int x, int y);
  void setPgraphIndividualHovered(tlp::node n, int x, int y);


  // Interactions on the score table (and maybe graph later)
  void setFilterGallery(const QString gallery_id);
  void setFilterProbe(const QString probe_id);
  void setFilterSample(const QString probe_id, const QString sample_id);
  void clearFilters();


  void showAbout();
  void enterWhatsthis();

  void setInfoToggled(bool);
  void individualOrSampleSelected(tlp::node);
};

#endif // BIOMETRICPOWERGRAPHWINDOW_H
