/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include "bpg/BiometricDatabaseForm.h"
#include "ui_BiometricDatabaseForm.h"

#include <QMenu>
#include <QPoint>

#include <iostream>
#include <cassert>
#include <QMessageBox>

BiometricDatabaseForm::BiometricDatabaseForm(QWidget *parent)
: QGroupBox(parent),
  ui(new Ui::BiometricDatabaseForm) {
  ui->setupUi(this);

  { // Manage contextual menu
    ui->biometricDatabaseTableView->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(
      ui->biometricDatabaseTableView,
      SIGNAL(customContextMenuRequested(const QPoint &)),
      this,
      SLOT(showContextMenu(const QPoint &)));
  }

  {
      // Normal connections
      connect(
                  ui->searchButton,
                  SIGNAL(clicked()),
                  this,
                  SLOT(applySearch()));
  }
}

BiometricDatabaseForm::~BiometricDatabaseForm() {
  delete ui;
}


void BiometricDatabaseForm::setModel(QAbstractItemModel *model) {
  assert(model);
  _model = model;
  ui->biometricDatabaseTableView->setModel(_model);
}







void BiometricDatabaseForm::applySearch()
{
    const QString value = ui->searchValue->text();
    if (ui->radioButtonGallery->isChecked()) {
        emit filterGallery(value);
    }
    else if (ui->radioButtonProbe->isChecked()) {
        emit filterProbe(value);
    }
    else {
        QMessageBox::critical(this, "Impossible to filter", "You need to first select the type of filter");
    }
}

void BiometricDatabaseForm::showContextMenu(const QPoint &pos) {
  assert(_model);

  // for most widgets
  //QPoint globalPos = ui->biometricDatabaseTableView->mapToGlobal(pos);
  // for QAbstractScrollArea and derived classes you would use:
  const QPoint globalPos = ui->biometricDatabaseTableView->viewport()->mapToGlobal(pos);

  // Acquire the information necessary for the menu
  const QModelIndex idx = ui->biometricDatabaseTableView->indexAt(pos);

  if (!idx.isValid())
    return; // We have not selected a real thing

  const QModelIndex galleryIdx = _model->index(idx.row(), 0);
  const QModelIndex probeIdx = _model->index(idx.row(), 1);
  const QModelIndex sampleIdx = _model->index(idx.row(), 2);

  const QString gallery_value = _model->data(galleryIdx).toString();
  const QString probe_value = _model->data(probeIdx).toString();
  const QString sample_value = _model->data(sampleIdx).toString();


  // Build the specific menu
  QMenu myMenu;
  QAction *filter_gallery = is_gallery_filtered ? nullptr : myMenu.addAction(QString("Filter gallery user (") + gallery_value + QString(")"));
  QAction *clear_gallery = !is_gallery_filtered ? nullptr : myMenu.addAction(QString("Clear filter for gallery user (") + gallery_value + QString(")"));

  QAction *filter_probe = is_probe_filtered ? nullptr : myMenu.addAction(QString("Filter probe user (") + probe_value + QString(")"));
  QAction *clear_probe = !is_probe_filtered ? nullptr : myMenu.addAction(QString("Clear filter for probe user (") + probe_value + QString(")"));

  QAction *filter_sample = is_sample_filtered ? nullptr : myMenu.addAction(QString("Filter sample (") + probe_value + QString(",") + sample_value + QString(")"));
  QAction *clear_sample = !is_sample_filtered ? nullptr : myMenu.addAction(QString("Clear filter for sample (") + probe_value + QString(",") + sample_value + QString(")"));

  myMenu.addSeparator();

  QAction *focus_gallery = nullptr, *focus_probe=nullptr, *focus_sample=nullptr;
  if (this->_pgDisplayed) {
      focus_gallery = myMenu.addAction(QString("Focus on gallery individual ") +  gallery_value);
      focus_probe = myMenu.addAction(QString("Focus on probe individual ") +  probe_value);
      focus_sample = myMenu.addAction(QString("Focuse on sample ") + sample_value + " of individual " + probe_value);
  }
  QAction *selectedItem = myMenu.exec(globalPos);


  if (nullptr != selectedItem) {

    // Filtering cases
    if (selectedItem == filter_gallery) {
      is_gallery_filtered = true;
      is_probe_filtered = false;
      is_sample_filtered = false;

      emit filterGallery(gallery_value);
    } else if (selectedItem == filter_probe) {
      is_gallery_filtered = false;
      is_probe_filtered = true;
      is_sample_filtered = false;

      emit filterProbe(probe_value);
    } else if (selectedItem == filter_sample) {
      is_gallery_filtered = false;
      is_probe_filtered = false;
      is_sample_filtered = true;

      emit filterSample(probe_value, sample_value);
    }

    // Cleanup cases
    else if (selectedItem == clear_gallery || selectedItem == clear_probe || selectedItem == clear_sample) {
        is_gallery_filtered = is_probe_filtered = is_sample_filtered = false;
      emit clearFilters();
    }

    else if (selectedItem == focus_gallery) {
    	emit requestFocusIndividual(gallery_value.toInt());
    }

    else if (selectedItem == focus_probe) {
    	emit requestFocusIndividual(probe_value.toInt());
    }
    else if (selectedItem == focus_sample) {
    	emit requestFocusSample(probe_value.toInt(), sample_value.toInt());
    }

  } else {
    // nothing was chosen
  }
}
