/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include "bpg/PowerGraphForm.h"
#include "ui_PowerGraphForm.h"

#include <QString>
#include <QSplitter>
#include <qfiledialog.h>
#include <QMenu>

#include <tulip/GlGraphComposite.h>
#include <tulip/GlGraphRenderingParameters.h>
#include <tulip/MouseInteractors.h>
#include <tulip/MouseSelectionEditor.h>
#include <tulip/View.h>
#include <tulip/QtGlSceneZoomAndPanAnimator.h>

#include <tulip/GlMainView.h>
#include <tulip/GlMainWidget.h>
#include <tulip/GlGraphComposite.h>
#include <tulip/GlGraphRenderingParameters.h>

#include "../../plugins/interactor/BiometricNeighborhoodHighlighter/BiometricNeighborhoodHighlighterInteractor.h"
#include "bpg/common.h"
#include "bpg/ExecutionMonitor.h"

PowerGraphForm::PowerGraphForm(QWidget *parent, tlp::Graph *dummyGraph)
: QWidget(parent),
  ui(new Ui::PowerGraphForm),
_layoutGraph{nullptr},
graph{nullptr}{
  ui->setupUi(this);

#ifdef _WIN32
  ui->buttonUp->setText("^");
  ui->buttonDown->setText("v");
  ui->buttonLeft->setText("<");
  ui->buttonRight->setText(">");
  ui->buttonZoomIn->setText("+");
  ui->buttonZoomOut->setText("-");
#endif

  constexpr auto node_link_name = "Node Link Diagram view";
  constexpr auto interactor_name = "BiometricNeighborhoodHighlighterInteractor";

/*
   for(const auto &plugin: tlp::PluginLister::availablePlugins()) {
       std::cout << plugin << std::endl;
   }
   */

  assert(tlp::PluginLister::instance()->pluginExists(node_link_name) && "node link daigram view not found");
  assert(tlp::PluginLister::instance()->pluginExists(interactor_name) && "BiometricNeighborhoodHighlighterInteractor not found");

  _pgraphView = tlp::PluginLister::instance()->getPluginObject< tlp::View >(node_link_name, nullptr);
  assert(nullptr != _pgraphView && "node link diagram not found");

  _pgraphView->setupUi();

  QList< tlp::Interactor * > interactors;
  neighInteractor = static_cast< BiometricNeighborhoodHighlighterInteractor * >(tlp::PluginLister::instance()->getPluginObject< tlp::Interactor >(interactor_name, nullptr));
  assert(neighInteractor&& "interactor not found");
  neighInteractor->addPgInteractor(this->pgInteractor());
  interactors.push_back(neighInteractor);
  _pgraphView->setInteractors(interactors);
  _pgraphView->setCurrentInteractor(interactors[0]);

  /*
    _pgraphWidget = static_cast<tlp::GlMainView*>(_pgraphView)->getGlMainWidget();
    assert(nullptr != _pgraphWidget);
    */

  this->ui->labelFMR->setText("-");
  this->ui->labelFNMR->setText("-");
  this->ui->labelThreshold->setText("-");

  this->setPowerGraph(dummyGraph);
  this->ui->powerGraphLayout->addWidget(_pgraphView->graphicsView() /*_pgraphWidget*/);

  _pgraphView->graphicsView()->setContextMenuPolicy(Qt::CustomContextMenu);
  connect(_pgraphView->graphicsView(), SIGNAL(customContextMenuRequested(const QPoint&)),
      this, SLOT(showContextMenu(const QPoint&)));

  connect(ui->pushButtonInfo, SIGNAL(toggled(bool)),
          this, SIGNAL(infoToggled(bool)));
  connect(ui->showEdgesCheckbox, SIGNAL(toggled(bool)),
          this, SLOT(showEdges(bool)));
    connect(this->ui->pushButtonImportLayout, SIGNAL(clicked()),
          this, SLOT(clickOnImportLayout()));
  connect(this->ui->pushButtonExportLayout, SIGNAL(clicked()),
	  this, SIGNAL(clickOnExportLayout()));


  connect(neighInteractor->neighInteractor(), SIGNAL(individualOrSampleSelected(tlp::node)),
          this, SIGNAL(individualOrSampleSelected(tlp::node)));

  connect(this, SIGNAL(selectIndividualOrSample(tlp::node)),
          neighInteractor->neighInteractor(), SLOT(selectIndividualOrSample(tlp::node)));

   connect(this->ui->buttonCenter, SIGNAL(clicked()),
           this, SLOT(centerView()));


   connect(this->ui->buttonLeft, SIGNAL(clicked()), this, SLOT(moveLeft()));
   connect(this->ui->buttonRight, SIGNAL(clicked()), this, SLOT(moveRight()));
   connect(this->ui->buttonUp, SIGNAL(clicked()), this, SLOT(moveUp()));
   connect(this->ui->buttonDown, SIGNAL(clicked()), this, SLOT(moveDown()));
   connect(this->ui->buttonZoomIn, SIGNAL(clicked()), this, SLOT(zoomIn()));
   connect(this->ui->buttonZoomOut, SIGNAL(clicked()), this, SLOT(zoomOut()));


  this->setEnabled(false);
}

PowerGraphForm::~PowerGraphForm() {
  delete ui;
}


/**
 * @brief PowerGraphForm::setPowerGraph
 * @param graph
 * @param threshold
 * @param fnmr
 * @param fmr
 *
 *
 * @todo See if this method for renderingParameters would be usefull: setDisplayFilteringProperty(BooleanProperty *filteringProperty) instead of modifying the graph
 */
void PowerGraphForm::setPowerGraph(tlp::Graph *graph, double threshold, double fnmr, double fmr) {

  if (graph->numberOfNodes()>0) execution_monitor::start("Draw::draw");
  this->graph = graph;

  if (graph->numberOfNodes() > 0) {
	  this->setEnabled(true);
  }
  else {
  	this->setEnabled(false);
  }

  this->ui->labelFMR->setText(QString::number(100 * fmr, 'g', 3) + "%");
  this->ui->labelFNMR->setText(QString::number(100 * fnmr, 'g', 3) + "%");
  this->ui->labelThreshold->setText(QString::number(threshold));


  _renderingParameters.setDisplayEdges(ui->showEdgesCheckbox->isChecked());
  _renderingParameters.setViewArrow(false);
  _renderingParameters.setEdgeColorInterpolate(false);
  _renderingParameters.setEdgeSizeInterpolate(false);
  _renderingParameters.setEdge3D(false);
  _renderingParameters.setEdgesMaxSizeToNodesSize(true);
  _renderingParameters.setLabelScaled(true);
  _renderingParameters.setNodesLabelStencil(0x00);
  _renderingParameters.setViewNodeLabel(true);
  _renderingParameters.setEdgesStencil(0x0f);
  _renderingParameters.setNodesStencil(0x01);
  _renderingParameters.setMetaNodesStencil(0x00);
  _renderingParameters.setElementZOrdered(false);
  _renderingParameters.setElementOrdered(true);
  _renderingParameters.setElementOrderedDescending(false);
  _renderingParameters.setElementOrderingProperty(graph->getProperty< tlp::DoubleProperty >("order"));
  _renderingParameters.setEdgeFrontDisplay(false);

  _pgraphView->setGraph(graph);

  static_cast< tlp::GlMainView * >(_pgraphView)
    ->getGlMainWidget()
    ->getScene()
    ->getGlGraphComposite()
    ->setRenderingParameters(_renderingParameters);
  if (graph->numberOfNodes()>0) execution_monitor::stop();
}




void PowerGraphForm::showEdges(bool state) {
  execution_monitor::start("Draw::draw");
  _renderingParameters.setDisplayEdges(state);
  static_cast< tlp::GlMainView * >(_pgraphView)
    ->getGlMainWidget()
    ->getScene()
    ->getGlGraphComposite()
    ->setRenderingParameters(_renderingParameters);
  _pgraphView->draw();
  execution_monitor::stop();
  execution_monitor::print();
}



void PowerGraphForm::clickOnImportLayout() {
        QString fname = QFileDialog::getOpenFileName(this,
                                                 "Select exported file to retreive its coordinates",
                                                 _settings.value("import layout/fname", "").toString(),
                                                 "Tulip files (*.tlp *.tlp.gz *.tlpb *.tlpb.gz)");

    if (fname.isEmpty()) {
      _layoutGraph = nullptr;
      this->ui->pushButtonImportLayout->setChecked(false);
    }
    else {
        _settings.setValue("import layout/fname", fname);
      _layoutGraph = tlp::loadGraph(fname.toStdString());
      this->ui->pushButtonImportLayout->setChecked( _layoutGraph != nullptr);
    }


    emit layoutImported();
}





void PowerGraphForm::showContextMenu(const QPoint& p) {
    QMenu myMenu(_pgraphView->graphicsView());


    myMenu.addAction(
                "Take a snapshot",
                this,
                SLOT(takeScreenshot()),
                0
                );
    myMenu.addAction(
		    "Center view",
		    this,
		    SLOT(centerView()),
		    0
		    );

    myMenu.move(_pgraphView->graphicsView()->mapToGlobal(p));
    myMenu.exec();
}



void PowerGraphForm::takeScreenshot() {
 QString fname = QFileDialog::getSaveFileName(this,
                                                 "Take a snapshot",
                                                 _settings.value("save screenshot/fname", "").toString(),
                                                 "PNG files (*.png)");

    if (!fname.isEmpty()) {
        constexpr int image_width = 4000;
        constexpr int image_height = 4000;
        static_cast< tlp::GlMainView * >(_pgraphView)
          ->getGlMainWidget()->createPicture(fname.toStdString(), image_width, image_height);
    }
}




void PowerGraphForm::focusIndividual(int id) {
	if (!graph) return;

	const tlp::node indiv = getNodeForIndividual(graph->getRoot(), id);
	const auto viewLayout = graph->getProperty<tlp::LayoutProperty>("viewLayout");
	const auto viewSize = graph->getProperty<tlp::LayoutProperty>("viewLayout");

	const tlp::Coord c = viewLayout->getNodeValue(indiv);
	const tlp::Size s = viewSize->getNodeValue(indiv);

	// Manage the movment
	tlp::BoundingBox bb;
	bb.expand(tlp::Vec3f{c.x() + s.width()/2, c.y()+s.height()/2, c.z()});
	bb.expand(tlp::Vec3f{c.x() - s.width()/2, c.y()+s.height()/2, c.z()});
	bb.expand(tlp::Vec3f{c.x() + s.width()/2, c.y()-s.height()/2, c.z()});
	bb.expand(tlp::Vec3f{c.x() - s.width()/2, c.y()-s.height()/2, c.z()});



	tlp::QtGlSceneZoomAndPanAnimator animator(static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget(), bb);
	animator.animateZoomAndPan();

	emit selectIndividualOrSample(indiv);
}

void PowerGraphForm::focusSample(int individual, int sample) {
	if (!graph) return;

	const tlp::node indiv = getNodeForIndividual(graph->getRoot(), individual);
	const tlp::node smpl = getNodeSampleOfIndividualForId(graph->getRoot(), indiv, sample);

	const auto viewLayout = graph->getProperty<tlp::LayoutProperty>("viewLayout");
	const auto viewSize = graph->getProperty<tlp::LayoutProperty>("viewLayout");

	const tlp::Coord c = viewLayout->getNodeValue(smpl);
	const tlp::Size s = viewSize->getNodeValue(smpl);

	// Manage the movment
	tlp::BoundingBox bb;
	bb.expand(tlp::Vec3f{c.x() + s.width()/2, c.y()+s.height()/2, c.z()});
	bb.expand(tlp::Vec3f{c.x() - s.width()/2, c.y()+s.height()/2, c.z()});
	bb.expand(tlp::Vec3f{c.x() + s.width()/2, c.y()-s.height()/2, c.z()});
	bb.expand(tlp::Vec3f{c.x() - s.width()/2, c.y()-s.height()/2, c.z()});



	tlp::QtGlSceneZoomAndPanAnimator animator(static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget(), bb);
	animator.animateZoomAndPan();

	emit selectIndividualOrSample(smpl);
}



void PowerGraphForm::moveLeft(){
    std::cout << "left <" << std::endl;

    int delta = 1;
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->getScene()->translateCamera(
                static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->screenToViewport(delta * 2),
                0,
                0);
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->draw();
}
void PowerGraphForm::moveRight(){
    int delta = 1;
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->getScene()->translateCamera(
                static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->screenToViewport(-1*delta * 2),
                0,
                0);
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->draw();

}
void PowerGraphForm::moveUp(){
    int delta = 1;
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->getScene()->translateCamera(
                0,
                static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->screenToViewport(-1*delta * 2),
                0);
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->draw();

}
void PowerGraphForm::moveDown(){
    int delta = 1;
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->getScene()->translateCamera(
                0,
                static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->screenToViewport(+1*delta * 2),
                0);
    static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->draw();
}

void PowerGraphForm::centerView() {
  static_cast< tlp::GlMainView * >(_pgraphView)->centerView(false);
}

void PowerGraphForm::zoomIn() {
    Camera& camera = static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->getScene()->getGraphCamera();
    auto cameraScaleFactor = camera.getZoomFactor();
    static const auto scaleFactor = 1/0.75;
    camera.setZoomFactor(cameraScaleFactor * scaleFactor);
        static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->draw();

}


void PowerGraphForm::zoomOut() {
    Camera& camera = static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->getScene()->getGraphCamera();
    auto cameraScaleFactor = camera.getZoomFactor();
    static const auto scaleFactor = 0.75;
    camera.setZoomFactor(cameraScaleFactor * scaleFactor);
        static_cast< tlp::GlMainView * >(_pgraphView)->getGlMainWidget()->draw();

}



