/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include "bpg/ROCForm.h"
#include "ui_ROCForm.h"
#include "bpg/common.h"


#include <tulip/GlMainWidget.h>
#include <tulip/GlGraphComposite.h>


#include <QWidget>


ROCForm::ROCForm(QWidget *parent)
: QWidget(parent),
  ui(new Ui::ROCForm),
  _root{nullptr},
  _rocLayer{nullptr},
  _rocNavigator(new ROCInteractor) {
  ui->setupUi(this);

  _rocLayer = getGlMainWidget()
                ->getScene()
                ->createLayer("Main");

  getGlMainWidget()->installEventFilter(_rocNavigator.get());

  ui->actionEER_choice->setStatusTip(ui->actionEER_choice->toolTip() + "(" + ui->actionEER_choice->shortcut().toString() + ")");
  ui->eerButton->setAction(ui->actionEER_choice);


  ui->actionEER_choice->setEnabled(false);
  ui->manualThresholdButton->setEnabled(false);
  ui->manualThresholdInput->setEnabled(false);

  connect(
    ui->actionEER_choice, SIGNAL(triggered()),
    this, SLOT(eerSelected()));

  connect(
    ui->manualThresholdButton, SIGNAL(pressed()),
    this, SLOT(manualThresholdButtonSelected()));


}

ROCForm::~ROCForm() {
  delete ui;
}




tlp::GlMainWidget *ROCForm::getGlMainWidget() {
  return ui->rocWidget;
}


void ROCForm::manualThresholdButtonSelected() {
  emit this->manualThresholdValueSelected(ui->manualThresholdInput->value());
}

void ROCForm::eerSelected() {
  if (nullptr == _root)
    return;

  const tlp::node EERnode = getEERRocNode(_root);
  assert(EERnode.isValid());
  assert(_root->isElement(EERnode));


  std::cout << "EER node " << EERnode << std::endl;

  emit rocNodeSelected(EERnode);
}

void ROCForm::setGraph(tlp::Graph *roc) {

  _root = roc->getRoot();
  auto rocWidget = getGlMainWidget();

  // Add the graph to the display stuff
  _rocLayer->addGraph(roc, "ROC");

  // Force some parameters
  auto renderingParameters = getGlMainWidget()->getScene()->getGlGraphComposite()->getRenderingParametersPointer();
  assert(renderingParameters);

  renderingParameters->setViewArrow(false);
  renderingParameters->setEdgeColorInterpolate(false);
  renderingParameters->setEdgeSizeInterpolate(false);
  renderingParameters->setLabelScaled(true);
  renderingParameters->setNodesLabelStencil(0x0000);
  renderingParameters->setViewNodeLabel(true);
  renderingParameters->setEdgesStencil(0xFFFF);
  renderingParameters->setNodesStencil(0xFFFF);
  renderingParameters->setMetaNodesStencil(0x0000);
  renderingParameters->setElementZOrdered(true);


  // Do rendering
  rocWidget->show();
  rocWidget->centerScene();
  rocWidget->draw();


  ui->actionEER_choice->setEnabled(true);

  // Get the minimum and maxium score values
  double min = 0, max = 0;
  bool first = true;
  tlp::DoubleProperty *thresholdProperty = roc->getProperty< tlp::DoubleProperty >(PROP_THR);
  for (tlp::node n : roc->nodes()) {
    const double value = thresholdProperty->getNodeValue(n);
    if (first) {
      min = value;
      max = value;
      first = false;
    } else {
      min = std::min(min, value);
      max = std::max(max, value);
    }
  }

  // go a bit lower than the limits
  constexpr unsigned int nbThresholds = 1000;
  const double delta = (max - min) / static_cast< double >(nbThresholds);
  min = min - delta;
  max = max + delta;

  this->ui->manualThresholdInput->setMinimum(min);
  this->ui->manualThresholdInput->setMaximum(max);
  this->ui->manualThresholdInput->setDecimals(4);
  this->ui->manualThresholdInput->setSingleStep(delta);

  this->ui->manualThresholdButton->setEnabled(true);
  this->ui->manualThresholdInput->setEnabled(true);
}
