#include "bpg/ScoreDistributionWindow.h"
#include "ui_ScoreDistributionWindow.h"

#include "bpg/common.h"
#include "bpg/powergraphmanager.h"
#include <QMenu>
#include <QPoint>
#include <QFileDialog>
#include <qwt_plot_renderer.h>
#include <qwt_legend.h>

#include <map>
#include <thread>
#include <omp.h>

ScoreDistributionWindow::ScoreDistributionWindow(QWidget *parent, PowerGraphManager &pg) :
		QMainWindow(parent),
		ui(new Ui::ScoreDistributionWindow)
{
		ui->setupUi(this);


		// Build the graph
		{
				// lambda to compute the histogram
				auto treat = [&](decltype(_intraHist) &histo, const std::vector<double> & list, const double min, const double bucket_size) {
						// Assign each element to a bucket
						for(const double & val: list) {
								const size_t bucket = std::min<size_t>((val - min) / bucket_size, nb_buckets-1);
								assert (bucket<nb_buckets);
								++histo[bucket];
						}

						// Transform them as probabilities
						const size_t nb_elements = list.size();
						std::for_each(
										std::begin(histo),
										std::end(histo),
										[&](double& count){
											count /= nb_elements;
										}
						);

				};

				// Compute histgram limits
				const auto scores = pg.getIntraInterScores();
				const std::vector<double> &intraScores = scores.first;
				const std::vector<double> &interScores = scores.second;

				const auto minmax1 = std::minmax_element(intraScores.begin(), intraScores.end());
				const auto minmax2 = std::minmax_element(interScores.begin(), interScores.end());
				const double min = std::min<double>( *minmax1.first, *minmax2.first);
				const double max = std::max<double>( *minmax1.second, *minmax2.second);
				const double width = max-min;
				const double bucket_size = width/nb_buckets;

				// Really compute the histograms

				std::thread t1{treat, std::ref(_intraHist), std::ref(intraScores), min, bucket_size};
				std::thread t2{treat, std::ref(_interHist), std::ref(interScores), min, bucket_size};
				t1.join();
				t2.join();

				// Build the graphical objects

#pragma omp parallel for
				for (size_t i=0; i< nb_buckets; ++i) {
						_intraData[i] = QwtIntervalSample(_intraHist[i], min + i*width, min + (i+1)*width);
						_interData[i] = QwtIntervalSample(_interHist[i], min + i*width, min + (i+1)*width);
				}

				_intraCurve.setSamples(_intraData);
				_interCurve.setSamples(_interData);


				QPen interPen(QColor(241,163,64), 4);
				QPen intraPen(QColor(153,142,195), 4);

				_intraCurve.setPen(interPen);
				_interCurve.setPen(intraPen);

				ui->widgetPlot->setStyleSheet("background-color: white");
				ui->widgetPlot->setAxisTitle(QwtPlot::xBottom, "Comparison score");
				ui->widgetPlot->setAxisTitle(QwtPlot::yLeft, "Probability");

				_intraCurve.setTitle("Intra scores");
				_interCurve.setTitle("Inter scores");

				const auto selectedStyle = QwtPlotHistogram::HistogramStyle::Outline;
				_intraCurve.setStyle(selectedStyle);
				_interCurve.setStyle(selectedStyle);

				_intraCurve.attach(ui->widgetPlot);
				_interCurve.attach(ui->widgetPlot);

				ui->widgetPlot->insertLegend(new QwtLegend(ui->widgetPlot), QwtPlot::TopLegend);


		}
}

ScoreDistributionWindow::~ScoreDistributionWindow()
{
		delete ui;
}
