
/**
 * Copyright University Bordeaux - CNRS UMR 5800 
 * contributor(s) : Romain Giot (2017)
 * 
 * Romain.Giot@u-bordeaux.fr
 * 
 * Biometric Power Graph (bpg) is a computer program whose purpose is to evaluate biometric authentication
 * systems by displaying with a graph-based representation after having transformed the data.
 * 
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * http://www.cecill.info. 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include "bpg/BiometricDatabaseForm.h"
#include "bpg/BiometricDatabaseModel.h"
#include "bpg/biometricpowergraphwindow.h"
#include "bpg/ClusterColorLegend.h"
#include "bpg/SampleColorLegend.h"
#include "bpg/common.h"
#include "bpg/ImportDialog.h"
#include "bpg/interactors/PowerGraphInteractor.h"
#include "bpg/ROCForm.h"
#include "bpg/ROCWindow.h"
#include "bpg/ZooGraphWindow.h"
#include "bpg/ZooPlotWindow.h"
#include "bpg/ThresholdToErrorsWindow.h"
#include "bpg/ScoreDistributionWindow.h"
#include "bpg/ExecutionMonitor.h"

#include <iostream>

#include <QDirIterator>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <QSizePolicy>
#include <QSortFilterProxyModel>
#include <QSplitter>
#include <QTabWidget>
#include "qtooltip.h"
#include <QUrl>
#include <QWhatsThis>
#include <string>
#include <tulip/GlGraphComposite.h>
#include <tulip/GlGraphRenderingParameters.h>
#include <tulip/TlpTools.h>

#include "ui_biometricpowergraphwindow.h"

namespace std {
string to_string(const char *buff) {
  return string(buff);
}
} /* end of namespace std */



HelpBrowser::HelpBrowser(QHelpEngine* helpEngine,
                         QWidget* parent):QTextBrowser(parent),
                         helpEngine(helpEngine)
{
}

QVariant HelpBrowser::loadResource(int type, const QUrl &name){
    if (name.scheme() == "qthelp")
        return QVariant(helpEngine->fileData(name));
    else
        return QTextBrowser::loadResource(type, name);
}




/**
 * Verifies if source and current are compatible in order to let current use the same layout of source
 */
bool areGraphsCompatibles(tlp::Graph * const source, tlp::Graph * const current) {
    if (source == nullptr || current == nullptr) return false;
    if (source->getSubGraph(LAYER_SAMPLES)->numberOfNodes() != current->getSubGraph(LAYER_SAMPLES)->numberOfNodes()) return false;
   // if (source->getSubGraph(LAYER_USERS)->getSubGraph("original")->numberOfNodes() != current->getSubGraph(LAYER_USERS)->getSubGraph("original")->numberOfNodes()) return false;
    if (source->getSubGraph(LAYER_QUERY2GALLERY)->numberOfEdges() != current->getSubGraph(LAYER_QUERY2GALLERY)->numberOfEdges()) return false;


    tlp::Graph * const q2g = current->getSubGraph(LAYER_QUERY2GALLERY);
    for(tlp::edge e: source->getSubGraph(LAYER_QUERY2GALLERY)->edges()) {
        if (!q2g->isElement(e)) return false;
    }
    return true;
}


BiometricPowerGraphWindow::BiometricPowerGraphWindow(QWidget *parent)
: QMainWindow(parent),
  ui(new Ui::BiometricPowerGraphWindow),

  _datasetInfo{new DatasetInformation(this)},
  _graphParameters{new GraphParameters(this)},
  _biometricDatabaseForm{new BiometricDatabaseForm(this)},

  _rocWidget(nullptr),

  _pgraphWidget{nullptr},
  dummyGraph{tlp::newGraph()}

{
  // Build the UI
  ui->setupUi(this);

  // create the visualization panels
  {
	  _rocWidget = new ROCForm(this);
	  _pgraphWidget = new PowerGraphForm(this, dummyGraph);

	  constexpr bool displayScoreEvolution = true;
	  _genuineScoreEvolution = new GenuineScoreEvolution(this);
	  _impostorScoreEvolution = new ImpostorScoreEvolution(this);

	  _samplesWidget = new SamplesWidget(this);


    QSplitter *splitter = new QSplitter(nullptr);
    QSplitter *leftSplitter = new QSplitter(nullptr);
    QSplitter *rightSplitter = new QSplitter(nullptr);
    QWidget *leftContainer = new QWidget();
    QWidget *rightContainer = new QWidget();
    QTabWidget *tabs = new QTabWidget();

    QVBoxLayout *leftContainer_layout = new QVBoxLayout;
    QVBoxLayout *rightContainer_layout = new QVBoxLayout;

    leftContainer->setLayout(leftContainer_layout);
    rightContainer->setLayout(rightContainer_layout);

    leftSplitter->setOrientation(Qt::Vertical);
    leftContainer_layout->addWidget(leftSplitter);
    leftSplitter->addWidget(_rocWidget);


    rightContainer_layout->addWidget(_pgraphWidget);

    _filterDatabase.setSourceModel(&getUniqueQtBiometricDataModel());
    _biometricDatabaseForm->setModel(&_filterDatabase);

    tabs->addTab(_datasetInfo, "Dataset information");
    tabs->addTab(_graphParameters, "Graph parameters");
    tabs->addTab(_biometricDatabaseForm, "Biometric database");
    tabs->addTab(_samplesWidget, "Samples");
    tabs->addTab(_genuineScoreEvolution, "Genuine scores");
    tabs->addTab(_impostorScoreEvolution, "Impostor scores");


    leftSplitter->addWidget(tabs);

    ui->innerSpace->addWidget(splitter);
    splitter->addWidget(leftContainer);
    splitter->addWidget(rightContainer);
  }

  // Add the interractors
  {

    /*
        _pgraphNavigator.reset(new PowerGraphInteractor);
        static_cast<tlp::GlMainWidget*>(_pgraphWidget)->installEventFilter(_pgraphNavigator.get());
        */
  }

  ui->actionExport->setEnabled(false);
  ui->actionHelpContent->setShortcut(QKeySequence::HelpContents);

  connect(ui->actionHelpContent, SIGNAL(triggered()), SLOT(helpContent()));

  // Manage the signal/slot stuff
  connect(ui->actionImport, SIGNAL(triggered()), SLOT(importCSV()));
  connect(ui->actionExport, SIGNAL(triggered()), SLOT(exportTLP()));

  connect(ui->actionAbout, SIGNAL(triggered()), SLOT(showAbout()));
  connect(ui->actionWhatsthis, SIGNAL(triggered()), SLOT(enterWhatsthis()));

  connect(ui->actionCluster_colors, SIGNAL(triggered()), SLOT(openClusterColorLegend()));
  connect(ui->actionSamples_colors, SIGNAL(triggered()), SLOT(openSampleColorLegend()));


  connect(ui->actionROC_curve, SIGNAL(triggered()), SLOT(openRoc()));
  connect(ui->actionZoo_plot, SIGNAL(triggered()), SLOT(openZooPlot()));
  connect(ui->actionZoo_graph, SIGNAL(triggered()), SLOT(openZooGraph()));
  connect(ui->actionFNMR_FMR_curve, SIGNAL(triggered()), SLOT(openFnmrFmrGraph()));
  connect(ui->actionScore_distribution, SIGNAL(triggered()), SLOT(openScoresDistribution()));


  ui->actionROC_curve->setEnabled(false);
  ui->actionZoo_plot->setEnabled(false);
  ui->actionZoo_graph->setEnabled(false);
  ui->actionFNMR_FMR_curve->setEnabled(false);
  ui->actionScore_distribution->setEnabled(false);

  connect(&_rocWidget->getROCInteractor(), SIGNAL(rocNodeHovered(tlp::node, int, int)), SLOT(setRocNodeHovered(tlp::node, int, int)));
  connect(&_rocWidget->getROCInteractor(), SIGNAL(rocNodeSelected(tlp::node)), SLOT(setRocNodeSelected(tlp::node)));
  connect(_rocWidget, SIGNAL(rocNodeSelected(tlp::node)), SLOT(setRocNodeSelected(tlp::node)));
  connect(_rocWidget, SIGNAL(manualThresholdValueSelected(double)),
          this, SLOT(setDecisionThreshold(double)));

  connect(_pgraphWidget->pgInteractor(), SIGNAL(pgraphSampleHovered(tlp::node, int, int)),
          SLOT(setPgraphSampleHovered(tlp::node, int, int)));
  connect(_pgraphWidget->pgInteractor(), SIGNAL(pgraphIndividualHovered(tlp::node, int, int)),
          SLOT(setPgraphIndividualHovered(tlp::node, int, int)));
  connect(_pgraphWidget, SIGNAL(infoToggled(bool)),
          this, SLOT(setInfoToggled(bool)));
  connect(_pgraphWidget, SIGNAL(individualOrSampleSelected(tlp::node)),
          this, SLOT(individualOrSampleSelected(tlp::node)));
  connect(_pgraphWidget, SIGNAL(layoutImported()),
          this, SLOT(layoutImported()));
  connect(_pgraphWidget, SIGNAL(clickOnExportLayout()), SLOT(exportTLP()));

  _pgraphWidget->setEnabled(false);

  // Connection to manage the filtering of the biometric table of scores
  connect(_biometricDatabaseForm, SIGNAL(filterGallery(const QString)),
          this, SLOT(setFilterGallery(const QString)));
  connect(_biometricDatabaseForm, SIGNAL(filterProbe(const QString)),
          this, SLOT(setFilterProbe(const QString)));
  connect(_biometricDatabaseForm, SIGNAL(filterSample(const QString, const QString)),
          this, SLOT(setFilterSample(const QString, const QString)));
  connect(_biometricDatabaseForm, SIGNAL(clearFilters()),
          this, SLOT(clearFilters()));

  connect(_biometricDatabaseForm, SIGNAL(requestFocusSample(int, int)),
	  _pgraphWidget, SLOT(focusSample(int, int)));
  connect(_biometricDatabaseForm, SIGNAL(requestFocusIndividual(int)),
	  _pgraphWidget, SLOT(focusIndividual(int)));


  setAcceptDrops(true);
}

BiometricPowerGraphWindow::~BiometricPowerGraphWindow() {
  delete ui;
}


// Manage drag and drop (mainly to import a CSV file)

void BiometricPowerGraphWindow::dragEnterEvent(QDragEnterEvent *event) {
  //   std::cout << event->format() << std::endl;
  if (event->mimeData()->hasFormat("text/csv") || event->mimeData()->hasFormat("x-special/gnome-icon-list"))
    event->acceptProposedAction();
}

void BiometricPowerGraphWindow::dropEvent(QDropEvent *event) {
  const bool isScore = true; // TODO open a dialog box to ask if it is a score file or not
#if 0
    const QString fname = event->mimeData()->text()
                            .remove(QRegExp("^file://"))
                            .remove(QRegExp("\\s*$"));
#else
  const QString fname = QUrl(event->mimeData()->text()).toLocalFile();
#endif

  std::cout << "*" << fname.toStdString() << "*" << std::endl;
  importCSV(fname, isScore, "");

  event->acceptProposedAction();
}

// Slots

void BiometricPowerGraphWindow::openClusterColorLegend() {
  ClusterColorLegend legend;
  legend.exec();
}


void BiometricPowerGraphWindow::openSampleColorLegend() {
  SampleColorLegend * legend = new SampleColorLegend();
  legend->setAttribute(Qt::WA_DeleteOnClose);
  legend->show();
}




// TODO manage slow loading of big files
void BiometricPowerGraphWindow::importCSV() {
  ImportDialog diag;

  if (diag.exec() == QDialog::Accepted) {
    const QString fileName = diag.filename();
    const bool isScore = diag.isScore();
    const QString thumbnailFname = diag.thumbnailFname();

    importCSV(fileName, isScore, thumbnailFname);
  }
}


bool fileExists(QString path) {
  QFileInfo check_file(path);
  // check if file exists and if yes: Is it really a file and no directory?
  return check_file.exists() && check_file.isFile();
}

void BiometricPowerGraphWindow::exportTLP() {

  bool loopAgain = false;

  do {
    QString fname = QFileDialog::getSaveFileName(this,
                                                 "Select file to export Tulip data structure",
                                                 "",
                                                 "*.tlp, *.tlp.gz, *.tlpb,*.tlpb.gz");

    if (fname.isEmpty())
      break;


    if (fileExists(fname)) {
      int res = QMessageBox::question(this,
                                      "Attention, " + fname + " already exists",
                                      "Do you want to replace " + fname + "?",
                                      QMessageBox::Yes,
                                      QMessageBox::No);

      if (res == QMessageBox::No) {
        loopAgain = true;
      }
    }

    if (!loopAgain) {
      exportTLP(fname);
    }
  } while (loopAgain);
}


void BiometricPowerGraphWindow::exportTLP(const QString fname) {
  QApplication::setOverrideCursor(Qt::WaitCursor);
  this->setEnabled(false);

  bool res = tlp::saveGraph(_pgManager.getRoot(), fname.toStdString());

  if (!res) {
    QMessageBox::critical(this, "Error while exporting", "For an unknown reason, the graph data structure has not been exported.", QMessageBox::Ok);
  }
    QApplication::restoreOverrideCursor();
    this->setEnabled(true);
}

void BiometricPowerGraphWindow::importCSV(const QString fileName, const bool isScore, const QString thumbnailFname, const bool openEER) {
  QApplication::setOverrideCursor(Qt::WaitCursor);

  // Unload current graph
  _rocWidget->setGraph(dummyGraph);
  _pgraphWidget->setPowerGraph(dummyGraph);
  _pgManager.safelyRealeaseData();

  this->setEnabled(false);

  try {
    bool res = false;
    res = _pgManager.importCSV(fileName.toLocal8Bit().constData(), isScore, thumbnailFname.toLocal8Bit().constData());
    _datasetInfo->setFilename(fileName.toStdString());


    if (!res) {
      QApplication::restoreOverrideCursor();
      QMessageBox msgbox;
      msgbox.setText("[ERROR] Impossible to load the file. Check the error console.");
      msgbox.exec();
      this->setEnabled(true);

    } else {
      fileImported(openEER);
    }

  } catch (const std::exception &e) {
    QMessageBox::critical(this, "Unable to load file", e.what());
    QApplication::restoreOverrideCursor();
    this->setEnabled(true);
  }
}


void BiometricPowerGraphWindow::fileImported(const bool openEER) {
  // Retreive the ROC graph
  tlp::Graph *roc = _pgManager.getROCGraph();
  std::cout << roc << roc->getName() << std::endl;

  tlp::Graph *_root = roc->getRoot();
  int nbIntra = 0, nbInter = 0;
  _root->getAttribute(ATTR_NB_INTRA, nbIntra);
  _root->getAttribute(ATTR_NB_INTER, nbInter);

  double EERthr = -1, EERval = -1;
  _root->getAttribute(ATTR_EER, EERval);
  _root->getAttribute(ATTR_THR, EERthr);

  _datasetInfo->setNbSamples(getNumberOfSamples(_root));
  _datasetInfo->setNbUsers(getNumberOfUsers(_root));
  _datasetInfo->setNbInterScores(nbInter);
  _datasetInfo->setNbIntraScores(nbIntra);
  _datasetInfo->setNbDifferentScores(roc->numberOfNodes());
  _datasetInfo->setEERThr(EERthr);
  _datasetInfo->setEERVal(EERval);

  _rocWidget->setGraph(roc);

  QApplication::restoreOverrideCursor();
  this->setEnabled(true);

  ui->actionExport->setEnabled(true);
  ui->actionROC_curve->setEnabled(true);
  ui->actionZoo_plot->setEnabled(true);
  ui->actionFNMR_FMR_curve->setEnabled(true);
  ui->actionScore_distribution->setEnabled(true);

  if (openEER) _rocWidget->eerSelected();

}

void BiometricPowerGraphWindow::setInfoToggled(bool val) {
  this->infoToggled = val;
}


void BiometricPowerGraphWindow::setPgraphIndividualHovered(tlp::node n, int /*x*/, int /*y*/) {
  if (infoToggled == false)
    return;

  auto res = _pgManager.getInformationFromIndividualNode(n);

  std::string representation = "Individual ";
  representation += std::to_string(res.id);
  representation += std::to_string(" FNMR: ") + std::to_string(100 * res.FNMR) + std::to_string("%");
  representation += std::to_string(" IN FMR: ") + std::to_string(100 * res.FMR) + std::to_string("%");
  representation += std::to_string(" OUT FMR: ") + std::to_string(100 * res.FMR2) + std::to_string("%");
  statusBar()->showMessage(QString::fromStdString(representation));

	if (_genuineScoreEvolution) {
	  const std::vector< tlp::edge > & intraEdges = getIntraComparisonsToUser(_pgManager.getRoot(), n);
	  std::vector< double > scores(intraEdges.size());
	  std::vector< bool > acceptance(intraEdges.size());

	  /** Seems to be sorted
	  std::for_each(
			  intraEdges.begin(),
			  intraEdges.end(),
			  [&](const tlp::edge e) {
				std::cout << _pgManager.getRoot()->source(e) << std::endl;;
			  }
			  );
	  */
	  std::transform(
		intraEdges.begin(),
		intraEdges.end(),
		scores.begin(),
		[&](const tlp::edge e) {
		  return getComparisonScore(_pgManager.getRoot(), e);
		});

	  std::transform(
		scores.begin(),
		scores.end(),
		acceptance.begin(),
		[&](const double score) {
		  return isComparisonScoreAccepted(score, _pgManager.getDecisionThreshold());
		});
		
		
	   _genuineScoreEvolution->setGenuineScores(
		res.id,
		n,
		_pgManager.getDecisionThreshold(),
		scores,
		acceptance);
	}

  std::vector< std::tuple<unsigned int, unsigned int, std::string> > thumbnails;
  const tlp::StringProperty * thumbnailProp = _pgManager.getRoot()->getProperty<tlp::StringProperty>("thumbnail");
  const tlp::IntegerProperty * sampleIdProp = _pgManager.getRoot()->getProperty<tlp::IntegerProperty>(PROP_SAMPLE);
  tlp::node sample;
  for(tlp::node sample: getSamples(_pgManager.getRoot(), n)){
  thumbnails.push_back({
                           res.id,
                           sampleIdProp->getNodeValue(sample),
                           thumbnailProp->getNodeValue(sample)
                       });
  }
  _samplesWidget->setThumbnails(thumbnails);
}

void BiometricPowerGraphWindow::setPgraphSampleHovered(tlp::node n, int /*x*/, int /*y*/) {
  if (infoToggled == false)
    return;

  auto res = _pgManager.getInformationFromSampleNode(n);

  {
	  std::string status_representation = "Sample ";
	  status_representation += std::to_string(res.sampleId);
	  status_representation += " of user " ;
	  status_representation += std::to_string(res.ownerId);

      if (0 == res.nbGenuineComparison) {
          status_representation += "[Never used for genuine verification]";
      }
      else {
        status_representation += res.isVerified ?( ". [Properly verified]") : (". [Wrongly rejected]");
      }

      if (0 == res.nbImpersonnateComparison) {
          status_representation += "[Never used for impersonnation]";
      }
      else {
        status_representation += ". Impersonation rate of ";
        status_representation += std::to_string(100. * res.nbImpersonnation / res.nbImpersonnateComparison) + "%";
      }
	  statusBar()->showMessage(QString::fromStdString(status_representation));
  }


  std::vector< std::tuple<unsigned int, unsigned int, std::string> > thumbnails;
  const tlp::StringProperty * thumbnailProp = _pgManager.getRoot()->getProperty<tlp::StringProperty>("thumbnail");
  thumbnails.push_back({
                           res.ownerId,
                           res.sampleId,
                           thumbnailProp->getNodeValue(n)
                       });
  _samplesWidget->setThumbnails(thumbnails);



  const tlp::node owner = getOwner(_pgManager.getRoot(), n);
  std::vector< double > scores;
  std::vector< bool > acceptances;
  tlp::edge e;
  forEach(e, _pgManager.getRoot()->getSubGraph(LAYER_QUERY2GALLERY)->getOutEdges(n)) {
    const tlp::node reference = _pgManager.getRoot()->target(e);
    if (reference == owner) continue;

    assert(_pgManager.getRoot()->getSubGraph(LAYER_SAMPLE2OWNER)->isElement(reference));

    const double score    = getComparisonScore(_pgManager.getRoot(), e);
    const bool acceptance = isComparisonScoreAccepted(score, _pgManager.getDecisionThreshold());

    scores.push_back(score);
    acceptances.push_back(acceptance);

    _impostorScoreEvolution->setImpostorScores(
            res.ownerId,
            res.sampleId,
            owner,
            n,
            _pgManager.getDecisionThreshold(),
            scores,
            acceptances
            );
  }





}

void BiometricPowerGraphWindow::setRocNodeHovered(tlp::node n, int x, int y) {
  const RocNodeInformation res = _pgManager.getInformationFromRocNode(n);
  std::string representation = "";

  representation = representation + std::to_string("Threshold: ") + std::to_string(res.threshold) + std::to_string(" ");
  representation = representation + std::to_string("FNMR: ") + std::to_string(res.FNMR) + std::to_string(" ");
  representation = representation + std::to_string("FMR: ") + std::to_string(res.FMR) + std::to_string(" ");

  QToolTip::showText(
    QPoint{x, y}, // XXX  BUG need to translate these points
    QString::fromStdString(representation),
    _rocWidget);

  statusBar()->showMessage(QString::fromStdString(representation));
}


void BiometricPowerGraphWindow::setRocNodeSelected(tlp::node n) {
  const bool graphsCompatible = areGraphsCompatibles(_pgManager.getRoot(), _pgraphWidget->getLayoutToCloneGraph());
  if (!graphsCompatible && _pgraphWidget->getLayoutToCloneGraph() != nullptr) {
    QMessageBox::warning(this, "Impossible to use imported layout", "It is impossible to use the imported layout. Please double check you are using compatible data.");
  }


  QApplication::setOverrideCursor(Qt::WaitCursor);
  this->setEnabled(false);
  _pgraphWidget->setPowerGraph(dummyGraph, -1, -1, -1);

  const RocNodeInformation res = _pgManager.getInformationFromRocNode(n);


  const bool success = _pgManager.buildDisplayGraphForThreshold(
    res.threshold,
    _graphParameters->getApproximated(),
    graphsCompatible ? _pgraphWidget->getLayoutToCloneGraph() : nullptr,
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FNMR, GraphParameters::PartitioningThresholdKind::Low),
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FNMR, GraphParameters::PartitioningThresholdKind::High),

              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_in, GraphParameters::PartitioningThresholdKind::Low),
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_in, GraphParameters::PartitioningThresholdKind::High),

              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_out, GraphParameters::PartitioningThresholdKind::Low),
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_out, GraphParameters::PartitioningThresholdKind::High)

              );

  if (!success) {
    QApplication::restoreOverrideCursor();
    QMessageBox msgbox;
    msgbox.setText("[ERROR] Impossible to draw the graph. Check the error console.");
    msgbox.exec();
  }

  std::cout << "Computation done" << std::endl;

  // Add the graph to the display stuff
  //_pgraphLayer->addGraph(_pgManager.getDisplayGraph(), "graph");
  _pgraphWidget->setPowerGraph(
    _pgManager.getDisplayGraph(),
    res.threshold,
    res.FNMR,
    res.FMR);

  QApplication::restoreOverrideCursor();
  this->setEnabled(true);
  this->_biometricDatabaseForm->pgIsDisplayed();
    getUniqueQtBiometricDataModel().updateUniqueDecisionThreshold(res.threshold);
    execution_monitor::print();
}


void BiometricPowerGraphWindow::setDecisionThreshold(double thr) {

  const bool graphsCompatible = areGraphsCompatibles(_pgManager.getRoot(), _pgraphWidget->getLayoutToCloneGraph());
  if (!graphsCompatible && _pgraphWidget->getLayoutToCloneGraph() != nullptr) {
    QMessageBox::warning(this, "Impossible to use imported layout", "It is impossible to use the imported layout. Please double check you are using compatible data.");
  }

  _pgraphWidget->setPowerGraph(dummyGraph, -1, -1, -1);
  const bool success = _pgManager.buildDisplayGraphForThreshold(
    thr,
    _graphParameters->getApproximated(),
    graphsCompatible ? _pgraphWidget->getLayoutToCloneGraph() : nullptr,
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FNMR, GraphParameters::PartitioningThresholdKind::Low),
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FNMR, GraphParameters::PartitioningThresholdKind::High),

              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_in, GraphParameters::PartitioningThresholdKind::Low),
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_in, GraphParameters::PartitioningThresholdKind::High),

              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_out, GraphParameters::PartitioningThresholdKind::Low),
              this->_graphParameters->getPartitioningThreshold(GraphParameters::PartitioningThresholdTarget::FMR_out, GraphParameters::PartitioningThresholdKind::High)

              );


  if (!success) {
    QApplication::restoreOverrideCursor();
    QMessageBox msgbox;
    msgbox.setText("[ERROR] Impossible to draw the graph. Check the error console.");
    msgbox.exec();
  }

  // XXX it would probably more efficient to use the ROC curve for that

  double FNMR = -1;
  double FMR = -1;
  std::tie(FNMR, FMR) = computeFNMRandFMRforThreshold(_pgManager.getRoot(), thr);


  // Add the graph to the display stuff
  //_pgraphLayer->addGraph(_pgManager.getDisplayGraph(), "graph");
  _pgraphWidget->setPowerGraph(
    _pgManager.getDisplayGraph(),
    thr,
    FNMR,
    FMR);

  QApplication::restoreOverrideCursor();
  this->setEnabled(true);
  this->_biometricDatabaseForm->pgIsDisplayed();
  getUniqueQtBiometricDataModel().updateUniqueDecisionThreshold(thr);
    execution_monitor::print();
}


void BiometricPowerGraphWindow::layoutImported() {
    if (_pgManager.displayAvailable()){
        setDecisionThreshold(_pgManager.getDecisionThreshold());
    }
}


void BiometricPowerGraphWindow::individualOrSampleSelected(tlp::node n) {
  if (_pgManager.isIndividual(n)) {
    setFilterGallery(QString::number(_pgManager.getIndividualId(n))); // XXX Probe ? GAllery & Probe ?
  } else {
    assert(_pgManager.isSample(n));
    setFilterSample(
      QString::number(_pgManager.getIndividualId(_pgManager.getOwnerOfSample(n))),
      QString::number(_pgManager.getSampleId(n)));
  }
}


// Inform the application a filtering occured

void BiometricPowerGraphWindow::setFilterGallery(const QString gallery_id) {

  _filterDatabase.setGalleryFilter(QString("^") + gallery_id + QString("$"));
  _biometricDatabaseForm->informIsGalleryFiltered();


  /*
   * XXX code supposed to die
  _pgraphWidget->setPowerGraph(dummyGraph);
  _pgManager.setGalleryToFilter(gallery_id.toInt(), true);
  _pgraphWidget->setPowerGraph(_pgManager.getDisplayGraph());
  */
}

void BiometricPowerGraphWindow::setFilterProbe(const QString probe_id) {
  _filterDatabase.setProbeFilter(QString("^") + probe_id + QString("$"));
  _biometricDatabaseForm->informIsProbeFiltered();
}

void BiometricPowerGraphWindow::setFilterSample(const QString probe_id, const QString sample_id) {
  _filterDatabase.setSampleFilter(
    QString("^") + probe_id + QString("$"),
    QString("^") + sample_id + QString("$"));
  _biometricDatabaseForm->informIsSampleFiltered();
}


void BiometricPowerGraphWindow::clearFilters() {
  _filterDatabase.setNoFilter();

  /*
   * XXX code supposed to die
  if(!_pgManager.displayAvailable()) return;
  _pgraphWidget->setPowerGraph(dummyGraph);
  _pgManager.clearFilters();
  _pgraphWidget->setPowerGraph(_pgManager.getDisplayGraph());
  */
}


void BiometricPowerGraphWindow::showAbout() {
  QMessageBox::about(this, "About \"Biometric Power Graph display\"",
                     "Citation:  Romain Giot, \"Evaluation of Biometric Authentication Systems through Visualisation of Partitionned and Bundled Power-graphs\", Information Visualisation 2017 (IV2017), pp 7, 2017.\n\n"
                     "Licence: CECILL\n\n"
                     "Author: Romain Giot <romain.giot@u-bordeaux.fr>\n\n"
                     "Uses:\n"
                     " - Tulip: http://tulip.labri.fr/\n"
                     " - Fast CPP CSV Parser: https://github.com/ben-strasser/fast-cpp-csv-parser\n");
}


void BiometricPowerGraphWindow::enterWhatsthis() {
  QWhatsThis::enterWhatsThisMode();
}


void BiometricPowerGraphWindow::openZooGraph() {
}

void BiometricPowerGraphWindow::openZooPlot() {
    auto window = new ZooPlotWindow(this, _pgManager);
    window->setAttribute(Qt::WA_DeleteOnClose);
    window->show();
}

void BiometricPowerGraphWindow::openRoc() {
  auto window = new ROCWindow(this, _pgManager);
  window->setAttribute(Qt::WA_DeleteOnClose);
  window->show();
}

void BiometricPowerGraphWindow::openFnmrFmrGraph() {
  auto window = new ThresholdToErrorsWindow(this, _pgManager);
  window->setAttribute(Qt::WA_DeleteOnClose);
  window->show();
}

void BiometricPowerGraphWindow::openScoresDistribution() {
  auto window = new ScoreDistributionWindow(this, _pgManager);
  window->setAttribute(Qt::WA_DeleteOnClose);
  window->show();
}




void BiometricPowerGraphWindow::helpContent() {
    static QHelpEngine* helpEngine = nullptr;
    static auto helpWindow = new QDockWidget(tr("Help"), this);

    static const QString fname = QString::fromStdString(tlp::TulipShareDir) + QString::fromStdString("/BiometricPowerGraphVisualisation.qhc");


    if (nullptr == helpEngine){
        auto fileInfo = QFileInfo(fname);
        helpEngine = new QHelpEngine(
                         fname);
        helpEngine->setupData();

        QTabWidget* tWidget = new QTabWidget;
        tWidget->setMaximumWidth(200);
        tWidget->addTab(helpEngine->contentWidget(), "Contents");
        tWidget->addTab(helpEngine->indexWidget(), "Index");

        HelpBrowser *textViewer = new HelpBrowser(helpEngine);
        textViewer->setSource(
                    QUrl("qthelp://org.sphinx.biometricpowergraphvisualisation.0.1/doc/index.html"));
        connect(helpEngine->contentWidget(),
                SIGNAL(linkActivated(QUrl)),
                textViewer, SLOT(setSource(QUrl)));

        connect(helpEngine->indexWidget(),
                SIGNAL(linkActivated(QUrl, QString)),
                textViewer, SLOT(setSource(QUrl)));

        QSplitter *horizSplitter = new QSplitter(Qt::Horizontal);
        horizSplitter->insertWidget(0, tWidget);
        horizSplitter->insertWidget(1, textViewer);
        horizSplitter->hide();

        helpWindow->setWidget(horizSplitter);
        // helpWindow->hide();
        addDockWidget(Qt::RightDockWidgetArea, helpWindow);
    }
    else {
        if (helpWindow->isVisible())
        {
            helpWindow->hide();
        }
        else {
            helpWindow->show();
        }
    }
}
